/*******************************************************\
* File: Hunter.h									    *
* header file for Hunter class							*
*														*
* Programer: Andrew Schlei								*
*														*
* A hunter is an agent that gathers food and lives in	*
* a 2-D world.  A hunter can move around, gather food,	*
* and interact with other agents by either attacking	*
* them, sharing with them, or ignoring them.			*
*														*
* Ver 1.0												*
\*******************************************************/

/*********************************************************************************\
* =====METHODS & FUNCTIONS SUMMARY=====
* 
* CONSTRUCTOR
* 
*	Hunter()
*	pre: none
*	post: none
*
* DESTRUCTOR
*
*	~Hunter()
*	pre: none
*	post: memory used by interactedWith array has been freed
*
* MODIFY METHODS
*
*	void setPos(int x, int y)
*	this method set the x and y position
*	pre: x and y are coordinates in the 2-D world, with x increasing from
*	     left to right and y increasing from top to bottom
*	post: xPos and yPos have changed to reflect agent's new position
*
*	void setPower(int pow)
*   inline 
*   this method sets the agents's power
*   pre: pow is the number to set the agent's power to
*   post: power has been set to pow
*   
*	void setFoodCount(int fc)
*   inline
*   this method sets how much food an agent has
*   pre: fc is the amount of food the agent is to possess
*   post: foodCount has been set to fc
*
*	void setID(int idNum)
*   inline
*   this method sets the agent's id number
*   pre: idNum is the agent's new id number
*   post: hID has been set to idNum
*
*	void setAlive(bool a)
*   inline
*   this method sets the agent's alive status
*   pre: a is the agent's new status (true = alive, false = dead)
*   post: alive has been set to a
*
*	void setStrengthAttack(int str)
*   inline
*   this method is used to set the agent's strength associated with attacking
*   pre: str is the new strength to be assigned to attacking
*   post: strengthAttack has been set to str
*
*	void setStrengthShare(int str)
*   inline
*   this method is used to set the agent's strength associated with sharing
*   pre: str is the new strength to be assigned to sharing
*   post: strengthShare has been set to str
*
*	void setStrengthHoard(int str)
*   inline
*   this method is used to set the agent's strength associated with hoarding
*   pre: str is the new strength to be assigned to hoarding
*   post: strengthHoard has been set to str
*
*	void setCostAttack(int cost)
*   inline
*   this method sets the agent's material cost for attacking
*   pre: cost is the amount of food used up when attacking
*   post: costAttack has been set to cost
*
*	void setUtilityAttack(int util)
*   inline
*   this method set the agent's utility offset for attacking
*   pre: util is the additional amount to adjust strength by when attacking
*   post: utilityAttack has been set to util
*
*	void setUtilityShare(int util)
*   inline
*   this method set the agent's utility offset for sharing
*   pre: util is the additional amount to adjust strength by when sharing
*   post: utilityShare has been set to util
*
*	void setUtilityHoard(int util)
*   inline
*   this method set the agent's utility offset for hoarding
*   pre: util is the additional amount to adjust strength by when hoarding
*   post: utilityHoard has been set to util
*
*	void setVision(int vis)
*	inline
*	this method is used to set the agent's vision range
*	pre: vis is the desired vision range
*	post: vision has been set to vis
*
*	void setMetabolism(int met)
*	inline
*	this method is used to set the agent's metabolism
*	pre: met is the desired metabolism
*	post: metabolism has been set to met
*
*	void createInteract(int numHunters)
*	this method is used to create the interactedWith array
*	pre: numHunters is the number of hunter agents in the simulation
*	post: the agent's interactedWith array has been allocated
*
*	void initInteract(int numHunters)
*	this method initializes the hunters interactedWith array
*	pre: numHunters is the number of hunter agents in the simulation
*	post: the agent's interactedWith array has been initialized to false
*
*	void setInteracted(int i)
*   inline
*   this method is used to adjust the agent's interactedWith array
*   to show that the agent has interacted with another hunter agent
*   pre: i is the id number of the agent that was interacted with
*   post: interactedWith[i] has been set to true, showing that the
*		  agent has interacted with agent i
*
*	void move(int* worldMatrix, int WSX, int WSY, int EMPTY, int FOOD)
*	this method determines how the agents move
*	pre: -worldMatrix is a pointer to a 1 dimensional array that represents
*	      the 2 dimensional world the hunter agents live in
*	     -WSX is the world size in the x dimension (in cells)
*	     -WSY is the world size in the y dimension (in cells)
*	     -EMPTY is the number denoting an empty cell in worldMatrix
*		 -FOOD is the number denoting food in worldMatrix
*	post: worldMatrix has been modified to reflect the new positions of the
*  		  hunter agents.  The agent's xPos and yPos coordinates have also 
*		  been altered to reflect the agent's new position.
*
*	int getScore(int* worldMatrix, int WSX, int EMPTY, int FOOD, int cellX, int cellY)
*	this method is used to find the score for a given cell, which agents use to determine how to move
*	pre: -worldMatrix is a pointer to a 1 dimensional array that represents
*	      the 2 dimensional world the hunter agents live in
*	     -WSX is the world size in the x dimension (in cells)
*	     -EMPTY is the number denoting an empty cell in worldMatrix
*	     -FOOD is the number denoting food in worldMatrix
*	     -cellX is the x coordinate of the cell to be examined
*	     -cellY is the y coordinate of the cell to be examined
*	post: returns score for given cell
*
*	void gather(int* worldMatrix, int WSX, int WSY, int EMPTY, 
*				int FOOD,int FOOD_VALUE)
*	this method is used by the agents to gather food
*	pre: -worldMatrix is a pointer to a 1 dimensional array that represents
* 	      the 2 dimensional world the hunter agents live in
*	     -WSX is the world size in the x dimension (in cells)
*	     -WSY is the world size in the y dimension (in cells)
*	     -EMPTY is the number denoting an empty cell in worldMatrix
*	     -FOOD is the nubmer denoting a cell containing food in worldMatrix
*	     -FOOD_VALUE is the amount of food gained from collecting from a 
*	      food cell
*	post: The agent's foodCount has been adjusted to reflect any food gathered.
*	      worldMatrix has been modified to remove any food cells that were
*		  gathered (by setting them to EMPTY).
*
*	void removeTies()
*   This method is used to remove any ties for the highest strength.  If any
*	strengths are tied for highest, it randomly selects one to increment by
*	one.
*	pre: none
*	post: the agent's action strengths have been adjusted to remove any ties
*
* 
* CONSTANT METHODS
*
*	int getPower()
*   inline
*   pre: none
*   post: returns the agent's power
*
*	int getFoodCount()
*	inline
*	pre: none
*	post: returns the agent's foodCount
*
*	int getStrengthAttack()
*	inline
*	pre: none
*	post: returns the agent's strengthAttack
*
*	int getStrengthShare()
*	inline
*	pre: none
*	post: returns the agent's strengthShare
*
*	int getStrengthHoard()
*	inline
*	pre: none
*	post: returns the agent's strengthHoard
*
*	int getY()
*	inline
*	pre: none
*	post: returns the agent's yPos
*
*	int getX()
*	inline
*	pre: none
*	post: returns the agent's xPos
*
*	int getCostAttack()
*	inline
*	pre: none
*	post: returns the agent's costAttack
*
*	int getUtilityAttack()
*	inline
*	pre: none
*	post: returns the agent's utilityAttack
*
*	int getUtilityShare()
*	inline
*	pre: none
*	post: returns the agent's utilityShare
*
*	int getUtilityHoard()
*	inline
*	pre: none
*	post: returns the agent's utilityHoard
*
*	int getVision()
*	inline
*	pre: none
*	post: returns the agent's vision range
*
*	int getMetabolism()
*	inline
*	pre: none
*	post: returns the agent's metabolism
*
*	bool isAlive()
*	inline
*	pre: none
*	post: returns the agent's alive status
*
*	bool interacted(int i) {return interactedWith[i];}
*	inline
*	pre: i is a hunter id number
*	post: returns true if the agent has interacted with agent i
*
*	bool isAttacking()
*	This method is used to determine if the agent is currently in "attack mode"
*	pre: none
*	post: returns true if the the agent's strengthAttack is its highest strength
*
*	bool isSharing()
*	This method is used to determine if the agent is currently in "share mode"
*	pre: none
*	post: returns true if the the agent's strengthShare is its highest strength
*
*	bool isHoarding(); 
*	This method is used to determine if the agent is currently in "hoard mode"
*	pre: none
*	post: returns true if the the agent's strengthHoard is its highest strength
*
*
* NON-MEMBER FUNCTIONS
*
*	void hunterInteraction(int* world,int WSY,int WSX,Hunter* hunters, int i, ofstream &log)
*	This function is used by hunter i to search for nearby hunters it can
*	interact with.  When it finds a hunter it has not yet interacted with, 
*	hunterInteractionResolver is called.
*	pre: -worldMatrix is a pointer to a 1 dimensional array that represents
*	      the 2 dimensional world the hunter agents live in
*	     -WSX is the world size in the x dimension (in cells)
*	     -WSY is the world size in the y dimension (in cells)
*	     -hunters is a pointer to the array of hunter agents
*	     -i is the id number of the agent currently acting
*	     -log is an open output file stream
*	post: agent i will have interacted with any agents nearby if it hasn't already
*
*	void hunterInteractionResolver(Hunter* hunters, int i, int target, ofstream &log)
*	This function is used to resolve interactions between hunters
*	pre: -hunters is a pointer to the array of hunter agents
*	     -i is the id number of the agent currently acting
*	     -targetID is the id number of the agent being interacted with
*	     -logOutput is an open file output stream
*	post: An action between two agents has been resolved, including adjustments to
*		  action strengths, foodCounts, and the interactedWith arrays.
*
*	void hunterEvolution(Hunter* hunters, int NUM_HUNTERS, int numAlive)
*	This function handles the evolution of the agents.  First, the id numbers
*	of all surviving agents are gathered.  Then each dead agent is replaced
*	with a new agent.  The new agent is created by first choosing two "parents"
*	from among the living agents, and then averaging their power, stengthAttack,
*	stengthShare, and strengthHoard member variables.  It is possible for a single
*	agent to be both "parents"
*	pre: -hunters is a pointer to the array of hunter agents
*	     -NUM_HUNTERS is the total number of hunters in the simulation
*	     -numAlive is the number of agents currently "alive"
*	post: all dead hunter agents have been replaced with new agents, although
*		  they are still dead.
\***********************************************************************************************/

#include <fstream>
#include <iostream>
using namespace std;

#ifndef HUNTER_H
#define HUNTER_H

class Hunter
{
	public:
		//constructors and destructors
		Hunter();
		~Hunter();
		//modify methods
		void setPos(int x, int y);
		void setPower(int pow) {power = pow;}
		void setFoodCount(int fc) {foodCount = fc;}
		void setID(int idNum) {hID = idNum;}
		void setAlive(bool a) {alive = a;}
		void setStrengthAttack(int str) {strengthAttack = str;}
		void setStrengthShare(int str) {strengthShare = str;}
		void setStrengthHoard(int str) {strengthHoard = str;}
		void setCostAttack(int cost) {costAttack = cost;}
		void setUtilityAttack(int util) {utilityAttack = util;}
		void setUtilityShare(int util) {utilityShare = util;}
		void setUtilityHoard(int util) {utilityHoard = util;}
		void setVision(int vis) {vision = vis;}
		void setMetabolism(int met) {metabolism = met;}
		void createInteract(int numHunters);
		void initInteract(int numHunters);
		void setInteracted(int i) {interactedWith[i] = true;}
		void move(int* worldMatrix, int WSX, int WSY, int EMPTY, int FOOD);
		int getScore(int* worldMatrix, int WSX, int EMPTY, int FOOD, int cellX, int cellY);
		void gather(int* worldMatrix, int WSX, int WSY, int EMPTY, 
					int FOOD,int FOOD_VALUE);
		void removeTies();
		//constant methods
		int getPower() {return power;}
		int getFoodCount() {return foodCount;}
		int getStrengthAttack() {return strengthAttack;}
		int getStrengthShare() {return strengthShare;}
		int getStrengthHoard() {return strengthHoard;}
		int getY() {return yPos;}
		int getX() {return xPos;}
		int getCostAttack() {return costAttack;}
		int getUtilityAttack() {return utilityAttack;}
		int getUtilityShare() {return utilityShare;}
		int getUtilityHoard() {return utilityHoard;}
		int getVision() {return vision;}
		int getMetabolism() {return metabolism;}
		bool isAlive() {return alive;}
		bool interacted(int i) {return interactedWith[i];}
		bool isAttacking();
		bool isSharing();
		bool isHoarding();
	private:
		int xPos,yPos;			//position in world
		int power;				//overall power of hunter, its ability
								//to defeat other hunters.
		int hID;				//hunters id number
		int foodCount;			//amount of food hunter has
		int costAttack;			//cost (in food) of an attack
		int vision;				//how far can the agent "see"?
		int metabolism;			//how much food does the hunter consume each round?
		bool alive;				//is the agent alive?
		bool* interactedWith;	//array to keep track of which other agents
								//this agent has interacted with this round.
		//Utilities
		int utilityAttack;		//utility gained from attacking
		int utilityShare;		//utility gained from sharing
		int utilityHoard;		//utility gained from hoarding

		//strengths
		int strengthAttack;		//Strength associated with attacking other agents
		int strengthShare;		//Strength associated with sharing with other agents
		int strengthHoard;		//Strength associated with hoarding food
};

//Non-member functions
void hunterInteraction(int* world,int WSY,int WSX,Hunter* hunters, int i, ofstream &log);
void hunterInteractionResolver(Hunter* hunters, int i, int target, ofstream &log);
void hunterEvolution(Hunter* hunters, int NUM_HUNTERS, int numAlive);


#endif