//file: Hunter.cpp
//implements Hunter class
//
//Programer: Andrew Schlei
//
//Ver 1.0

#include "Hunter.h"
#include <stdlib.h>
#include <time.h>

//===================CONSTRUCTORS & DESTRUCTORS=========================
Hunter::Hunter()
//pre: none
//post: none
{
}

Hunter::~Hunter()
//pre: none
//post: memory used by interactedWith array has been freed 
{
	delete [] interactedWith;
}
//===============================METHODS=================================
void Hunter::setPos(int x, int y)
//this method set the x and y position
//pre: x and y are coordinates in the 2-D world, with x increasing from
//	   left to right and y increasing from top to bottom
//post: xPos and yPos have changed to reflect agent's new position
{
	xPos = x;
	yPos = y;
}

void Hunter::createInteract(int numHunters)
//this method is used to create the interactedWith array
//pre: numHunters is the number of hunter agents in the simulation
//post: the agent's interactedWith array has been allocated
{
	interactedWith = new bool[numHunters];
}

void Hunter::initInteract(int numHunters)
//this method initializes the hunters interactedWith array
//pre: numHunters is the number of hunter agents in the simulation
//post: the agent's interactedWith array has been initialized to false
{
	for(int i=0; i<numHunters; i++)
	{
		interactedWith[i] = false;
	}
}

void Hunter::move(int* worldMatrix, int WSX, int WSY, int EMPTY, int FOOD)
//this method determines how the agents move
//pre: -worldMatrix is a pointer to a 1 dimensional array that represents
//	    the 2 dimensional world the hunter agents live in
//	   -WSX is the world size in the x dimension (in cells)
//	   -WSY is the world size in the y dimension (in cells)
//	   -EMPTY is the number denoting an empty cell in worldMatrix
//	   -FOOD is the number denoting food in worldMatrix
//post: worldMatrix has been modified to reflect the new positions of the
//		hunter agents.  The agent's xPos and yPos coordinates have also 
//		been altered to reflect the agent's new position.
{
	int random;		//holds random number
	enum direction_t{WEST,NORTH,EAST,SOUTH,STAY};
	direction_t moveDirection = STAY;
	bool squareOK = false;	//is the square being moved to open?
	int scoreW;		//score for west direction
	int scoreN;		//score for west direction
	int scoreE;		//score for west direction
	int scoreS;		//score for west direction

	//agent "looks" in each direction and records score for each direction
	for(int i=1;i<vision;i++)
	{
		if(xPos-i >=0) scoreW += getScore(worldMatrix,WSX,EMPTY,FOOD,xPos-i,yPos);
		if(yPos-i >=0) scoreN += getScore(worldMatrix,WSX,EMPTY,FOOD,xPos,yPos-1);
		if(xPos+i < WSX) scoreE += getScore(worldMatrix,WSX,EMPTY,FOOD,xPos+i,yPos);
		if(yPos+i < WSY) scoreS += getScore(worldMatrix,WSX,EMPTY,FOOD,xPos,yPos+i);
	}
	
	//agent figures out which direction to move
	if(scoreW > scoreN && scoreW > scoreE && scoreW > scoreS) 
	{
		moveDirection = WEST;
		squareOK = xPos != 0 && worldMatrix[yPos*WSX+xPos-1] == EMPTY;
	}
	else if(scoreN > scoreE && scoreN > scoreS && scoreN > scoreW) 
	{
		moveDirection = NORTH;
		squareOK = yPos != 0 && worldMatrix[(yPos-1)*WSX+xPos] == EMPTY;
	}
	else if(scoreE > scoreS && scoreE > scoreW && scoreE > scoreN) 
	{
		moveDirection = EAST;
		squareOK = xPos != WSX-1 && worldMatrix[yPos*WSX+xPos+1] == EMPTY;
	}
	else if(scoreS > scoreW && scoreS > scoreN && scoreS > scoreE) 
	{
		moveDirection = SOUTH;
		squareOK = yPos != WSY-1 && worldMatrix[(yPos+1)*WSX+xPos] == EMPTY;
	}

	if(!squareOK || moveDirection == STAY)
	{
		do
		{
			random = rand() % 5;
				switch(random)
				{
					case 0:
						moveDirection = WEST;
						squareOK = xPos != 0 && worldMatrix[yPos*WSX+xPos-1] == EMPTY;
						break;
					case 1:
						moveDirection = NORTH;
						squareOK = yPos != 0 && worldMatrix[(yPos-1)*WSX+xPos] == EMPTY;
						break;
					case 2:
						moveDirection = EAST;
						squareOK = xPos != WSX-1 && worldMatrix[yPos*WSX+xPos+1] == EMPTY;
						break;
					case 3:
						moveDirection = SOUTH;
						squareOK = yPos != WSY-1 && worldMatrix[(yPos+1)*WSX+xPos] == EMPTY;
						break;
					case 4:
						moveDirection = STAY;
						squareOK = true;
						break;
				}
		}while(!squareOK);
	}

	switch(moveDirection)
	{
		case WEST:
			worldMatrix[yPos*WSX+xPos] = EMPTY;
			worldMatrix[yPos*WSX+xPos-1] = hID;
			xPos = xPos - 1;
			break;
		case NORTH:
			worldMatrix[yPos*WSX+xPos] = EMPTY;
			worldMatrix[(yPos-1)*WSX+xPos] = hID;
			yPos = yPos - 1;
			break;
		case EAST:
			worldMatrix[yPos*WSX+xPos] = EMPTY;
			worldMatrix[yPos*WSX+xPos+1] = hID;
			xPos = xPos + 1;
			break;
		case SOUTH:
			worldMatrix[yPos*WSX+xPos] = EMPTY;
			worldMatrix[(yPos+1)*WSX+xPos] = hID;
			yPos = yPos + 1;
			break;
		case STAY:
			break;
	}
}

int Hunter::getScore(int* worldMatrix, int WSX, int EMPTY, int FOOD, int cellX, int cellY)
//this method is used to find the score for a given cell, which agents use to determine how to move
//pre: -worldMatrix is a pointer to a 1 dimensional array that represents
//	    the 2 dimensional world the hunter agents live in
//	   -WSX is the world size in the x dimension (in cells)
//	   -EMPTY is the number denoting an empty cell in worldMatrix
//	   -FOOD is the number denoting food in worldMatrix
//	   -cellX is the x coordinate of the cell to be examined
//	   -cellY is the y coordinate of the cell to be examined
//post: returns score for given cell
		
{
	const int foodScore = 5;			//score for cell with food
	const int agentScoreAttack = 4;		//score for a cell with another agent for an attacking agent
	const int agentScoreShare = 4;		//score for a cell with another agent for a sharing agent
	const int agentScoreHoard = 0;		//score for a cell with another agent for a hoarding agent

	int score=0;
	int cell;
	
	cell = worldMatrix[cellY*WSX+cellX];

	if(cell == EMPTY) score = 0;
	else if(cell == FOOD) score = foodScore;
	else
	{
		if(isAttacking())
		{
			score = agentScoreAttack;
		}
		else if(isSharing())
		{
			score = agentScoreShare;
		}
		else
		{
			score = agentScoreHoard;
		}
	}

	return score;
}

void Hunter::gather(int* worldMatrix,int WSX,int WSY,int EMPTY,
					int FOOD, int FOOD_VALUE)
//this method is used by the agents to gather food
//pre: -worldMatrix is a pointer to a 1 dimensional array that represents
//	    the 2 dimensional world the hunter agents live in
//	   -WSX is the world size in the x dimension (in cells)
//	   -WSY is the world size in the y dimension (in cells)
//	   -EMPTY is the number denoting an empty cell in worldMatrix
//	   -FOOD is the nubmer denoting a cell containing food in worldMatrix
//	   -FOOD_VALUE is the amount of food gained from collecting from a 
//	    food cell
//post: The agent's foodCount has been adjusted to reflect any food gathered.
//	    worldMatrix has been modified to remove any food cells that were
//		gathered (by setting them to EMPTY).
{
	if(xPos != 0 && (worldMatrix[yPos*WSX+xPos-1] == FOOD))
	{
		foodCount += FOOD_VALUE;
		worldMatrix[yPos*WSX+xPos-1] = EMPTY;
	}

	if(xPos != (WSX-1) && (worldMatrix[yPos*WSX+xPos+1] == FOOD))
	{
		foodCount += FOOD_VALUE;
		worldMatrix[yPos*WSX+xPos+1] = EMPTY;
	}

	if(yPos != 0 && (worldMatrix[(yPos-1)*WSX+xPos] == FOOD))
	{
		foodCount += FOOD_VALUE;
		worldMatrix[(yPos-1)*WSX+xPos] = EMPTY;
	}

	if(yPos != (WSY-1) && (worldMatrix[(yPos+1)*WSX+xPos] == FOOD))
	{
		foodCount += FOOD_VALUE;
		worldMatrix[(yPos+1)*WSX+xPos] = EMPTY;
	}
}

void Hunter::removeTies()
//This method is used to remove any ties for the highest strength.  If any
//strengths are tied for highest, it randomly selects one to increment by
//one.
//pre: none
//post: the agent's action strengths have been adjusted to remove any ties
{
	int random;		//used to store random numbers
	if(strengthAttack == strengthShare && strengthAttack > strengthHoard)
	{
		//Attack and Share are tied
		random = rand() % 2;
		if(random == 0) strengthAttack++; else strengthShare++;
	}
	else if(strengthAttack == strengthHoard && strengthAttack > strengthShare)
	{
		//Attack and Hoard are tied
		random = rand() % 2;
		if(random == 0) strengthAttack++; else strengthHoard++;
	}
	else if(strengthShare == strengthHoard && strengthShare > strengthAttack)
	{
		//Share and Hoard are tied
		random = rand() % 2;
		if(random == 0) strengthShare++; else strengthHoard++;
	}
	else if(strengthAttack == strengthShare && strengthShare == strengthHoard)
	{
		//Attack, Share, and Hoard are all tied
		random = rand() %3;
		if (random == 0) strengthAttack++;
		else if (random == 1) strengthShare++;
		else strengthHoard++;
	}
}
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
bool Hunter::isAttacking()
//This method is used to determine if the agent is currently in "attack mode"
//pre: none
//post: returns true if the the agent's strengthAttack is its highest strength
{
	return(strengthAttack > strengthShare && strengthAttack > strengthHoard);
}

bool Hunter::isSharing()
//This method is used to determine if the agent is currently in "share mode"
//pre: none
//post: returns true if the the agent's strengthShare is its highest strength
{
	return(strengthShare > strengthAttack && strengthShare > strengthHoard);
}

bool Hunter::isHoarding()
//This method is used to determine if the agent is currently in "hoard mode"
//pre: none
//post: returns true if the the agent's strengthHoard is its highest strength
{
	return(strengthHoard > strengthAttack && strengthHoard > strengthShare);
}

//=======================NON-MEMBER FUNCTIONS=================================
void hunterInteraction(int* world,int WSY,int WSX,Hunter* hunters, int i, ofstream &log)
//This function is used by hunter i to search for nearby hunters it can
//interact with.  When it finds a hunter it has not yet interacted with, 
//hunterInteractionResolver is called.
//pre: -worldMatrix is a pointer to a 1 dimensional array that represents
//	    the 2 dimensional world the hunter agents live in
//	   -WSX is the world size in the x dimension (in cells)
//	   -WSY is the world size in the y dimension (in cells)
//	   -hunters is a pointer to the array of hunter agents
//	   -i is the id number of the agent currently acting
//	   -log is an open output file stream
//post: agent i will have interacted with any agents nearby if it hasn't already.
{
	int x = hunters[i].getX();
	int y = hunters[i].getY();
	int targetID;

	if(x != 0)
	{
		targetID = world[y*WSX+x-1]; //left
		if(targetID >= 0 && !hunters[i].interacted(targetID))
		{
			hunterInteractionResolver(hunters,i,targetID,log);
		}
	}

	if(x != (WSX-1))
	{
		targetID = world[y*WSX+x+1]; //right
		if(targetID >= 0 && !hunters[i].interacted(targetID))
		{
			hunterInteractionResolver(hunters,i,targetID,log);
		}
	}

	if(y != 0)
	{
		targetID = world[(y-1)*WSX+x]; //up
		if(targetID >= 0 && !hunters[i].interacted(targetID))
		{
			hunterInteractionResolver(hunters,i,targetID,log);
		}
	}

	if(y != (WSY-1))
	{
		targetID = world[(y+1)*WSX+x]; //down
		if(targetID >= 0 && !hunters[i].interacted(targetID))
		{
			hunterInteractionResolver(hunters,i,targetID,log);
		}
	}
}

void hunterInteractionResolver(Hunter* hunters, int i, int targetID, ofstream &logOutput)
//This function is used to resolve interactions between hunters
//pre: -hunters is a pointer to the array of hunter agents
//	   -i is the id number of the agent currently acting
//	   -targetID is the id number of the agent being interacted with
//	   -logOutput is an open file output stream
//post: An action between two agents has been resolved, including adjustments to
//		action strengths, foodCounts, and the interactedWith arrays.
{
	int oldFoodCountI = hunters[i].getFoodCount();				//hunter i's food count 
	int oldFoodCountTarget = hunters[targetID].getFoodCount();	//hunter target's food count
	int foodStolen;					//amount of food stolen
	int foodPooled;					//amount of food pooled
	int random;						//holds random numbers

	//resolve action
	if (hunters[i].isAttacking())	//***RESOLVE ATTACK ACTION***
	{
		logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food, " << hunters[i].getPower() <<" power) [" 
				  << hunters[i].getStrengthAttack() << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard()
				  << "] attacks hunter " << targetID  << "(" << hunters[targetID].getFoodCount() << " food, "<< hunters[targetID].getPower() 
				  <<" power) [" << hunters[targetID].getStrengthAttack() << "/" << hunters[targetID].getStrengthShare() << "/"
				  << hunters[targetID].getStrengthHoard() << "]"<< endl;
		hunters[i].setFoodCount(hunters[i].getFoodCount()-hunters[i].getCostAttack());
		if (hunters[targetID].isAttacking())
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount()-hunters[targetID].getCostAttack());

		if (hunters[i].getPower() > hunters[targetID].getPower())
		//Hunter i is the strongest
		{
			foodStolen = (int)(hunters[targetID].getFoodCount() / 2);
			hunters[i].setFoodCount(hunters[i].getFoodCount() + foodStolen);
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() - foodStolen);
			logOutput << "hunter " << i << " wins and steals " << foodStolen << " food" << endl;
		}
		else if (hunters[targetID].getPower() > hunters[i].getPower())
		//Hunter target is the strongest
		{
			if (hunters[targetID].isAttacking())	//hunter target is attacking
			{
				foodStolen = (int)(hunters[i].getFoodCount() /2);
				hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
				hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
				logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
			}
			else
			{
				logOutput << "hunter " << targetID << " wins, but not attacking so steals no food" << endl;
			}
		}
		else
		//Hunters have equal strength, winner determined randomly
		{
			random = rand() % 2;
			if(random == 0)	//hunter i wins
			{
				foodStolen = (int)(hunters[targetID].getFoodCount() / 2);
				hunters[i].setFoodCount(hunters[i].getFoodCount() + foodStolen);
				hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() - foodStolen);
				logOutput << "hunter " << i << " wins and steals " << foodStolen << " food" << endl;
			}
			else	//hunter target wins
			{
				if (hunters[targetID].isAttacking())
				//hunter target is attacking, and can get some of hunter i's food
				{
					foodStolen = (int)(hunters[i].getFoodCount() /2);
					hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
					hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
					logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
				}
				else
				{	
					logOutput << "hunter " << targetID << " wins, but not attacking so steals no food" << endl;
				}
			}
		}
	}

	else if (hunters[i].isSharing())	//***RESOLVE SHARE ACTION***
	{
		if(hunters[targetID].isSharing())
		//hunters agree to share
		//each hunter puts half their food into a pool, and splits it evenly between the two of them
		{
			logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food)[ "<< hunters[i].getStrengthAttack() 
					  << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard()<<"] shares with hunter " 
					  << targetID  << "(" << hunters[targetID].getFoodCount() << " food)[" << hunters[targetID].getStrengthAttack() 
					  << "/" << hunters[targetID].getStrengthShare() << "/" << hunters[targetID].getStrengthHoard()<< "]"<< endl;
			foodPooled = (int)(hunters[i].getFoodCount()/2) + (int)(hunters[targetID].getFoodCount()/2);
			hunters[i].setFoodCount(hunters[i].getFoodCount() - (int)(hunters[i].getFoodCount()/2));
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() - (int)(hunters[targetID].getFoodCount()/2));

			hunters[i].setFoodCount(hunters[i].getFoodCount() + (int)(foodPooled/2));
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + (int)(foodPooled/2));
			if(foodPooled % 2 == 1) //amount of food pooled was odd
			{
				random = rand() % 2;
				if(random == 0)	//hunter i gets the extra food
				{
					hunters[i].setFoodCount(hunters[i].getFoodCount() + 1);
				}
				else //hunter target gets the extra food
				{
					hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + 1);
				}
			}
		}
		else if(hunters[targetID].isAttacking())
		//hunter target is attacking hunter i
		{
			logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food)[" << hunters[i].getStrengthAttack() 
					  << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard() << "] tries to share, but hunter " 
					  << targetID  << "(" << hunters[targetID].getFoodCount() << " food)[" << hunters[targetID].getStrengthAttack() 
					  << "/" << hunters[targetID].getStrengthShare() << "/" << hunters[targetID].getStrengthHoard()<<"] attacks" << endl;
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount()-hunters[targetID].getCostAttack());
			if(hunters[targetID].getPower() > hunters[i].getPower())
			//hunter target is the strongest
			{
				foodStolen = (int)(hunters[i].getFoodCount() /2);
				hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
				hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
				logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
			}
			else if  (hunters[i].getPower() > hunters[targetID].getPower())
			//hunter i wins
			{
				logOutput << "hunter " << i << " wins, but not attacking so steals no food" << endl;
			}
			else
			//hunters have equal strength, determine winner randomly
			{
				random = rand () % 2;
				if (random == 0)	//hunter target wins
				{
					foodStolen = (int)(hunters[i].getFoodCount() /2);
					hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
					hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
					logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
				}
				else //hunter i wins
				{
					logOutput << "hunter " << i << " wins, but not attacking so steals no food" << endl;
				}
			}
		}
		else
		//hunter target is hoarding
		{
			logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food)[" << hunters[i].getStrengthAttack() 
					  << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard() << "] tries to share, but hunter " 
					  << targetID  << "(" << hunters[targetID].getFoodCount() << " food) [" << hunters[targetID].getStrengthAttack() 
					  << "/" << hunters[targetID].getStrengthShare() << "/" << hunters[targetID].getStrengthHoard() << "] hoards" << endl;
		}
	}

	else if (hunters[i].isHoarding())	//***RESOLVE HOARD ACTION***
	//nothing happens unless target is attacking hunter i
	{
		if(hunters[targetID].isAttacking())
		//hunter target is attacking hunter i
		{
			logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food)[" << hunters[i].getStrengthAttack() 
					  << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard() << "] tries to hoard, but hunter " 
					  << targetID  << "(" << hunters[targetID].getFoodCount() << " food)[" << hunters[targetID].getStrengthAttack() 
					  << "/" << hunters[targetID].getStrengthShare() << "/" << hunters[targetID].getStrengthHoard() << "] attacks" << endl;
			hunters[targetID].setFoodCount(hunters[targetID].getFoodCount()-hunters[targetID].getCostAttack());
			if(hunters[targetID].getPower() > hunters[i].getPower())
			//hunter target is the strongest
			{
				foodStolen = (int)(hunters[i].getFoodCount() /2);
				hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
				hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
				logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
			}
			else if  (hunters[i].getPower() > hunters[targetID].getPower())
			//hunter i wins
			{
				logOutput << "hunter " << i << " wins, but not attacking so steals no food" << endl;
			}
			else
			//hunters have equal strength, determine winner randomly
			{
				random = rand () % 2;
				if (random == 0)	//hunter target wins
				{
					foodStolen = (int)(hunters[i].getFoodCount() /2);
					hunters[targetID].setFoodCount(hunters[targetID].getFoodCount() + foodStolen);
					hunters[i].setFoodCount(hunters[i].getFoodCount() - foodStolen);
					logOutput << "hunter " << targetID << " wins and steals " << foodStolen << " food" << endl;
				}
				else //hunter i wins
				{
					logOutput << "hunter " << i << " wins, but not attacking so steals no food" << endl;
				}
			}
		}
		else
		{
			logOutput << "hunter " << i << "(" << hunters[i].getFoodCount() << " food) hoards." << endl;
		}
	}


	//adjust action strengths
	//RULES
	//===============================================================================================
	//| Hunter 1 | Hunter 2 | Adjust A1 | Adjust S1 | Adjust H1 | Adjust A2 | Adjust S2 | Adjust H2 |
	//===============================================================================================
	//| Attacked | Attacked |  + Diff   |     0     |     0     |  + Diff   |     0     |     0     |
	//| Attacked | Shared   |  + Diff   |	  0     |     0     |  +/- 1    |     0     |     0     |
	//| Attacked | Hoareded |  + Diff   |     0     |     0     |  +/- 1    |     0     |  + Diff   |
	//| Shared	 | Shared   |     0     |  + Diff   |     0     |     0     |  + Diff   |     0     |
	//| Shared	 | Hoarded  |     0     |     0     |     0     |     0     |     0     |     0     |
	//| Hoarded  | Hoarded  |     0     |     0     |     0     |     0     |     0     |     0     |
	//===============================================================================================
	//KEY
	//Adjust A1/A2 = adjust hunter's attack strength
	//Adjust S1/S2 = adjust hunter's share strength
	//Adjust H1/H2 = adjust hunter's hoard strength
	//+ Diff = adjust strength by difference between the hunter's current food count and original food count 
	//		   (note that this can be negative) and then adding the utility offset.
	//+/- 1 = adjust strength by adding 1 if the hunter won the battle, or subtracting 1 if the hunter lost
	//0 = no change to strength
	if (hunters[i].isAttacking())	//hunter i attacked
	{
		hunters[i].setStrengthAttack(hunters[i].getStrengthAttack() + (hunters[i].getFoodCount()-oldFoodCountI) + hunters[i].getUtilityAttack());
		if(hunters[targetID].isAttacking())	//hunter targetID attaked
		{
			hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() + (hunters[targetID].getFoodCount()-oldFoodCountTarget) 
				+ hunters[targetID].getUtilityAttack());
		}
		else if (hunters[targetID].isSharing())	//hunter targetID shared
		{
			if((hunters[targetID].getFoodCount()-oldFoodCountTarget) < 0)	//hunter lost food in battle
			{
				hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() - 1);
			}
			else	//hunter won battle
			{
				hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() + 1);
			}
		}
		else if (hunters[targetID].isHoarding())	//hunter targetID hoarded
		{
			hunters[targetID].setStrengthHoard(hunters[targetID].getStrengthHoard() + (hunters[targetID].getFoodCount()-oldFoodCountTarget) 
				+ hunters[targetID].getUtilityHoard());
			if((hunters[targetID].getFoodCount()-oldFoodCountTarget) < 0)	//hunter lost food in battle
			{
				hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() - 1);
			}
			else	//hunter won battle
			{
				hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() + 1);
			}
		}
	}
	else if (hunters[i].isSharing())	//hunter i shared
	{
		if(hunters[targetID].isAttacking())	//hunter targetID attacked
		{
			hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() + (hunters[targetID].getFoodCount()-oldFoodCountTarget) 
				+ hunters[targetID].getUtilityAttack());
			if((hunters[i].getFoodCount()-oldFoodCountTarget) < 0)	//hunter lost food in battle
			{
				hunters[i].setStrengthAttack(hunters[i].getStrengthAttack() - 1);
			}
			else	//hunter won battle
			{
				hunters[i].setStrengthAttack(hunters[i].getStrengthAttack() + 1);
			}
		}
		else if(hunters[targetID].isSharing())	//hunter targetID shared
		{
			hunters[i].setStrengthShare(hunters[i].getStrengthShare() + (hunters[i].getFoodCount()-oldFoodCountI) + hunters[i].getUtilityShare());
			hunters[targetID].setStrengthShare(hunters[targetID].getStrengthShare() + (hunters[targetID].getFoodCount()-oldFoodCountTarget) 
				+ hunters[targetID].getUtilityShare());
		}
	}
	else //hunter i hoarded
	{
		if(hunters[targetID].isAttacking())	//hunter targetID attacked
		{
			hunters[i].setStrengthHoard(hunters[i].getStrengthHoard() + (hunters[i].getFoodCount()-oldFoodCountI) + hunters[i].getUtilityHoard());
			hunters[targetID].setStrengthAttack(hunters[targetID].getStrengthAttack() + (hunters[targetID].getFoodCount()-oldFoodCountTarget) 
				+ hunters[targetID].getUtilityAttack());
			if((hunters[i].getFoodCount()-oldFoodCountTarget) < 0)	//hunter lost food in battle
			{
				hunters[i].setStrengthAttack(hunters[i].getStrengthAttack() - 1);
			}
			else	//hunter won battle
			{
				hunters[i].setStrengthAttack(hunters[i].getStrengthAttack() + 1);
			}
		}
	}

	logOutput << "hunter " << i << " now has " << hunters[i].getFoodCount() << " food and ["<< hunters[i].getStrengthAttack() 
			  << "/" << hunters[i].getStrengthShare() << "/" << hunters[i].getStrengthHoard() << "], and hunter "
			  << targetID << " now has " << hunters[targetID].getFoodCount() << " food and [" << hunters[targetID].getStrengthAttack() 
			  << "/" << hunters[targetID].getStrengthShare() << "/" << hunters[targetID].getStrengthHoard() << "]." << endl;


	//mark agents as having interaced and remove strength ties
	hunters[i].setInteracted(targetID);
	hunters[targetID].setInteracted(i);
	hunters[i].removeTies();
	hunters[targetID].removeTies();
}

void hunterEvolution(Hunter* hunters, int NUM_HUNTERS, int numAlive)
//This function handles the evolution of the agents.  First, the id numbers
//of all surviving agents are gathered.  Then each dead agent is replaced
//with a new agent.  The new agent is created by first choosing two "parents"
//from among the living agents, and then averaging their power, stengthAttack,
//stengthShare, and strengthHoard member variables.  It is possible for a single
//agent to be both "parents"
//pre: -hunters is a pointer to the array of hunter agents
//	   -NUM_HUNTERS is the total number of hunters in the simulation
//	   -numAlive is the number of agents currently "alive"
//post: all dead hunter agents have been replaced with new agents, although
//		they are still dead.
{
	int i,j;								//counters
	int* livingHunters = new int[numAlive];	//array to hold ids of surviving Hunters
	int parent1, parent2;				//"parent" agents

	//gather ids of living agents
	j = 0;
	for(i=0;i<NUM_HUNTERS;i++)
	{
		if(hunters[i].isAlive())
		{
			livingHunters[j] = i;
			j++;
		}
	}

	//replace dead agents
	for(i=0;i<NUM_HUNTERS;i++)
	{
		if(!hunters[i].isAlive())
		{
			parent1 = rand() % numAlive;
			parent2 = rand() % numAlive;

			hunters[i].setPower((hunters[livingHunters[parent1]].getPower() +
								 hunters[livingHunters[parent2]].getPower()) / 2);
			hunters[i].setStrengthAttack((hunters[livingHunters[parent1]].getStrengthAttack() +
										  hunters[livingHunters[parent2]].getStrengthAttack()) / 2);
			hunters[i].setStrengthShare((hunters[livingHunters[parent1]].getStrengthShare() +
										 hunters[livingHunters[parent2]].getStrengthShare()) / 2);
			hunters[i].setStrengthHoard((hunters[livingHunters[parent1]].getStrengthHoard() +
										 hunters[livingHunters[parent2]].getStrengthHoard()) / 2);
		}
	}

	//cleanup
	delete [] livingHunters;
}