require 'test/unit'
require 'StartBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('STARTagent.rb')
      FileUtils.copy('STARTagent.rb', 'STARTagent.rb.save')
      begin 
        FileUtils.copy('STARTagent-api.yaml', 'STARTagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/STARTbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("StartBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("StartBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("StartBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("StartBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("StartBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("StartBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("StartBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("StartBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("STARTbot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("STARTbot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("STARTbot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("STARTbot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('STARTagent.rb') unless !File.exist?('STARTagent.rb.save')
    begin; File.delete('STARTagent-api.yaml'); rescue; end
    if File.exist?('STARTagent.rb.save')
      FileUtils.copy('STARTagent.rb.save', 'STARTagent.rb')
      begin; FileUtils.copy('STARTagent-api.yaml.save', 'STARTagent-api.yaml'); rescue; end
    end
    if File.exist?('STARTagent.rb') then File.delete('STARTagent.rb.save') end
    if File.exist?('STARTagent-api.yaml.save') then File.delete('STARTagent-api.yaml.save') end
  end

end
