################################################################################
# Natural language interface to the Akaranga Sutra.
# Verses are referenced by Book:Lecture:Lesson:Verse.
#
# > @ak
# A monk or a nun should not wipe or rub, &c. (all as in  7).  II:3:2:12 
# 
# > find / II:11:2:9/
# Nor to towers, pathways, doors, or town gates;  II:11:2:9 
#
# > next
# Nor where three or four roads meet, nor to courtyards or squares;   II:11:2:10
# 
# >
################################################################################

require 'yaml'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/akagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^show (.*)'s api/, :show_one_method_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0
     
     @ak = YAML.load( File.open("ak.yaml", "r") )

     @found_verses = Array.new
     @save_find = ''
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     puts "input==#{input}" unless @DEBUG == 0
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   # a is a synonym for b, or /a/ is a synonym for /b/
   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     g1 = $1; g2 = $2; #puts "[add_synonym] g1==#{g1}; g2=#{g2}" 
     # Remove enclosing "/", if present.
     if g1 =~ /^\// then g1.sub!(/^\//, '') end
     if g1 =~ /\/$/ then g1.sub!(/\/$/, '') end
     if g2 =~ /^\// then g2.sub!(/^\//, '') end
     if g2 =~ /\/$/ then g2.sub!(/\/$/, '') end
     #puts "[add_synonym] g1==#{g1}; g2=#{g2}"
     grp1 = %r{(?i-mx:^#{g1})}
     grp2 = %r{(?i-mx:^#{g2})}
     #puts "[add_synonym] grp1==#{grp1}; grp2=#{grp2}"
     method =''
     index = -1 # To add the new pattern before the old one.
     @patterns.each_with_index { |p,i| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          index = i
          break
       end
     }
     # If still no match, try again with slightly different regexps.
    if method == '' then grp1 = %r{(?i-mx:#{g1})}; grp2 = %r{(?i-mx:#{g2})}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{#{g1}}; grp2 = %r{#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{^#{g1}}; grp2 = %r{^#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
 
     return unless method.to_s != nil
     puts "add_synonym: index==#{index}" unless @DEBUG == 0
     @patterns.insert(index, [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       puts "@new_regexp==#{@new_regexp}" unless @DEBUG == 0

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def show_one_method_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1 # Supplied method name.
     apiarr= Array.new # To hold methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p| # Collect all the regexps for the supplied method.
       regexp = p[0].to_s; method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}." unless @DEBUG == 0
       if method == grp1 then apiarr.push [regexp, method] end 
     } 
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

# Begin agent-specific methods

   def random_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     size = @ak.size
     verse = @ak[rand(size)]
     verse =~ / ((?:I|II){1,1}\:\d+\:\d+\:\d+)/
     @save_find = $1
     return verse
   end # method_def

   def find_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     
     #if not @more then @score += 3; return "No more." end

     if grp1 != @save_find then @more = false end

     if @found_verses.size == 1 and @more
       @save_find = grp1
       response = @found_verses[0]
       @found_verses.clear
       @more = false
       #@save_find = ''
       @score = 4
       return response
     end
     if @found_verses.size > 1 and @more
       @save_find = grp1
       first_verse = @found_verses[0]
       @found_verses = @found_verses[1..-1]
       @score = 4
       return first_verse + " [#{@found_verses.size} more]"
     end 

     returnArr = Array.new
     puts "[find_verse] grp1==#{grp1}" unless @DEBUG == 0
     @ak.each { |verse| if verse =~ /#{grp1}/i then returnArr.push(verse) end }
     @score += 4
     if returnArr.size == 1 then @save_find = grp1; return returnArr[0] end
     if returnArr.size > 1
       @found_verses = returnArr[1..-1]
       @save_find = grp1
       return returnArr[0] + " [#{@found_verses.size} more]"
     end
     return "Nothing found."
   end # method_def

   # _ is to make testakagent.rb work! (because of @more?)
   def more(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @save_find == '' then return "No more." end
     if @found_verses.empty? then return "No more." end
     @more = true
     return getResponse("find #{@save_find}")[0]
   end # method_def

   # underscore to prevent possible conflict with keyword next?
   def _next(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     puts "0. @save_find==#{@save_find}" unless @DEBUG == 0
     if @save_find == '' then return "No previous verse."; @score += 3 end
     @save_find =~ /(I|II){1,1}\:(\d+)\:(\d+)\:(\d+)/
     puts "1. @save_find==#{@save_find}" unless @DEBUG == 0
     bookno = $1; lectureno = $2.to_i; lessonno = $3.to_i; verseno = $4.to_i
     @score += 4
     # increment verseno
     verseno += 1
     @save_find = "/ #{bookno}:#{lectureno}:#{lessonno}:#{verseno}/"
     puts "@save_find==#{@save_find}" unless @DEBUG == 0
     toReturn = getResponse("find #{@save_find}")[0]
     if toReturn =~ /(.*) \[\d* more\]/ then toReturn = $1 end
     return toReturn unless toReturn =~ /^Nothing found/
     # roll over to next lesson
     verseno = 1
     lessonno += 1
     @save_find = "/ #{bookno}:#{lectureno}:#{lessonno}:#{verseno}/"
     toReturn = getResponse("find #{@save_find}")[0]
     if toReturn =~ /(.*) \[\d* more\]/ then toReturn = $1 end
     return toReturn unless toReturn =~ /^Nothing found/
     # roll over to next lecture
     lessonno = 1
     lectureno += 1
     @save_find = "/ #{bookno}:#{lectureno}:#{lessonno}:#{verseno}/"
     toReturn = getResponse("find #{@save_find}")[0]
     if toReturn =~ /(.*) \[\d* more\]/ then toReturn = $1 end
     return toReturn unless toReturn =~ /^Nothing found/
     # roll over to next book
     lectureno = 1
     if bookno == 'I' then bookno = 'II' end
     @save_find = "/ #{bookno}:#{lectureno}:#{lessonno}:#{verseno}/"
     toReturn = getResponse("find #{@save_find}")[0]
     if toReturn =~ /(.*) \[\d* more\]/ then toReturn = $1 end
     return toReturn unless toReturn =~ /^Nothing found/
     #if bookno == 'II' then @save_find = ''; return "That was the last verse!" end
     if bookno == 'II' then return "That was the last verse!" end
     if toReturn =~ /(.*) \[\d* more\]/ then toReturn = $1 end
     return toReturn
   end # method_def

   def update_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     begin
       require 'net/http'
       Net::HTTP.start("subbot.org") { |http|
         resp = http.get("/akagent/ak.yaml")
         open("ak.yaml", "wb") { |file| file.write(resp.body) }
         resp = http.get("/akagent/scraper/ak.yaml")
         open("scraper/ak.yaml", "wb") { |file| file.write(resp.body) }
         resp = http.get("/akagent/scraper/akaranga.txt")
         open("scraper/akaranga.txt", "wb") { |file| file.write(resp.body) }
       }
     rescue => err; return "Eep! #{err}"; end
     @score += 4
     @ak = YAML.load( File.open("ak.yaml", "r") )
     return 'ak.yaml updated.'
   end # method_def

   def load_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     file = 'ak.yaml'
     file = process $1 unless $1 == nil
     begin
       @ak = YAML.load(File.open(file, "r"))
       return "Okay, I have loaded #{file}."
     rescue Exception => e
       return e
     end
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
