require 'test/unit'
require 'akagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('akagent.rb')
      FileUtils.copy('akagent.rb', 'akagent.rb.save')
      begin 
        FileUtils.copy('akagent-api.yaml', 'akagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_show_one_method_api
    actual, s = @agent.send("show show_api's api")
    assert_match(/show api/, actual)
  end

  def test_random
    actual, s = @agent.send("@ak")
    assert_match(/ (I|II)\:\d+:\d+:\d+$/, actual)
  end

  def test_find
    actual, s = @agent.send("@ak / I:1:1:1/")
    assert_match(/Gambusvamin/, actual)
  end

  def test_more
    actual, s = @agent.send("find elep.ant")
    assert_match(/as elephants/, actual)

    #actual, s = @agent.send("more") # doesn't work
    actual, s = @agent.getResponse("more")
    assert_match(/Now look at those who are born/, actual)
    
    #actual, s = @agent.send("more")
    actual, s = @agent.getResponse("more")
    assert_match(/\[7 more\]/, actual)
  end

  def test_next
    actual, s = @agent.send("@ak")
    assert_match(/(I|II){1,1}\:\d+\:\d+\:\d+/, actual)

    actual, s = @agent.getResponse("next")
    assert_match(/(I|II){1,1}\:\d+\:\d+\:\d+/, actual)

    actual, s = @agent.getResponse("find / I:2:2:1/")
    assert_match(/A wise man should remove any aversion/, actual)

    actual, s = @agent.getResponse("next")
    assert_match(/\(But on the contrary\)/, actual)
  end

  def teardown
    File.delete('akagent.rb') unless !File.exist?('akagent.rb.save')
    begin; File.delete('akagent-api.yaml'); rescue; end
    if File.exist?('akagent.rb.save')
      FileUtils.copy('akagent.rb.save', 'akagent.rb')
      begin; FileUtils.copy('akagent-api.yaml.save', 'akagent-api.yaml'); rescue; end
    end
    if File.exist?('akagent.rb') then File.delete('akagent.rb.save') end
    if File.exist?('akagent-api.yaml.save') then File.delete('akagent-api.yaml.save') end
  end

end
