require 'test/unit'
require 'AkBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('akagent.rb')
      FileUtils.copy('akagent.rb', 'akagent.rb.save')
      begin 
        FileUtils.copy('akagent-api.yaml', 'akagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/akbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("AkBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("AkBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("AkBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("AkBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("AkBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("AkBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("AkBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("AkBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("akbot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("akbot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("akbot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("akbot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('akagent.rb') unless !File.exist?('akagent.rb.save')
    begin; File.delete('akagent-api.yaml'); rescue; end
    if File.exist?('akagent.rb.save')
      FileUtils.copy('akagent.rb.save', 'akagent.rb')
      begin; FileUtils.copy('akagent-api.yaml.save', 'akagent-api.yaml'); rescue; end
    end
    if File.exist?('akagent.rb') then File.delete('akagent.rb.save') end
    if File.exist?('akagent-api.yaml.save') then File.delete('akagent-api.yaml.save') end
  end

end


