################################################################################
# Provides a natural langauge interface to the Analects of Confucius.
#
# > @an
# Book IV, Chapter XXIII: The Master said, 'The cautious seldom err.'
#
# > @an Book IV, Chapter XXIII 
# Book IV, Chapter XXIII: The Master said, 'The cautious seldom err.'
#
# > find "err"
# Book IV, Chapter XXIII: The Master said, 'The cautious seldom err.'  [10 more]
#
# > more
# Book VI, Chapter XXVIII: 1. Tsze-kung said, 'Suppose the case of a man [...]
#
################################################################################

require 'yaml'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/analectsagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^show (.*)'s api/, :show_one_method_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0

     @analects = Array.new # Holds analects.txt
     get_analects
     @anarr = Array.new # Holds analects in an array. (See scraper.rb)
     get_anarr

     @found_verses = Array.new
     @save_find = ''
   end
   
   def get_analects
     Dir.chdir('scraper')
     begin; File.open("analects.txt", "r") { |f|
       @analects = f.readlines
     }; rescue Exception => e; puts e; end
     Dir.chdir('..')
   end # method_def

   def get_anarr
     Dir.chdir('scraper')
     begin; @anarr = YAML.load( File.open("analects.yaml", "r") )
     rescue Exception => e; puts e; end
     Dir.chdir('..')
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     puts "input==#{input}" unless @DEBUG == 0
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   # a is a synonym for b, or /a/ is a synonym for /b/
   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     g1 = $1; g2 = $2; #puts "[add_synonym] g1==#{g1}; g2=#{g2}" 
     # Remove enclosing "/", if present.
     if g1 =~ /^\// then g1.sub!(/^\//, '') end
     if g1 =~ /\/$/ then g1.sub!(/\/$/, '') end
     if g2 =~ /^\// then g2.sub!(/^\//, '') end
     if g2 =~ /\/$/ then g2.sub!(/\/$/, '') end
     #puts "[add_synonym] g1==#{g1}; g2=#{g2}"
     grp1 = %r{(?i-mx:^#{g1})}
     grp2 = %r{(?i-mx:^#{g2})}
     #puts "[add_synonym] grp1==#{grp1}; grp2=#{grp2}"
     method =''
     index = -1 # To add the new pattern before the old one.
     @patterns.each_with_index { |p,i| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          index = i
          break
       end
     }
     # If still no match, try again with slightly different regexps.
    if method == '' then grp1 = %r{(?i-mx:#{g1})}; grp2 = %r{(?i-mx:#{g2})}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{#{g1}}; grp2 = %r{#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{^#{g1}}; grp2 = %r{^#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
 
     return unless method.to_s != nil
     puts "add_synonym: index==#{index}" unless @DEBUG == 0
     @patterns.insert(index, [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       puts "@new_regexp==#{@new_regexp}" unless @DEBUG == 0

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def show_one_method_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1 # Supplied method name.
     apiarr= Array.new # To hold methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p| # Collect all the regexps for the supplied method.
       regexp = p[0].to_s; method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}." unless @DEBUG == 0
       if method == grp1 then apiarr.push [regexp, method] end 
     } 
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

# Begin agent-specific methods

   def find_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     re = %r{#{grp1}}i
     puts "re==#{re}" unless @DEBUG == 0
     verse = 'Default verse'

     # Deal with "more"
     if grp1 != @save_find then @more = false end
     if @found_verses.size == 1 and @more
       @save_find = grp1
       response = @found_verses[0]
       @found_verses.clear
       @more = false
       @score = 4
       return response
     end
     if @found_verses.size > 1 and @more
       @save_vind = grp1
       first_verse = @found_verses[0]
       @found_verses = @found_verses[1..-1]
       @score = 4
       return first_verse + " [#{@found_verses.size} more]"
     end 

     # Find verse
     verses = Array.new
     @anarr.each { |elem|
       if elem =~ re then verses.push(elem) end
     }
    
     # Return response (adding "more" if there are multiple matches) 
     @score += 4
     if verses.size == 1 then @save_find == grp1; return verses[0] end
     if verses.size > 1
       @found_verses = verses[1..-1]
       @save_find = grp1
       return verses[0] + " [#{@found_verses.size} more]"
     end
     return "Nothing found."
     #return verses.join("\n\n")
   end # method_def

   def more(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @save_find == '' then return "No more." end
     if @found_verses.empty? then return "No more." end
     @more = true
     return getResponse("@an #{@save_find}")[0]
   end # method_def

   # Tries to use the raw text file to do searches. Doesn't always work!
   # For example if the search term goes across lines...
   def find_verse_in_raw_text(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     re = %r{#{grp1}}i
     puts "re==#{re}" unless @DEBUG == 0
     verse = 'Default verse'
     verses = Array.new
     book = 'Default book'
     @analects.each_with_index { |elem, i|
       #if elem =~ re then puts "elem==#{elem}; re==#{re}" end
       if elem =~ re 
         #puts "elem==#{elem}"
         verse = get_verse(i)
         #puts "verse==#{verse}"
         book = get_book(verse)
         #puts "book==#{book}"
         verses.push(book + ' ' + verse.strip)
       end
     } 
     return verses.uniq.join("\n\n")
   end # method_def

   # Utility method to find the book that contains a verse.
   # WARNING: THIS IS A HACK!
   def get_book(verse = '')
     book = ''
     #puts "verse==#{verse}"
     s = @analects.join(' ')
     s.gsub!(/\n/, ' ')
     to_match = verse[0..50]
     to_match.strip!
     #puts "to_match==#{to_match}"
     s =~ /#{to_match}/
     find_book = $`.to_s
     #puts "find_book==#{find_book}"
     if find_book != nil then find_book =~ /(?:.*)(BOOK [IVXL]*)\.([^(?:BOOK)].*)$/; book = $1.to_s end
     
     return book
   end # method_def

   # Utility method to find the complete verse from a fragment.
   def get_verse(i = 0)
      #puts "get_verse: i==#{i}"
      orig_i = i
      verse = @analects[i] 
      #puts "verse=#{verse}"
      orig_verse = verse
      begin
      while verse !~ /\bCHAP(?:TER |\. )?/ do 
        if i < 0 then break end
        s = @analects[i-1]
        verse = s.strip + ' ' + verse.strip + ' '
        i -= 1 
      end

      #puts "get_verse: verse==#{verse}"
      pre_verse = verse
      pre_i = i
      i = orig_i
      verse = orig_verse

      #puts "****"; puts "orig_verse==#{orig_verse}"; puts "****"
      #puts "****"; puts "pre_verse== #{pre_verse}"; puts "****"
      #puts "verse == #{verse}"; puts "**********"
      pre_verse.gsub!(/\n/, ' ');
      if pre_verse =~ /(.*)#{orig_verse.strip}/ then pre_verse = $1.to_s end

      # To catch the case where the search word appears on the CHAPTER line.
      #puts "verse==#{verse}; pre_i==#{pre_i}; orig_i==#{orig_i}; i==#{i}"
      if verse =~ /\bCHAP(?:TER |\. )?/ and pre_i == orig_i
        verse = @analects[i+1]
      end

      while verse !~ /\bCHAP(?:TER |\. )?/ and verse !~ /\bBOOK / and verse !~ /^End of/ do 
        if i >= (@analects.size-1) then puts i; puts @analects.size-1; break end
        s = @analects[i+1]
        verse = verse + s.strip + ' '
        i += 1
      end
      if verse =~ /(.*)\bCHAP(?:TER |\. )?/ or verse =~ /(.*)\bBOOK / or verse !~ /^End of/
        verse = $1.to_s.strip
      end
      rescue Exception => e; puts e; end
      to_return = pre_verse.strip + ' ' + verse.strip
      if to_return.include? orig_verse.strip then return to_return end
      to_return = pre_verse.strip + ' ' + orig_verse.strip + ' ' + verse.strip
      return to_return
      #return to_return.strip!
   end # method_def

   def random_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     size = @anarr.size
     return @anarr[rand(size)]
   end # method_def

   def shortest_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     min = 1000; min_i = 0
     @anarr.each_with_index { |elem, i|
       if elem.size < min then min = elem.size; min_i = i end
     }
     return @anarr[min_i]
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
