require 'test/unit'
require 'AnalectsBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('analectsagent.rb')
      FileUtils.copy('analectsagent.rb', 'analectsagent.rb.save')
      begin 
        FileUtils.copy('analectsagent-api.yaml', 'analectsagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/analectsbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("AnalectsBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("AnalectsBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("AnalectsBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("AnalectsBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("AnalectsBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("AnalectsBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("AnalectsBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("AnalectsBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("analectsbot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("analectsbot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("analectsbot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("analectsbot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('analectsagent.rb') unless !File.exist?('analectsagent.rb.save')
    begin; File.delete('analectsagent-api.yaml'); rescue; end
    if File.exist?('analectsagent.rb.save')
      FileUtils.copy('analectsagent.rb.save', 'analectsagent.rb')
      begin; FileUtils.copy('analectsagent-api.yaml.save', 'analectsagent-api.yaml'); rescue; end
    end
    if File.exist?('analectsagent.rb') then File.delete('analectsagent.rb.save') end
    if File.exist?('analectsagent-api.yaml.save') then File.delete('analectsagent-api.yaml.save') end
  end

end
