################################################################################
# Implements backward chaining.
#
# > if A recommends silver and B recommends silver, then C buys silver.
# Okay ...
#
# > add A recommends silver
# Okay, A recommends silver
#
# > add B recommends silver
# Okay, B recommends silver
#
# > is 'C buys silver' true?
# Yes, C buys silver is true because A recommends silver and B recommends silver
#
# >
#
################################################################################

require 'yaml'
require 'drb'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/bcagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''

     @Chain = Hash.new
     @Propositions = Array.new # Built-in database of assertions
     @gagent_started = false
     @gagent = startAgent
     @score = 2 # default score
   end

   # attempt to start the agent, if it isn't started. NOT WORKING.
   def startAgent
     DRb.start_service()
     gagent = DRbObject.new(nil, 'druby://localhost:9011')
     paths = ['../gagent/gagentbotd.rb', 'c:\controlbot\controller\gagent\gagentbotd.rb']
     begin
       test = gagent.getResponse('gagent: help')
       if test =~ /aristotelian/i 
         puts "Gagent is available." 
         @gagent_started = true
         return gagent 
       end
     rescue NoMethodError, DRb::DRbConnError
       puts "Gagent not started, trying #{paths[0]}"
     end
     begin; gagentd = IO.popen("-" << paths[0], "r+"); rescue; end
     gagent = DRbObject.new(nil, 'druby://localhost:9011')
     begin
       test = gagent.getResponse('gagent: help')
       if test =~ /aristotelian/i 
         puts "Gagent is available."
         @gagent_started = true
         return gagent 
       end
     rescue NoMethodError, DRb::DRbConnError
       puts "Gagent not started, trying #{paths[1]}"
     end
     begin; gagentd= IO.popen("-" << paths[1], "r+"); rescue; end
     gagent = DRbObject.new(nil, 'druby://localhost:9011')
     begin
       test = gagent.getResponse('gagent: help')
       if test =~ /aristotelian/i 
         puts "Gagent is available." 
         @gagent_started = true
       end
     rescue NoMethodError, DRb::DRbConnError
       puts "Gagent not working."
     end
     return gagent
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s
     grp2 = $2.to_s
     grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def} added."
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     #puts "input==#{input}"
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     @score = 2
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        file = File.basename(@filename)
        return "#{file} loaded."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     #grp1 = %r{^#{$1}}
     grp1 = %r{(?i-mx:^#{$1})}
     #grp2 = %r{^#{$2}}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s.strip) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def if_p_then_q(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1; grp2 = process $2; grp2 = grp2.to_s
     ants = grp1.split(' and ')
     #puts "ants==#{ants.inspect}"
     if @Chain[grp2] == nil
       @Chain[grp2] = [ants]
     else
       arr = @Chain[grp2]
       #puts "arr1 == #{arr.inspect}"
       arr.push(ants)
       arr.uniq!
       #puts "arr2 == #{arr.inspect}"
       @Chain[grp2] = arr
     end

     return "Okay I have added \"if #{ants.join(' and ')}, then #{grp2}\" to the chain."
   end # method_def

   def is_q(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1; grp2 = process $2
     self.send("is \"#{grp1} is #{grp2} true?")
   end

   # This method does not correctly handle conjunctions/disjunctions.
   def is_q_true(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     prop = false
     grp1 = process $1
     to_return = ''
     ants = @Chain.fetch(grp1)
     values = ''
     #puts "ants==#{ants.inspect}"
     ants.each { |val|
       if val.class == Array and val.size > 1
         if values.size > 0 
           values << ' or (' << val.join(' and ') << ')' 
         else 
           values << val.join(' and ')
         end
         val.each { |statement|
           puts "1. is #{statement} true?"

           # TODO: ask @gagent if each ant is true
           if @gagent_started 

             # turn "X is Y" into 'is "X" "Y"?'
             question = ""
             if statement =~ /(.*) (?:is|are)+ (.*)/
               _grp1 = process $1; _grp2 = process $2
               question = "is \"#{_grp1.downcase}\" \"#{_grp2.downcase}\"?"
	       puts "1. question==#{question}"
	     end

	     if question.size > 0
               response, s = @gagent.send(question.to_s)
	     else
               response, s = @gagent.send("is #{statement.downcase} true?")
	     end

             #puts "gagent response: #{response}"
             if response =~ /Yes/
               prop = true
             else
               prop = false
             end
           end
           
           # ask built-in database
           if @gagent_started == false or prop == false 
             if @Propositions.include?(statement.downcase)
               #puts "Yes, #{statement} is true."
               prop = true
             else
               prop = false
             end 
           end
         }
       else
         if values.size > 0
           values << ' or ' << val.to_s 
         else
           values << val.to_s
         end
         #puts "2. is #{val} true?"
 
         # TODO: ask @gagent if each ant is true
         if @gagent_started 

           # turn "X is Y" into 'is "X" "Y"?'
           question = ""
           if statement =~ /(.*) (?:is|are)+ (.*)/
             _grp1 = process $1; _grp2 = process $2
             question = "is \"#{_grp1.downcase}\" \"#{_grp2.downcase}\"?"
             puts "2. question==#{question}"
	   end

	   if question.size > 0 
             response, s = @gagent.send(question.to_s)
	   else
             response, s = @gagent.send("is #{val.to_s.downcase} true?")
	   end
   
           #puts "gagent response: #{response}"
	   if response =~ /Yes/
             prop = true
           else
             prop = false
           end
         end

         # ask built-in database 
         if @gagent_started == false or prop == false
           #puts "val==#{val}; @Propositions==#{@Propositions.inspect}"
           if @Propositions.include?(val.to_s.downcase)
             #puts "Yes, #{val} is true."
             prop = true
           else
             prop = false 
           end
         end
       end
     }
     if values == nil or values == ''
       to_return = "I don't know whether #{grp1} is true."
     end
     if prop then return "Yes, #{grp1} is true because #{values}" end
     return "#{grp1} is true if #{values} is true."
   end # method_def

   def delete_if_p_then_q(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1; grp2 = process $2
     #puts "grp1==#{grp1}; grp2==#{grp2}"
     if @Chain.delete(grp2) then return "Okay I have deleted #{grp2}." end
     return "I can't find the rule: if #{grp1} then #{grp2}" 
   end # method_def

   def show_if_p_then_q_rules(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     to_r = ''
     @Chain.each_pair { |key, value|
       values = ''
       #puts "key==#{key}, value==#{value.inspect}"
       value.each { |val|
         if val.class == Array and val.size > 1
           if values.size > 0 
             values << ' or (' << val.join(' and ') << ')' 
           else 
             values << val.join(' and ')
           end
         else
           if values.size > 0
             values << ' or ' << val.to_s 
           else
             values << val.to_s
           end
         end
       }
       to_r << "if #{values}, then #{key}\n"
     }
     to_r.chomp!
     return to_r
   end # method_def

   def delete_if_p_then_q_rules(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     count = @Chain.size
     s = 's'; if count == 1 then s = '' end
     @Chain.clear
     @score = 4
     return "Okay, I have deleted the #{count} item#{s} in the chain."
   end # method_def

   # add facts to the built-in database.
   def add_to_propositions(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     grp1.downcase!
     @Propositions.push(grp1)
     return "Okay, #{$1}"
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
