require 'test/unit'
require 'becauseagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('becauseagent.rb')
      FileUtils.copy('becauseagent.rb', 'becauseagent.rb.save')
      begin 
        FileUtils.copy('becauseagent-api.yaml', 'becauseagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_show_one_method_api
    actual, s = @agent.send("show show_api's api")
    assert_match(/show api/, actual)
  end

  def test_we_verb_x_because_y
    actual, s = @agent.send("we have daylight savings time because we want to save energy.")
    actual, s = @agent.send("why do we have daylight savings time?")
    assert_match(/we want to save energy/i, actual)

    actual, s = @agent.send("we want to save energy because it is expensive")
    actual, s = @agent.send("why do we want to save energy?")
    assert_match(/it is expensive/i, actual)
  end

  def test_x_verbs_y_because_z
    actual, s = @agent.send("the dog barks because he sees the mailman")
    actual, s = @agent.send("why does the dog bark?")
    assert_match(/he sees the mailman/i, actual)
  end

  def test_why_does_x_verb_y
    actual, s = @agent.send("Daylight Savings Time starts at 2am because that is the least disruptive time.")
    actual, s = @agent.send("why does daylight savings time start at 2am?")
    assert_match(/that is the least disruptive time/i, actual)
  end

  def teardown
    File.delete('becauseagent.rb') unless !File.exist?('becauseagent.rb.save')
    begin; File.delete('becauseagent-api.yaml'); rescue; end
    if File.exist?('becauseagent.rb.save')
      FileUtils.copy('becauseagent.rb.save', 'becauseagent.rb')
      begin; FileUtils.copy('becauseagent-api.yaml.save', 'becauseagent-api.yaml'); rescue; end
    end
    if File.exist?('becauseagent.rb') then File.delete('becauseagent.rb.save') end
    if File.exist?('becauseagent-api.yaml.save') then File.delete('becauseagent-api.yaml.save') end
  end

end
