require 'test/unit'
require 'BecauseBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('becauseagent.rb')
      FileUtils.copy('becauseagent.rb', 'becauseagent.rb.save')
      begin 
        FileUtils.copy('becauseagent-api.yaml', 'becauseagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/becausebot/, actual)
  end

  def test_volume
    actual, s = @agent.send("BecauseBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("BecauseBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("BecauseBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("BecauseBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("BecauseBot: delete rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("BecauseBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("BecauseBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("BecauseBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("becausebot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("becausebot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("becausebot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("becausebot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def test_why_is
    actual, s = @agent.send("The grass is wet because it's raining.")
    #actual, s = @agent.send("becausebot: loud")
    actual, s = @agent.send("why is the grass wet?")
    puts "actual==#{actual} s==#{s}"
    assert_match(/it's raining/i, actual)
  end

  def teardown
    File.delete('becauseagent.rb') unless !File.exist?('becauseagent.rb.save')
    begin; File.delete('becauseagent-api.yaml'); rescue; end
    if File.exist?('becauseagent.rb.save')
      FileUtils.copy('becauseagent.rb.save', 'becauseagent.rb')
      begin; FileUtils.copy('becauseagent-api.yaml.save', 'becauseagent-api.yaml'); rescue; end
    end
    if File.exist?('becauseagent.rb') then File.delete('becauseagent.rb.save') end
    if File.exist?('becauseagent-api.yaml.save') then File.delete('becauseagent-api.yaml.save') end
  end

end
