################################################################################
# Interactive chord dictionary.
#
# > A
# A - C# - E
#
# > F# minor
# F# - A - C#
#
# >
################################################################################

require 'yaml'
require 'drb'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/chordagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^show (.*)'s api/, :show_one_method_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0
     
     initialize_keys_hash
    
     DRb.start_service()
     @logic = DRbObject.new(nil, 'druby://localhost:9012') 
   end

   # http://en.wikipedia.org/wiki/Chord_notation  
   # Uses different note conventions (i.e., Gb/F#) for different scales.
   def initialize_keys_hash
     @keys = Hash.new

     # For F and C scales
     notes =['F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B', 'C', 'Db', 'D', 'Eb', 'E']
     notes.each_with_index { |key, i|
       if key !~ /[F|C]/ then next end
       scale = Array.new
       for i2 in (1..11)
         n =  notes[(i + i2) % 12]
         scale[i2-1] =  n
       end
       @keys[key] = scale #unless @keys[key] != nil
     }


     # For A, B, D, E, G scales
     notes =['A', 'Bb', 'B', 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#'] 
     notes.each_with_index { |key, i|
       if key =~ /#/ then next end
       if key =~ /B/ then next end
       scale = Array.new
       for i2 in (1..11)
         n =  notes[(i + i2) % 12]
         scale[i2-1] =  n
       end
       @keys[key] = scale unless @keys[key] != nil
     }

     # For Gb, Ab, Bb, Cb, Db, Eb scales
     notes = ['Gb', 'G', 'Ab', 'A', 'Bb', 'Cb', 'C', 'Db', 'D', 'Eb', 'E', 'F']
     #notes = ['Gb', 'Abb', 'Ab', 'Bbb', 'Bb', 'Cb', 'Dbb', 'Db', 'Ebb', 'Eb', 'Fb', 'F']
     notes.each_with_index{ |key, i|
       scale = Array.new
       for i2 in (1..11)
         n = notes[(i + i2) % 12]
         scale[i2-1] = n
       end
       @keys[key] = scale unless @keys[key] != nil
     }


     # For F#, G#, A#, B#, D#, E# scales
     notes = ['F#', 'G', 'G#', 'A', 'A#', 'B', 'B#', 'C#', 'D', 'D#', 'E', 'E#']
     notes.each_with_index{ |key, i|
       scale = Array.new
       for i2 in (1..11)
         n = notes[(i + i2) % 12]
         scale[i2-1] = n
       end
       @keys[key] = scale unless @keys[key] != nil
     }

   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     puts "input==#{input}" unless @DEBUG == 0
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       puts "@new_regexp==#{@new_regexp}" unless @DEBUG == 0

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def show_one_method_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1 # Supplied method name.
     apiarr= Array.new # To hold methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p| # Collect all the regexps for the supplied method.
       regexp = p[0].to_s; method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}." unless @DEBUG == 0
       if method == grp1 then apiarr.push [regexp, method] end 
     } 
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def major_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[3]; fifth = arr[6]
     @score = 4
     return "#{key} - #{third} - #{fifth}"
   end # method_def
  
   def major_scale(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     @score = 4
     return "#{key} - #{arr[1]} - #{arr[3]} - #{arr[4]} - #{arr[6]} - #{arr[8]} - #{arr[10]} - #{key}"
   end # method_def

   def minor_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[2]; fifth= arr[6]
     @score = 4
     return "#{key} - #{third} - #{fifth}"
   end # method_def


   def minor_scale(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     @score = 4
     return "#{key} - #{arr[1]} - #{arr[2]} - #{arr[4]} - #{arr[6]} - #{arr[7]} - #{arr[9]} - #{key}"
   end # method_def

   def dominant_seventh_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[3]; fifth = arr[6]; seventh = arr[9]
     @score = 4
     return "#{key} - #{third} - #{fifth} - #{seventh}"
   end # method_def

   def major_seventh_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[3]; fifth = arr[6]; seventh = arr[10]
     @score = 4
     return "#{key} - #{third} - #{fifth} - #{seventh}"
   end # method_def

   def power_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     fifth = arr[6]
     @score = 4
     return "#{key} - #{fifth} - #{key}"
   end # method_def

   def diminished_seventh_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[2]; fifth = arr[5]; seventh = arr[8]
     @score = 4
     return "#{key} - #{third} - #{fifth} - #{seventh}"
   end # method_def

   def augmented_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1
     arr = @keys[key]
     third = arr[3]; fifth = arr[6]
     @score = 4
     return "#{key} - #{third} - #{fifth}#"
   end # method_def

   def unknown_chord(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     key = process $1; chord_type = $2
     scale_arr = @keys[key]
     ans = "1 - 4 - 7"
     num_arr = ['1', '4', '7']
     toReturn = ''
     begin
       ans = @logic.send("what is #{chord_type}")[0]
       if ans =~ /don't/ then return ans end
       ans_arr = ans.split(',')
       ans_arr.each { |a|
         if a =~ / - /
           num_arr = a.split(' - ')
         end
       }
       num_arr.each_with_index { |s,i|
         if i == 0 then next end
         note = scale_arr[s.to_i - 1]
         #if i == 0 then toReturn.concat(note); next end
         toReturn.concat(" - #{note}")
       } 
     rescue => e; return "Eep! #{e}"; end
     @score = 4; return "#{key}#{toReturn}"
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
