require 'test/unit'
require 'chordagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('chordagent.rb')
      FileUtils.copy('chordagent.rb', 'chordagent.rb.save')
      begin 
        FileUtils.copy('chordagent-api.yaml', 'chordagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_show_one_method_api
    actual, s = @agent.send("show show_api's api")
    assert_match(/show api/, actual)
  end

  def test_major_chord
    actual, s = @agent.send("C")
    assert_match(/C - E - G/, actual)

    actual, s = @agent.send("D")
    assert_match(/D.*F#.*A/, actual)

    actual, s = @agent.send("F#")
    assert_match(/F#.*A#.*C#/, actual)

    actual, s = @agent.send("Bb")
    assert_match(/Bb.*D.*F/, actual)

    actual, s = @agent.send("Eb")
    assert_match(/Eb.*G.*Bb/, actual)
  end

  def test_minor_chord
    actual, s = @agent.send("Fm")
    assert_match(/F.*Ab.*C/, actual)

    actual, s = @agent.send("Dm")
    assert_match(/D.*F.*A/, actual)

    actual, s = @agent.send("Am")
    assert_match(/A.*C.*E/, actual)

    actual, s = @agent.send("F# minor")
    assert_match(/F#.*A.*C#/, actual)

    actual, s = @agent.send("Gb minor")
    assert_match(/Gb.*A.*Db/, actual)

    actual, s = @agent.send("A#m")
    assert_match(/A#.*C#.*E#/, actual)

    actual, s = @agent.send("Bb minor")
    assert_match(/Bb.*Db.*F/, actual)
  end

  def test_chords
    actual, s = @agent.send("Ebdim7")
    assert_match(/Eb.*Gb.*A.*C/, actual)

    actual, s = @agent.send("C5")
    assert_match(/C.*G.*C/, actual)

    actual, s = @agent.send("C7")
    assert_match(/C.*E.*G.*Bb/, actual)

    actual, s = @agent.send("F+")
    assert_match(/F.*A.*C#/, actual)
  end

  def teardown
    File.delete('chordagent.rb') unless !File.exist?('chordagent.rb.save')
    begin; File.delete('chordagent-api.yaml'); rescue; end
    if File.exist?('chordagent.rb.save')
      FileUtils.copy('chordagent.rb.save', 'chordagent.rb')
      begin; FileUtils.copy('chordagent-api.yaml.save', 'chordagent-api.yaml'); rescue; end
    end
    if File.exist?('chordagent.rb') then File.delete('chordagent.rb.save') end
    if File.exist?('chordagent-api.yaml.save') then File.delete('chordagent-api.yaml.save') end
  end

end
