require 'test/unit'
require 'ChordBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('chordagent.rb')
      FileUtils.copy('chordagent.rb', 'chordagent.rb.save')
      begin 
        FileUtils.copy('chordagent-api.yaml', 'chordagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/chordbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("ChordBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("ChordBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("chordbot: delete all rules")
    assert_match(/ok/i, actual)

    actual, s = @agent.send("ChordBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("ChordBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("ChordBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("ChordBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("ChordBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("ChordBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("chordbot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("chordbot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("chordbot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("chordbot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('chordagent.rb') unless !File.exist?('chordagent.rb.save')
    begin; File.delete('chordagent-api.yaml'); rescue; end
    if File.exist?('chordagent.rb.save')
      FileUtils.copy('chordagent.rb.save', 'chordagent.rb')
      begin; FileUtils.copy('chordagent-api.yaml.save', 'chordagent-api.yaml'); rescue; end
    end
    if File.exist?('chordagent.rb') then File.delete('chordagent.rb.save') end
    if File.exist?('chordagent-api.yaml.save') then File.delete('chordagent-api.yaml.save') end
  end

end
