# Wrapper for Controller, adding user commands to modify the score, etc. 
# This class can function both as a command-line program and 
# as an IRC bot when it is loaded by the RedTurnip IRC bot framework.

begin 
   require 'controller' # for when we start from the command line
rescue LoadError
   require 'mod/bot/controller' # for when we start from RedTurnip 
end

require 'readline'

class ControlBot 

   def initialize(creator = Object.new)
      #print "Hello\n"
      @creator = creator
      @agent = Controller.new(self)
      @THRESHHOLD = 0
      @MAX_RESPONSE = 10000
      @MY_NAME = "(?i)#{$irc_nickname}"
      #if @MY_NAME == "(?i)" then @MY_NAME = "(?i)controlbot" end
      if @MY_NAME == "(?i)" then @MY_NAME = "(?i)trane" end
      @agent.MY_NAME = @MY_NAME
      @DEBUG = 0
      @agent.DEBUG = @DEBUG
      @DELAY = 2.0 
      @agent.DELAY = @DELAY
      @SLEEP = 1.0 
      @agent.SLEEP = @SLEEP

      @rules = []
      puts "Reading rules.txt..."
      read_file("rules.txt") # read and execute commands in "rules.txt"
      # The following commented-out rules should be in rules.txt.
      #@rules.push("if response =~ /^default response/i; response = '...'; score = 2; end")
      #@rules.push("if input =~ /run (.*)/i; response = 'Okay I am running test ' +$1.to_s; score = 5; end")
      #@rules.push("if input =~ /what.*round/i then response = 'This is the qualifying round.'; score += 3 end")
      #@rules.push("if input =~ /when.*contest/ then response = 'The Loebner Contest is being held on Oct. 23 this year (2010).' end")
      @threshhold_rule = 'if response =~ // and score <= @THRESHHOLD then score = @THRESHHOLD + 1 end' # To try to force a response to every input.
      @SAFE_LEVEL = "low"
      @Low_safe_level_words = ["low", "default", "0", "off"]
      @High_safe_level_words = ["high", "on", "1"]
   end

   def delete_threshhold_rule
     @rules.delete(@threshhold_rule)
   end

   def add_threshhold_rule
     if @rules.include?(@threshhold_rule) then return end
     @rules.push(@threshhold_rule)
   end

   def setResponse(response = '', score = 1)
     #puts "[controlbot.setResponse] response==#{response}"
     @responded = true
     puts "[controlbot.setResponse] @responded==#{@responded}"
     begin
       @creator.setResponse(response, score)
     rescue Exception => e
       puts e unless @DEBUG == 0
       $response = response
     end
     puts "\n#{response}\n" #unless @DEBUG == 0
   end

   def getResponse(input)

      input = input.to_s
      print 'INPUT==' + input + "\n" unless @DEBUG == 0

      # Set threshhold.
      if input =~ /^what is your name?/i
         name = @MY_NAME.gsub(/\(\?i\)/, '')
         return name
      end
      if input =~ /^#{@MY_NAME}[:,!]* (quiet|be quiet|stfu|shut.*up|lower.*thresh.?old)/
         @THRESHHOLD = @THRESHHOLD + 1
         delete_threshhold_rule
         return "My threshhold is now " << @THRESHHOLD.to_s
      end
      if input =~ /^#{@MY_NAME}[:,!]* (speak.*up|loud|louder|i can't hear you|raise.*thresh.?old)/
         @THRESHHOLD = @THRESHHOLD - 1
         return "My threshhold is now " << @THRESHHOLD.to_s
      end
      # Return threshhold
      if input =~ /^#{@MY_NAME}[:,!]* what is your (threshhold|threshold|volume)/ or     input =~ /^#{@MY_NAME}[:,!]* (threshold|threshhold|volume)/
         return "My threshhold is currently " << @THRESHHOLD.to_s
      end
      # Add the rule that tries to force a response to each input
      if input =~ /^#{@MY_NAME}[:,!]* add.*thresh.?old rule/
        add_threshhold_rule
        return "Okay, I have added #{@threshhold_rule}."
      end 

      # Provide help message.
      if input =~ /^#{@MY_NAME}[:,!]* help/
         return "I act as a traffic control among agents."
      end

      # trap restart
      case input
        when /^#{@MY_NAME}[:,!]* restart$/, /^restart$/
          return "That command is disabled." # for safety
      end

      case input # trap exit
        when /^exit\b/i
          return "No exit"
      end

      # Set DEBUG level
      case input
        when /^#{@MY_NAME}[:,!]*.*on ?(?:the )?debug(?: mode)?/i,
             /^#{@MY_NAME}[:,!]*.*debug(?: mode)? on/i
          @DEBUG = 1; @agent.DEBUG = 1; return "I have turned on debug output."
        when /^#{@MY_NAME}[:,!]*.*off ?(?:the )?debug(?: mode|setting|level)?/i,
             /^#{@MY_NAME}[:,!]*.*debug(?: mode|setting|level)? off/i
          @DEBUG = 0; @agent.DEBUG = 0; return "I have turned debug output off."
      end

      # Set DELAY
      case input
        when /^#{@MY_NAME}[:,!]*.*(?:the )?delay(?: to) (\d+.?\d+)/i
          @DELAY = $1; @agent.DELAY = $1; return "I have set the delay to #{$1}."
        when /^#{@MY_NAME}[:,!]*.*\bwhat.*(?:the |your )?delay/i
          return "The delay is #{@DELAY}."
      end

      # Set SLEEP
      case input
        when /^#{@MY_NAME}[:,!]*.*\bwhat.*(?:the |your )?sleep/i
          return "The sleep value is set to #{@SLEEP}."
        when /^#{@MY_NAME}[:,!]*.*(?:the )?sleep(?: to) (\d+.?\d+)/i
          begin
            @SLEEP = $1
          rescue 
            @SLEEP = 2
          end
          @agent.SLEEP = @SLEEP; return "I have set the sleep value to #{@SLEEP}."
      end

      # add new if-then rules to post-process responses
      case input
        when /^#{@MY_NAME}[:,!]* (if .*? =~ .*?(?: then|;) .*)$/,
             /^#{@MY_NAME}[:,!]* (.*sub.?\(\/.*\/.*\))$/
          return add_if_rule($1)
      end

      # show the rules
      case input
        when /^#{@MY_NAME}[:,!]*.* show.* rules/
          return show_rules[0]
      end

      # delete a rule
      case input
        when /^#{@MY_NAME}[:,!]*.* delete.* rule.* (\d+)/
          return delete_rule($1.to_i)
      end

      # save rules
      case input
        when /^#{@MY_NAME}[:,!]*.* save.* rules/
          return save_rules
      end

      # set safe level
      case input
        when /^#{@MY_NAME}[:,! ]*.*turn (.*) (?:the )safe(?:ty)? level/,
             /^#{@MY_NAME}[:,! ]*(?:set )?.*safe(?:ty)? level (?:to )?(.*)/
          return set_safe_level($1)
      end

      # get safe level
      case input 
        when /^#{@MY_NAME}[:,! ]*what is.*?safe(?:ty)? level/
          return  get_safe_level
      end


      # Get the agent's response.
      begin
        r, s = @agent.getResponse(input)
      #rescue Errno::EBADF, DRb::DRbConnError
      rescue Exception
        r, s = ['Default response', 0]
      end
      if s == nil then s = 1 end
      if r == nil or r.strip == '' then r = 'Default response' end
      r, s = apply_if_rules(r, s, input)
      if r.to_s.size > @MAX_RESPONSE
         r = r[0..@MAX_RESPONSE]
      end

      if @DEBUG > 0
        puts "\n[controlbot] score==" << s.to_s << "; threshhold==" << @THRESHHOLD.to_s
      end

      if @responded then @responded = false; return end

      # May need Forced response mode for contests.
      if s > @THRESHHOLD 
         return postprocess(r)
      else
         puts "r==#{postprocess(r)}" unless @DEBUG == 0
      end

      return ""
   end

   def postprocess input
      r = input
      # post process here
      return r
   end

   # add an if-then rule
   def add_if_rule(rule = '')
     if rule == '' then return "No rule found." end
     if @SAFE_LEVEL =~ /#{@Low_safe_level_words.join('|')}$/ 
       @rules.push(rule)
       return "Okay I have added #{rule}."
     end
     return "I can't add any if rules because my safe level is #{@SAFE_LEVEL}."
   end

   # apply if-then rules
   def apply_if_rules(response, score, input)
     @rules.each { |rule|
       #puts "rule==#{rule}; response==#{response}; score==#{score}"
       begin
         eval(rule)
       rescue Exception; puts "Can't apply rule: #{rule}"; next; end
     }
     #puts "returning: response==#{response}; score==#{score}"
     return [response, score]
   end

   # return the rules
   def show_rules
     return [@rules.join("\n"), 4]
   end

   # delete a rule by its index
   def delete_rule(i)
     if i < 1 or i > @rules.size
       return "The rules go from 1 to #{@rules.size}."
     end
     rule = @rules[i-1]
     @rules.delete_at(i-1)
     return "Okay, \"#{rule}\" deleted."
   end

   # save rules to file
   def save_rules(file = "rules.txt")
     @rules.uniq!
     comments = Array.new

     # Need to make sure the directory is controlbot/controller
     orig_dir = Dir.pwd
     if orig_dir !~ /controlbot\/controller/
       # Assume you are in /redturnip/mod/bot directory
       Dir.chdir('../../../controlbot/controller')
     end

     begin
       File.open(file, "r+") { |f| 
         line = f.readline
         if line =~ /^#/ then comments.push(line) end
       }
       File.open(file, "w+") { |f|
         f.puts(comments.join("\n"))
         f.puts(@rules.join("\n"))
       }
     rescue => err; return "Eep! #{err}"; end
     Dir.chdir(orig_dir) # Change directory back
     return "Rules saved."
   end

   def read_file(file = "rules.txt")
     count = 0

     # Need to make sure the directory is controlbot/controller
     orig_dir = Dir.pwd
     #if orig_dir !~ /controlbot\/controller/
     if orig_dir =~ /redturnip/
       Dir.chdir('../controlbot/controller')
     end
     #end

     puts "Dir = #{Dir.pwd}"

     begin
       IO.foreach(file) { |statement|
         if statement =~ /^#/ then next end
         statement.chomp!
         if statement.strip == '' then next end
         count += 1
         puts "#{count}: #{statement}" #unless @DEBUG == 0
         @rules.push(statement)
         #response = getResponse(statement)
         #puts response unless @DEBUG == 0
       }
     rescue => boom
       return boom
     end
     Dir.chdir(orig_dir) # Change directory back
     @rules.uniq!
     s = 's'; if count == 1 then s = '' end
     return "Okay, #{file} read: #{count} statement#{s}."
   end

   def set_safe_level(level)
     case level 
       when /\b#{@Low_safe_level_words.join('|')}\b/ 
         @SAFE_LEVEL = "off"
         return "Okay I have turned off the safe level."
       when /\b#{@High_safe_level_words.join('|')}\b/
         @SAFE_LEVEL = "on"
         return "Okay I have turned the safe level on."
     end
     return "I don't understand #{level}."
   end

   def get_safe_level
     return "My safe level is set to #{@SAFE_LEVEL}."
   end

   # Respond to natural language method calls.
   def method_missing input
     return getResponse(input.to_s)
   end

end


if $0 == __FILE__
  b = ControlBot.new(self)
  quitwords = [":q", "quit", "exit", "bye"]
  while (line = Readline.readline("\n> ")) !~ /^(#{quitwords.join('|')})$/
    Readline::HISTORY.push(line)
    r = b.getResponse(line)
    if r == nil then puts $response; next end
    print "\n" + r + "\n"
  end
  puts "Bye!"
  if RUBY_PLATFORM =~ /win/i then exec("exit/rubykill.exe") end # Windows
  if RUBY_PLATFORM =~ /linux/i then exec("sh exit/rubykill.sh") end # Linux
end
