################################################################################
# >
################################################################################

require 'yaml'
require 'fileutils'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/createagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     #puts "input==#{input}"
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def create_agent(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     root = grp1 
     if grp1 =~ /(.*)agent/ then root = process $1 end
    
     error = ''

     Dir.chdir("..")
     begin Dir.mkdir(grp1); rescue => boom; puts boom; end
     begin
       files = ["MyAgent/MyAgent.rb", "MyAgent/MyBot.rb", "MyAgent/TestMyAgent.rb", "MyAgent/TestMyBot.rb", "MyAgent/MyBotd.rb"]
       FileUtils.cp(files, grp1)
     rescue => boom; puts boom; end
     Dir.chdir(grp1)
     begin FileUtils.mv 'MyAgent.rb', "#{grp1}.rb"; rescue => r; puts r; end
     begin FileUtils.mv 'MyBot.rb', "#{root}bot.rb"; rescue => r; puts r; end
     begin FileUtils.mv 'MyBotd.rb', "#{root}botd.rb"; rescue => r; puts r; end
     begin FileUtils.mv 'TestMyBot.rb', "test#{root}bot.rb"; rescue => r; puts r; end
     begin FileUtils.mv 'TestMyAgent.rb', "test#{grp1}.rb"; rescue => r; puts r; end
     
     # Now modify ___agent.rb
     begin file = IO.read("#{grp1}.rb")
       file.sub!(/MyAgent-api.yaml/, "#{grp1}-api.yaml")
       File.open("#{grp1}.rb", "w+") { |f|
         f.puts(file)
       }
     rescue => r; error << r; puts r; end

     # Now modify ___bot.rb
     begin 
       file = IO.read("#{root}bot.rb") 
       file.sub!(/Wrapper for MyAgent/, "Wrapper for #{grp1}")
       file.sub!(/require 'MyAgent'/, "require '#{grp1}'")
       file.sub!(/myagent\|mybot/, "#{grp1}|#{root}bot")
       file.sub!(/I am MyBot, a template for making new bots./, "I am #{root.capitalize}Bot.")
       file.gsub!(/\[MyBot/, "[#{root.capitalize}Bot")
       File.open("#{root}bot.rb", "w+") { |f|
         f.puts(file)
       }
     rescue => r; error << true; puts r; end

     # Now modify test___agent.rb
     begin
       file = IO.read("test#{grp1}.rb")

       file.gsub!(/MyA/, "#{root}a")
       file.sub!(Regexp.new(Regexp.escape( # change the class name back
         "class Test#{root}agent")),
         "class TestMyAgent")
       file.sub!(Regexp.new(Regexp.escape( # change the agent name back
         "@agent = #{root}agent.new")),
         "@agent = MyAgent.new")

       File.open("test#{grp1}.rb", "w+") { |f|
         f.puts(file)
       }
     rescue => r; error << r; puts r; end

     # Now modify test___bot.rb
     begin 
       file = IO.read("test#{root}bot.rb")
       
       file.gsub!(/MyA/, "#{root}a")
       file.gsub!(/MyB/, "#{root.capitalize}B")
       file.gsub!(/mybot/, "#{root}bot")
       file.sub!(Regexp.new(Regexp.escape( # change the class name back
         "class Test#{root.capitalize}Bot")), 
         "class TestMyBot")
       file.sub!(Regexp.new(Regexp.escape( # change the agent name back
         "@agent = #{root.capitalize}Bot.new")),
         "@agent = MyBot.new")

       File.open("test#{root}bot.rb", "w+") { |f|
         f.puts(file)
       }
     rescue => r; error << r; puts r; end 

     # Now modify ___botd.rb
     begin
       file = IO.read("#{root}botd.rb")

       file.sub!(Regexp.new(Regexp.escape("require 'MyBot'")), "require '#{root}bot'")

       File.open("#{root}botd.rb", "w+") { |f| f.puts(file) }
     rescue => r; error << r; puts r; end

     # Now modify ../controller.rb, startagents.rb; 
     # and create __wrapper.rb, batch file

     return "Error! #{error}" unless error == '' 
     return "#{grp1} has been created."
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
