require 'test/unit'
require 'CreateBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('createagent.rb')
      FileUtils.copy('createagent.rb', 'createagent.rb.save')
      begin 
        FileUtils.copy('createagent-api.yaml', 'createagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/createbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("CreateBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("CreateBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("CreateBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("CreateBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("CreateBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("CreateBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("CreateBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("CreateBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('createagent.rb') unless !File.exist?('createagent.rb.save')
    begin; File.delete('createagent-api.yaml'); rescue; end
    if File.exist?('createagent.rb.save')
      FileUtils.copy('createagent.rb.save', 'createagent.rb')
      begin; FileUtils.copy('createagent-api.yaml.save', 'createagent-api.yaml'); rescue; end
    end
    if File.exist?('createagent.rb') then File.delete('createagent.rb.save') end
    if File.exist?('createagent-api.yaml.save') then File.delete('createagent-api.yaml.save') end
  end

end
