################################################################################
# Recites verses from the Dhammapada.
# 
# > what is the Dhammapada?
#
#   the Dhammapada is a versified buddhist scripture traditionally ascribed 
#   to the buddha himself
#
# > recite a random verse please
#
#   From pleasure arises sorrow, from pleasure arises fear, but he who is freed 
#   from pleasure has no sorrow and certainly no fear. 214
#
# > recite verse 42 
#
#   One's own misdirected thought can do one more harm than an enemy or an 
#   ill-wisher.
#
# > find "verses"
#
#   Better than a thousand pointless verses is one stanza on hearing which one 
#   finds peace. 101  [more]
#
# > more
#
#   Better than reciting a hundred pointless verses is one verse of the 
#   teaching (one dhammapada) on hearing which one finds peace. 102
#
# > 
################################################################################

require 'yaml'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/dpagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^show (.*)'s api/, :show_one_method_api] 
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     #@DEBUG = 1
     @DEBUG = 0  # debug off

     @found_verses = Array.new
     @save_find = ''
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     #puts "input==#{input}"
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     # If still no match, try again with slightly different regexps.
    if method == '' then grp1 = %r{(?i-mx:#{$1})}; grp2 = %r{(?i-mx:#{$2})}; end
     @patterns.each {|p|; if p[0].to_s == grp2.to_s; method = p[1]; break; end}
     if method == '' then grp1 = %r{#{$1}}; grp2 = %r{#{$2}}; end
     @patterns.each {|p|; if p[0].to_s == grp2.to_s; method = p[1]; break; end}
     if method == '' then grp1 = %r{^#{$1}}; grp2 = %r{^#{$2}}; end
     @patterns.each {|p|; if p[0].to_s == grp2.to_s; method = p[1]; break; end}
 
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '') # strip initial regexp delimiter /
      str.sub!(/\/$/, '') # strip final regexp delimiter /
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil,input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apihash = Hash.new
     @patterns.each { |p|
       regexp = p[0].to_s
       method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       if apihash.has_key? method then next end
       apihash[method] = regexp
       puts "[show_api] apihash[method]=#{method}, regexp=#{regexp}" unless @DEBUG == 0
     }
     # Try to pretty-print.
     toReturn = ''
     apihash.each_pair { |key, value|
       puts "1. value=#{value}" unless @DEBUG == 0
       method = ''
       if value =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/
         puts "   2. value=#{value}; $1=#{$1}" unless @DEBUG == 0
         method = $1.to_s
       end
       if key == nil or key == '' then key = "*" end
       toReturn << key.to_s << " | " << method.to_s << "\n"
     }
     return toReturn
   end # method_def

   def show_one_method_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1 # Supplied method name.
     apiarr= Array.new # To hold methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p| # Collect all the regexps for the supplied method.
       regexp = p[0].to_s; method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}." unless @DEBUG == 0
       if method == grp1 then apiarr.push [regexp, method] end 
     } 
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

# Begin dpagent methods.

   def recite_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     dpfile = File.new("dhammapada.txt", "r")
     while (line = dpfile.gets)
       line.chomp!
       case line
         # search for lines with the verse number at the end
         # sometimes there are two numbers: 58, 59
         when /(.*) #{grp1},(?: )*\d+(?: )*$/,
              /(.*) \d+,(?: )*#{grp1}(?: )*$/,
              /(.*) #{grp1}(?: )*$/
           @score = 4
           return $1
       end
     end
     return "I can't find verse #{grp1}."
   end # method_def

   def find_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1

     #puts "[find_verse] grp1==#{grp1}; @save_find==#{@save_find}; @more==#{@more}"

     if grp1 != @save_find then @more = false end

     if @found_verses.size == 1 and @more
       @save_find = grp1
       response = @found_verses[0]
       @found_verses.clear
       @more = false
       @save_find = ''
       @score = 4
       return response
     end
     if @found_verses.size > 1 and @more
       @save_find = grp1
       first_verse = @found_verses[0]
       @found_verses = @found_verses[1..-1]
       @score = 4
       return first_verse + " [more]"
     end

     dpfile = File.new("dhammapada.txt")
     toreturn = Array.new
     while (line = dpfile.gets)
       if line =~ /#{grp1}/i
         toreturn.push line.chomp
       end
     end
     if toreturn.empty? then return "I can't find #{grp1}." end 
     @score = 4
     #puts "[find_verses] toreturn==#{toreturn}"
     if toreturn.size == 1 then return toreturn[0] end
     if toreturn.size > 1
       @found_verses = toreturn[1..-1]
       @save_find = grp1
       return toreturn[0] + " [more]"
     end
   end # method_def

   def more(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @save_find == '' then return end
     @more = true
     #puts "[more] @save_find==#{@save_find}"
     return getResponse("find #{@save_find}")[0]
   end # method_def

   def random_verse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     dpfile = File.new("dhammapada.txt")
     r = rand(423)
     count = 0
     while (line = dpfile.gets)
       if count == r
         @score = 4
         return line.chomp
       end
       count = count + 1
     end 
     return "Eep! I couldn't find a verse!"
   end # method_def

   def on_irc_channel(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from on_irc_channel'
   end # method_def


   def on_irc_channel(pattern=nil, input=nil)
     return self.send("find irc channel")[0]
   end # method_def

   def on_ban(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return self.send("find irc channel")[0]
   end # method_def

   def what_is_the_dp(pattern=nil,input=nil)
     @score = 4
     return "The Dhammapada is a versified Bhuddhist scripture traditionally assigned to the Buddha himself."
   end # method_def

   def add(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     dpfile = File.new("dhammapada.txt", "a+")
     dpfile.seek(-10, IO::SEEK_END)
     ending = dpfile.readline
     if ending =~ / (\d+)/
       last = $1.to_i
       new = last + 1
     end 
     if new != 0
       dpfile.puts(grp1 + ". #{new}")
       dpfile.close
       return "Okay, I have added #{grp1} as verse #{new}."
     end
     dpfile.puts(grp1)
     dpfile.close
     return "I have added #{grp1}."
   end # method_def

   def how_many_x_is_y(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from how_many_x_is_y'
   end # method_def

   def on_namokar_mantra(pattern=nil, input=nil)
     return self.send("find namokar")[0]
   end # method_def


  def on_ban(pattern=nil, input=nil); if input =~ /\bban\b/ then self.send("@dp 424") end; end # method_def

     def on_bhasa_samiti(pattern=nil, input=nil)
       @score += 3; return self.send("find bhasa")[0]
     end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
