require 'test/unit'
require 'dpagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('dpagent.rb')
      FileUtils.copy('dpagent.rb', 'dpagent.rb.save')
      begin 
        FileUtils.copy('dpagent-api.yaml', 'dpagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_recite
    actual, s = @agent.send("recite verse 100")
    assert_match(/a thousand pointless words/, actual)
  end

  def test_random
    actual, s = @agent.send("@dp")
    assert_match(/\d+(?: )*$/, actual)
  end

  def test_find
    actual, s = @agent.send("find mind")
    assert_match(/\[more\]/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def teardown
    File.delete('dpagent.rb') unless !File.exist?('dpagent.rb.save')
    begin; File.delete('dpagent-api.yaml'); rescue; end
    if File.exist?('dpagent.rb.save')
      FileUtils.copy('dpagent.rb.save', 'dpagent.rb')
      begin; FileUtils.copy('dpagent-api.yaml.save', 'dpagent-api.yaml'); rescue; end
    end
    if File.exist?('dpagent.rb') then File.delete('dpagent.rb.save') end
    if File.exist?('dpagent-api.yaml.save') then File.delete('dpagent-api.yaml.save') end
  end

end
