require 'test/unit'
require 'emoagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('emoagent.rb')
      FileUtils.copy('emoagent.rb', 'emoagent.rb.save')
      begin 
        FileUtils.copy('emoagent-api.yaml', 'emoagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
    @agent.send("reset graph")
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_pqr
    actual, s = @agent.send("suppose: P")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("P -> Q")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("Q -> R")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("But R is absurd!")
    assert_match(/Therefore, not-P/i, actual)
  end

  def test_if_then
    actual, s = @agent.send("Suppose AI succeeds.")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("If AI succeeds, then people skills become irrelevant.")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("If people skills become irrelevant, I wasted time learning people skills.")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("IF I wasted time learning people skills then I feel bad.")
    assert_match(/Okay/i, actual)

    actual, s = @agent.send("But I feel bad is absurd! We cannot allow me to feel bad!")
    assert_match(/Therefore/i, actual)
  end

  def test_x_is_not_y
    actual, s = @agent.send("Assume: We should watch this game.")
    actual, s = @agent.send("if we should watch this game, it should be interesting.")
    actual, s = @agent.send("if it should be interesting, this scrimmage should be interesting.")
    actual, s = @agent.send("This scrimmage is not interesting.")
    assert_match(/Therefore, we should not watch this game/i, actual)
  end

  def test_i_cannot_be_wrong
    actual, s = @agent.send("Assume: We should not watch this game.")
    actual, s = @agent.send("If we should not watch this game, then I was wrong to recommend it so enthusiastically.")
    actual, s = @agent.send("But I can't be wrong!")
    assert_match(/Therefore,/i, actual)

    actual, s = @agent.send("But I was wrong that one time...")
    assert_match(/Therefore, we should not have to watch this game/i, actual)
  end

  def test_forget
    actual, s = @agent.send("P -> Q")
    actual, s = @agent.send("Q -> R")
    actual, s = @agent.send("does P -> R?")
    assert_match(/Yes/i, actual)
    actual, s = @agent.send("forget that Q -> R.")
    assert_match(/Okay/i, actual)
    actual, s = @agent.send("Does P -> R?")
    assert_match(/no knowledge/, actual)
  end

  def test_questions
    actual, s = @agent.send("A -> B")
    actual, s = @agent.send("B -> C")
    actual, s = @agent.send("C -> D")
    actual, s = @agent.send("Assume: A")
    assert_match(/Okay/i, actual)
    actual, s = @agent.send("D is absurd!")
    assert_match(/not-A/i, actual)
    actual, s = @agent.send("any questions?")
    assert_match(/Yes/i, actual)
  end

  def test_clear_questions
    actual, s = @agent.send("a -> b")
    actual, s = @agent.send("b -> c")
    actual, s = @agent.send("c -> d")
    actual, s = @agent.send("assume a")
    actual, s = @agent.send("d is absurd!")
    actual, s = @agent.send("do you have any questions?")
    assert_match(/Yes/i, actual)
    actual, s = @agent.send("clear questions")
    assert_match(/cleared.*3/, actual)
    actual, s = @agent.send("do you have any questions?")
    assert_match(/No questions/i, actual)
    actual, s = @agent.send("clear questions")
    assert_match(/don't have/i, actual)
  end

  def teardown
    File.delete('emoagent.rb') unless !File.exist?('emoagent.rb.save')
    begin; File.delete('emoagent-api.yaml'); rescue; end
    if File.exist?('emoagent.rb.save')
      FileUtils.copy('emoagent.rb.save', 'emoagent.rb')
      begin; FileUtils.copy('emoagent-api.yaml.save', 'emoagent-api.yaml'); rescue; end
    end
    if File.exist?('emoagent.rb') then File.delete('emoagent.rb.save') end
    if File.exist?('emoagent-api.yaml.save') then File.delete('emoagent-api.yaml.save') end
  end

end
