require 'test/unit'
require 'emobot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('emoagent.rb')
      FileUtils.copy('emoagent.rb', 'emoagent.rb.save')
      begin 
        FileUtils.copy('emoagent-api.yaml', 'emoagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/emobot/i, actual)
  end

  def test_volume
    actual, s = @agent.send("EmoBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("EmoBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("EmoBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("EmoBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("EmoBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("EmoBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("EmoBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("EmoBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("EmoBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('emoagent.rb') unless !File.exist?('emoagent.rb.save')
    begin; File.delete('emoagent-api.yaml'); rescue; end
    if File.exist?('emoagent.rb.save')
      FileUtils.copy('emoagent.rb.save', 'emoagent.rb')
      begin; FileUtils.copy('emoagent-api.yaml.save', 'emoagent-api.yaml'); rescue; end
    end
    if File.exist?('emoagent.rb') then File.delete('emoagent.rb.save') end
    if File.exist?('emoagent-api.yaml.save') then File.delete('emoagent-api.yaml.save') end
  end

end
