# Scrape assertions from openmind.org's "Highest rated" listings.

require 'net/http'
require 'uri'

$Delay = 5

assertions = Array.new

if ARGV[0] =~ /(\?|h)/i 
  puts "USAGE: ruby #{File.basename($0)} startpage# endpage#" 
  exit
end

# get start and end pages from the command-line
startp = ARGV[0]
endp = ARGV[1]
if startp =~ /(\d+)-(\d+)/; startp = $1; endp = $2; end
if startp == nil or startp.to_i <= 0 then startp = 1 end
if endp == nil or endp.to_i <= 0 then endp = 1 end

# scrape the web pages
url = URI.parse('http://commons.media.mit.edu')
puts "#{startp} - #{endp}"
(startp.to_i).upto(endp.to_i) do |p|
  puts "#{p}: starting HTTP..."
  res = Net::HTTP.start(url.host, url.port) { |http|
    puts "sleeping for #{$Delay} seconds..."
    sleep($Delay)
    http.get("/en/assertion/best/?page=#{p}")
  }
  page = res.body

  # Read files from disk, for testing.
  #page = String.new
  #file = File.new("page6.htm", "r")
  #file.each_line { |line|; page << line; }
  
  # extract assertions
  startString = Regexp.escape("class=\"predlink\">")
  endString = Regexp.escape("<td class=\"creator\">")
  started = false
  count = 0
  page.each_line { |line|
    if line =~ /#{startString}/
      started = true
      next
    end
    if started == true and line =~ /#{endString}/
      started = false
    end
    if line == '' or line == nil then next end
    if started
      if line =~ /<\/td>$/
        line.gsub!(/\<.*?\>/, '')
        assertions.push(line.chomp.strip)
      end
    end
  }
end

# write to disk
outfile = File.new("assertions-#{startp}-#{endp}", "w")
assertions.each_with_index { |assertion, index|
  puts "assertion ##{index}: #{assertion}"
  outfile.puts(assertion)
}
outfile.close
