require 'test/unit'
require 'GoogleBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('googleagent.rb')
      FileUtils.copy('googleagent.rb', 'googleagent.rb.save')
      begin 
        FileUtils.copy('googleagent-api.yaml', 'googleagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/googlebot/, actual)
  end

  def test_volume
    actual, s = @agent.send("GoogleBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("GoogleBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("GoogleBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("GoogleBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("GoogleBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("GoogleBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("GoogleBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("GoogleBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("googlebot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("googlebot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("googlebot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("googlebot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('googleagent.rb') unless !File.exist?('googleagent.rb.save')
    begin; File.delete('googleagent-api.yaml'); rescue; end
    if File.exist?('googleagent.rb.save')
      FileUtils.copy('googleagent.rb.save', 'googleagent.rb')
      begin; FileUtils.copy('googleagent-api.yaml.save', 'googleagent-api.yaml'); rescue; end
    end
    if File.exist?('googleagent.rb') then File.delete('googleagent.rb.save') end
    if File.exist?('googleagent-api.yaml.save') then File.delete('googleagent-api.yaml.save') end
  end

end
