################################################################################
# Parses arbitrary S-V-O sentences into a graph, where:
#   S is a vertex, V is the label on an edge, O is a vertex
#
# Then answers questions about the sentence.
#
# > John loves Mary.
# Okay
#
# > does John love Mary?
# Yes
#
# > who does John love?
# Mary
#
# > who is loved by John?
# Mary
#
# > does Mary love John
# I don't know
#
# > is Mary loved by John
# Yes
#
# > why does John love Mary?
# Because you said John loves Mary 
#
################################################################################

require 'yaml'
require 'graph'
require 'drb'

BasicSocket.do_not_reverse_lookup = true

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :plurals, :singulars, :verbs, :verb_forms, :names 

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/graphagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''

     @graph = Graph.new

     @plurals = Array.new
     @plurals << "you"; @plurals << "I"
     @singulars = Array.new
     @verbs = [ "is", "=", "are", "has", "have", "eat", "eats", "is had by", "are had by", "student teach", "student teaches", "contains" ]
     @verb_forms = { "is" => "are", "has" => "have", "eats" => "eat", "is had by" => "are had by", "student teaches" => "student teach", "contains" => "contain" }

     @names = Array.new
     @save = false
     @already_asked = false
     @quit_question_state = ''

     DRb.start_service()
     @monty = DRbObject.new(nil, 'druby://localhost:9070')
     @link = DRbObject.new(nil, 'druby://localhost:9050')

     @those = nil

     # @last_file_read must be the last line in initialize 
     @last_file_read = "highest.txt"
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     input = strip_names(input)

     #puts "input==#{input}"
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), 2] rescue return r
       end
     }
     return r
   end

   # Strip names from input (don't strip if input is "NAME is not a name")
   # i.e., an input: "John: Socrates is a man" -> "Socrates is a man"
   def strip_names input
     @names.each { |name|
       escaped_name = Regexp.escape(name)
       if input =~ /^#{escaped_name}[:, ]*(.*)/
         new_input = $1.strip
         break if new_input =~ /^is not a name/
         input = new_input 
         break
       end
     }
     return input
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end


   def quit(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @save then return 'quit' end
     if @already_asked then return 'quit' end
     #if @already_asked then return end
     @already_asked = true
     @quit_question_state = "waiting for yes/no"
     return "Do you wish to save the graph to graph.yaml?" 
   end

   def no(pattern=nil, input=nil)
     #puts "@quit_question_state==#{@quit_question_state}"
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @quit_question_state == "waiting for yes/no"
       @quit_question_state = 'answered'
       return 'quit' 
     end
     return ''
   end

   def yes(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @quit_question_state == "waiting for yes/no"
       @quit_question_state = 'answered'
       self.send("save graph")
       return 'quit'
     end
     return ''
   end

   def make_verb_agree(s)
     begin
       if s =~ /(.*) (#{@verbs.join('|')}) (.*)/
         a = process $1; r = process $2; b = process $3
         verb = getForm(a, r)
         return "#{a} #{verb} #{b}"
       end
     rescue
       return s
     end
     return s
   end

   def getForm(a, verb = "is")
     if isPlural?(a)
       return getPlural(verb)
     end
     return verb
   end

   def getPlural(verb)
     @verb_forms.fetch(verb, verb)
   end

   def isPlural?(word = '')
     @plurals.each { |plural|
       if word =~ /^#{plural}$/i then return true end
     }
     if word =~ /socrates$/i; return false; end
     if word =~ /men$/i; return true; end
     @singulars.each { |singular|
       if word =~ /^#{singular}$/i then return false end
     }
     if word =~ /s$/i; return true; end
   end

   def print_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     s = @graph.to_s
     arr2 = Array.new
     arr = s.split("\n")
     #puts "arr==#{arr}"
     arr.each { |phrase|
       phrase = make_verb_agree(phrase)
       arr2 << phrase
     }
     #return s unless s == nil or s.size == 0
     return arr2.join("\n") unless arr2.empty?
     return "The graph is empty."
   end # method_def

   def add_to_plurals(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     @plurals.push(grp1)
     return "Okay I have recorded #{grp1} as a plural form."
   end # method_def

   def add_to_singulars(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     @singulars.push(grp1)
     return "Okay I have recorded #{grp1} as a singular form."
   end # method_def

   def save_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     file = nil 
     begin
       file = process $1
     rescue; end
     puts "\n... Saving graph (this may take minutes) ..."
     r = @graph.save(file)
     return "#{r}"
   end # method_def

   def load_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     begin
       file = process $1
     rescue; end
     return @graph.load(file)
   end

   def reset_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @graph = Graph.new(false)
     @save = false
     return "Okay, I have reset the graph."
   end # method_def

   def how_many_facts(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     begin; a = @graph.vertex_hash.keys; rescue => boom; puts boom; end
     puts 'hello'
     n = a.size
     vertices = "vertices"; if n == 1 then vertices = "vertex" end
     return "I have #{n} #{vertices} stored in my database." 
   end # method_def

   def add_name(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     #grp1 = process $1
     grp1 = $1.strip
     grp1.sub!(/^(?:"|')/, '')
     grp1.sub!(/(?:"|')$/, '')
     grp1.sub!(/(?:,)$/, '')
     @names.push(grp1)
     return "Okay, #{grp1} is a name."
   end # method_def

   def what_are_names(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return @names.join(", ")
   end # method_def

   def set_save_mode_off(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @save = false
     return "Okay I have turned off save mode."
   end # method_def

   def set_save_mode_on(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @save = true
     return "Okay I have turned on save mode."
   end # method_def

   def x_is_not_a_name(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     name_to_del = $1
     result = @names.delete(name_to_del)
     if result == nil
       return "I can't find #{name_to_del} among the names I have stored." 
     end
     return "Okay, I have deleted #{result}."
   end # method_def

   def is_x_a_name(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     name = $1
     if @names.include?(name)
       return "Yes, #{name} is a name."
     end
     return "No, #{name} is not a name."
   end # method_def
   alias is_x_a_name? is_x_a_name

   def what_is_save_mode(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     save_mode = "on"; save_mode = "off" unless @save
     return "The save mode is #{save_mode}."
   end # method_def

   def read_file(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     file = process $1
     if file !~ /^highest.txt/
       if file !~ /^openmind(\\|\/)/ then file = "openmind/" + file end
     end
     turn_save_back_on = false
     if @save; @save = false; turn_save_back_on = true; end
     count = 0
     IO.foreach(file) { |assertion|
       if assertion =~ /^#/ then next end
       if assertion.strip == '' then next end
       count += 1
       puts "#{count}: " + assertion
       self.send(assertion)
     }
     if turn_save_back_on; self.send("save graph"); @save = true; end
     @last_file_read = file
     write_last_file(file) #unless @save == false
     plural = 's'; if count == 1 then plural = ''; end
     return "Okay, #{file} read; #{count} assertion#{plural}."
   end # method_def

   def write_last_file(file)
     orig = IO.read(__FILE__)
     orig =~ /(.*?)\@last_file_read \= (.*?)(   end.*)/m
     grp1 = $1; grp2 = $2; grp3 = $3
     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + "@last_file_read = \"" + file + "\"\n" + grp3
     }
   end

   def what_was_last_file_read(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return "The last file I read was: #{@last_file_read}."
   end # method_def

   def how_many_factoids(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     hash = @graph.vertex_hash
     count = 0
     keys = hash.keys
     keys.each { |key|
       values = @graph.vertex_hash.fetch(key)
       values.each { |value|
         count += 1
         #puts "#{count}: #{value}"
       }
     }
     return "I know #{count} factoids."
   end # method_def

   def how_many_factoids(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     hash = @graph.vertex_hash
     count = 0
     keys = hash.keys
     keys.each { |key|
       values = @graph.vertex_hash.fetch(key)
       values.each { |value|
         count += 1
         #puts "#{count}: #{value}"
       }
     }
     return "I know #{count} factoids."
   end # method_def

   def what_are_those(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @those != nil
       arr = @those
       @those = nil
       return "#{arr.join(', ')}"
     end
     if @those != nil and @those.size == 0
       @those = nil
       return "Nothing"
     end
     return "I don't know what 'those' refers to."
   end # method_def


   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def s_v_o(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     sentence = process $1

     subj, verb, obj = parse_using_monty(sentence)
     if subj =~ /I can't process /i then return subj end
     
     @graph.a_r_b!(subj, verb, obj)
     if @save then @graph.save end
     return "Okay, #{subj} #{verb} #{obj}"
   end # method_def

   def parse_using_monty sentence 
     subj = @monty.send("subject: #{sentence}")
     verb = @monty.send("verb: #{sentence}")
     obj  = @monty.send("object: #{sentence}")
     puts "subj==#{subj}; verb==#{verb}; obj==#{obj}"
     abort = false
     if subj =~ /I don't know what the subject /i then abort = true end
     if verb =~ /I don't know what the verb /i then abort = true end
     if obj =~ /I don't know what the object /i then abort = true end
     if abort then return "I can't process #{sentence}" end
     return subj, verb, obj
   end

   def parse_using_link sentence
     subj = @link.send("subject: #{sentence}")
     verb = @link.send("verb: #{sentence}")
     obj = @link.send("obj: #{sentence}")
     puts "[graphagent.parse_using_link] subj=#{subj}; verb=#{verb}; obj=#{obj}"
    
     abort = false 
     if subj =~ /I don't /i then abort = true end
     if verb =~ /I don't /i then abort = true end
     if obj =~ /I don't /i then abort = true end
     if abort then return "I can't process #{sentence}" end

     if obj =~ /Default response/i
       sentence =~ /#{subj}.*#{verb} (.*)/
       obj = process $1.to_s 
     end

     return subj, verb, obj
   end

   def does_a_r_b(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     sentence = process $1

     subj, verb, obj = parse_using_link(input)
     puts "[graphagent.does_a_r_b] subj==#{subj}; verb==#{verb}; obj==#{obj}"

     #if verb =~ /\bhave\b/ then verb.gsub!(/have/, 'has') end
     if verb =~ /\w+/ then verb = verb + "s" end
     
     if @graph.a_r_b?(subj, verb, obj) then return "Yes" end

     if @graph.a_r_b?(subj.downcase, verb.downcase, obj.downcase) then return "yes" end
     
     return "I don't know whether #{subj} #{verb} #{obj}."
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
