require 'test/unit'
require 'graphagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('graphagent.rb')
      FileUtils.copy('graphagent.rb', 'graphagent.rb.save')
      begin 
        FileUtils.copy('graphagent-api.yaml', 'graphagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("reset graph")
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    #assert_match(/again/, actual) # problem: MyAgent.s_v_o is called
  end

  def test_save_load_reset
    actual, s = @agent.send("reset graph")
    actual = @agent.send("print graph")[0]
    assert_match(/empty/, actual)
    actual = @agent.send("a = b")[0]
    assert_match(/Okay/, actual)
    actual = @agent.send("save graph test.graph")[0]
    assert_match(/saved/, actual)
    actual = @agent.send("reset graph")[0]
    assert_match(/Okay/, actual)
    actual = @agent.send("y = z")[0]
    assert_match(/Okay/, actual)
    actual = @agent.send("print graph")[0]
    assert_match(/y = z/, actual)
    actual = @agent.send("load graph test.graph")[0]
    actual = @agent.send("print graph")[0]
    assert_match(/a = b/, actual)
    File.delete("test.graph")
  end

  def test_save_mode
    actual, s = @agent.send("what is the save mode?")
    assert_match(/\boff\b/, actual)
    actual, s = @agent.send("set save mode on")
    actual, s = @agent.send("what is the save mode?")
    assert_match(/\bon\b/, actual)
  end

  def test_names
    actual, s = @agent.send("joe is a name")
    assert_match(/Okay\b/, actual)
    actual, s = @agent.send("joe: a is b")
    assert_match(/Okay\b/, actual)
    actual, s = @agent.send("what is a?")
    #assert_match(/\bb\b/, actual)

    actual, s = @agent.send("what are the names?")
    assert_match(/\bjoe\b/, actual)

    actual, s = @agent.send("joe is not a name")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("what are the names?")
    assert_match(//, actual)
  end

=begin
  def test_read_file
    actual, s = @agent.send("read highest.txt")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("what is something you find at the beach?")
    assert_match(/sand/, actual)

    #actual, s = @agent.send("read openmind\\assertions-1-1")
    #assert_match(/Okay/, actual)

    actual, s = @agent.send("reset graph")
    #actual, s = @agent.send("read openmind/highest.txt")
    actual, s = @agent.send("read highest.txt")
    assert_match(/Okay/, actual)
    actual, s = @agent.send("how many facts?")
    assert_match(/164/, actual)

    actual, s = @agent.send("what was the last file read")
    assert_match(/highest.txt/, actual)

  end
=end

  def test_John_loves_Mary
    @agent.send("reset graph")
    actual, s = @agent.send("John loves Mary")
    actual, s = @agent.send("Does John love Mary?")
    assert_match(/Yes/i, actual)
    actual, s = @agent.send("who does John love?")
    assert_match(/Mary/i, actual)
    actual, s = @agent.send("why does John love Mary?")
    assert_match(/because/i, actual)
    actual, s = @agent.send("who is loved by John?")
    assert_match(/Mary/i, actual)
    actual, s = @agent.send("who loves Mary?")
    assert_match(/John/i, actual)
    actual, s = @agent.send("is Mary loved by John?")
    assert_match(/Yes/i, actual)
    actual, s = @agent.send("who is Mary loved by?")
    assert_match(/John/i, actual)
  end

  def teardown
    File.delete('graphagent.rb') unless !File.exist?('graphagent.rb.save')
    begin; File.delete('graphagent-api.yaml'); rescue; end
    if File.exist?('graphagent.rb.save')
      FileUtils.copy('graphagent.rb.save', 'graphagent.rb')
      begin; FileUtils.copy('graphagent-api.yaml.save', 'graphagent-api.yaml'); rescue; end
    end
    if File.exist?('graphagent.rb') then File.delete('graphagent.rb.save') end
    if File.exist?('graphagent-api.yaml.save') then File.delete('graphagent-api.yaml.save') end
  end

end
