###############################################################################
# Interface to the Hunters program. See READMEs for full output of example runs.
#
# > start hunters
# [Runs hunters.exe, prompts for 16 parameters]
#
# > show output
# [Displays gen_output.dat, generated by each run of hunters.exe]
#
# > run hunters with a default of 10
# [Runs hunters putting 10 for all parameters except rounds, where it puts 1]
#
# > repeat [last run]
# [Runs hunters.exe again with the same parameters as the previous run]
# 
# > who won?
# A=0; S=6; H=4; Sharers won [this run].
# Total wins: Attackers: 0; Sharers: 3; Hoarders: 0.
# 
################################################################################

require 'yaml'
require 'hunters/hunters.rb'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/huntersagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^show (.*)'s api/, :show_one_method_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0

     @last = Array.new
     @attackers = 0; @sharers = 0; @hoarders = 0
     @last_output = Array.new
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     puts "input==#{input}" unless @DEBUG == 0
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   # a is a synonym for b, or /a/ is a synonym for /b/
   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     g1 = $1; g2 = $2; #puts "[add_synonym] g1==#{g1}; g2=#{g2}" 
     # Remove enclosing "/", if present.
     if g1 =~ /^\// then g1.sub!(/^\//, '') end
     if g1 =~ /\/$/ then g1.sub!(/\/$/, '') end
     if g2 =~ /^\// then g2.sub!(/^\//, '') end
     if g2 =~ /\/$/ then g2.sub!(/\/$/, '') end
     #puts "[add_synonym] g1==#{g1}; g2=#{g2}"
     grp1 = %r{(?i-mx:^#{g1})}
     grp2 = %r{(?i-mx:^#{g2})}
     #puts "[add_synonym] grp1==#{grp1}; grp2=#{grp2}"
     method =''
     index = -1 # To add the new pattern before the old one.
     @patterns.each_with_index { |p,i| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          index = i
          break
       end
     }
     # If still no match, try again with slightly different regexps.
    if method == '' then grp1 = %r{(?i-mx:#{g1})}; grp2 = %r{(?i-mx:#{g2})}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{#{g1}}; grp2 = %r{#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
     if method == '' then grp1 = %r{^#{g1}}; grp2 = %r{^#{g2}}; end
     @patterns.each_with_index {|p,i| 
       if p[0].to_s == grp2.to_s; method = p[1]; index = i; break; end}
 
     return unless method.to_s != nil
     puts "add_synonym: index==#{index}" unless @DEBUG == 0
     @patterns.insert(index, [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       puts "@new_regexp==#{@new_regexp}" unless @DEBUG == 0

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def show_one_method_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1 # Supplied method name.
     apiarr= Array.new # To hold methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p| # Collect all the regexps for the supplied method.
       regexp = p[0].to_s; method = p[1].to_s
       puts "regexp=#{regexp}; method=#{method}." unless @DEBUG == 0
       if method == grp1 then apiarr.push [regexp, method] end 
     } 
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

# Begin agent-specific methods.

   def kill_hunters(pattern=nil, input=nil)
     begin Dir.chdir('hunters') rescue Exception end
     hunters = Hunters.new
     Dir.chdir('..')
     return hunters.kill_hunters
   end # method_def

   def run_hunters(pattern=nil, input=nil)
     toReturn = ''
     @last = Array.new
     Dir.chdir('hunters')
     hunters = Hunters.new
     toReturn << hunters.getstr
     print '     ' << toReturn
     end_loop = 0
     while (line = gets) !~ /^(#{$Quitwords.join('|')})$/
       @last.push(line)
       hunters.putstr('     ' << line.to_s)
       response = hunters.getstr
       if response =~ /(.*)\n(.*)/m then response = "#{$1}\n     #{$2}" end
       print '     ' << response 
       toReturn << response
       if end_loop == 1 then break end
       if response =~ /delay/ then end_loop += 1 end
     end
     Dir.chdir('..')
     return hunters.getstr
   end # method_def

   def run_hunters_with_default(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @last = Array.new
     response = ''
     grp1 = process $1
     Dir.chdir('hunters')
     hunters = Hunters.new
     14.times do
        print hunters.getstr; puts " #{grp1}"
        hunters.putstr(grp1); @last.push(grp1)
     end
     print hunters.getstr; puts " 1" 
     hunters.putstr("1"); @last.push('1')
     print hunters.getstr; puts " #{grp1}" 
     hunters.putstr(grp1); @last.push(grp1)
     response = hunters.getstr
     Dir.chdir('..')
     return response
   end # method_def

   def repeat_last(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if @last.size != 16 then return "No default! #{@last.join(', ')}." end
     response = ''
     Dir.chdir('hunters')
     hunters = Hunters.new
     @last.each { |e|
       e.chomp!
       print hunters.getstr; puts "#{e}"
       hunters.putstr(e)
     }
     response = hunters.getstr
     Dir.chdir('..')
     @last_output = '' # to get accurate totals?
     winners, s = self.send("who won?")
     return response << "\n" << winners
   end # method_def

   def print_gen_output(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     Dir.chdir('hunters')
     begin
       gen = File.new("gen_output.dat", "r")
     rescue; Dir.chdir('..'); return "No gen_output.dat!"; end
     Dir.chdir('..')
     return gen.readlines
   end # method_def

   def pick_winner(a,s,h,same)
     a1 = a >= s; if a1 then a1 = a >= h end
     s1 = s >= a; if s1 then s1 = s >= h end
     h1 = h >= a; if h1 then h1 = h >= s end
     winners = Array.new
     if same # "who won" was asked twice without an intervening simulation
       if a1 then winners.push("Attackers") end
       if s1 then winners.push("Sharers") end
       if h1 then winners.push("Hoarders") end
       return "#{winners.join(' and ')} won. \nTotal wins: Attackers: #{@attackers}; Sharers: #{@sharers}; Hoarders: #{@hoarders}." 
     end
     if a1 then winners.push("Attackers"); @attackers += 1 end
     if s1 then winners.push("Sharers"); @sharers += 1 end
     if h1 then winners.push("Hoarders"); @hoarders += 1 end
     return "#{winners.join(' and ')} won. \nTotal wins: Attackers: #{@attackers}; Sharers: #{@sharers}; Hoarders: #{@hoarders}."
   end # method_def

   def who_won(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     output,s = self.send("print output")
     #puts output 
     if @last_output == output then same = true end
     @last_output = output
     last = ''; a, s, h = 0
     if output.join =~ /(?:.*)Stategies at end(.*)$/m then last = $1.to_s end
     if last =~ /attacking.*\(Alive\: (.*)\,/ then a = $1.to_i end
     if last =~ /sharing.*\(Alive\: (.*)\,/ then s = $1.to_i end
     if last =~ /hoarding.*\(Alive\: (.*)\,/ then h = $1.to_i end
     total = pick_winner(a, s, h, same)
     return "A=#{a}; S=#{s}; H=#{h}; #{total}" 
   end # method_def

   def repeat_x_times(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     begin 
       x = grp1.to_i 
     rescue Exception; return "I don't understand #{grp1}."; end
     x.times do
       self.send("repeat")
       puts self.send("who won?")[0]
     end
     return self.send("who won?")[0]
   end # method_def

   def reset_totals(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @attackers = 0; @sharers = 0; @hoarders = 0
     return "Okay, I have reset the total wins: Attackers #{@attackers}; Sharers #{@sharers}; Hoarders #{@hoarders}."
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  bot.kill_hunters('', '')
  puts "Bye"
end # of file
