require 'test/unit'
require 'ifbot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('ifagent.rb')
      FileUtils.copy('ifagent.rb', 'ifagent.rb.save')
      begin 
        FileUtils.copy('ifagent-api.yaml', 'ifagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/ifbot/i, actual)
  end

  def test_volume
    actual, s = @agent.send("IfBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("IfBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("IfBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("IfBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("IfBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("IfBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("IfBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("IfBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("IfBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('ifagent.rb') unless !File.exist?('ifagent.rb.save')
    begin; File.delete('ifagent-api.yaml'); rescue; end
    if File.exist?('ifagent.rb.save')
      FileUtils.copy('ifagent.rb.save', 'ifagent.rb')
      begin; FileUtils.copy('ifagent-api.yaml.save', 'ifagent-api.yaml'); rescue; end
    end
    if File.exist?('ifagent.rb') then File.delete('ifagent.rb.save') end
    if File.exist?('ifagent-api.yaml.save') then File.delete('ifagent-api.yaml.save') end
  end

end
