# Wrapper for LingAgent, adding user commands to modify the score, etc. 
# This class can function both as a command-line program and 
# as an IRC bot when it is loaded by the RedTurnip IRC bot framework.

begin 
   require 'lingagent' # for when we start from the command line
rescue LoadError
   require 'mod/bot/lingagent' # for when we start from RedTurnip 
end

class LingBot 

   def initialize
      print "Hello\n"
      @agent = MyAgent.new
      @THRESHHOLD = 1
      @SCORE_MOD = 0
      @MAX_RESPONSE = 10000
      @MY_NAME = "(?i)(?:lingagentbot|lingbot|lingagent|linguistics|ling)"
   end

   def getResponse(input)

      input = input.to_s
      #print 'INPUT==' + input + "\n"

      # Set threshhold.
      if input =~ /^what is your name?/
         begin
         name = @MY_NAME.gsub!(/\(\?\)\(\?\:/, '')
         name.gsub!(/\|/, '')
         name.gsub!(/\)/, '')
         rescue => err; puts "[getResponse] err==#{err}"; name = @MY_NAME end
         return name.to_s
      end
      if input =~ /^#{@MY_NAME}[:,!]* (quiet|be quiet|stfu|shut.*up)/
         @THRESHHOLD = @THRESHHOLD + 1
         @SCORE_MOD -= 1
         return "My threshhold is now " << @THRESHHOLD.to_s << ", and my score_mod is #{@SCORE_MOD}."
      end
      if input =~ /^#{@MY_NAME}[:,!]* (speak.*up|loud|louder|i can't hear you)/
         @THRESHHOLD = @THRESHHOLD - 1
         @SCORE_MOD += 1
         return "My threshhold is now " << @THRESHHOLD.to_s << ", and my score_mod is #{@SCORE_MOD}."
      end
      # Return threshhold
      if input =~ /^#{@MY_NAME}[:,!]* what is your (threshhold|threshold|volume)/ or input =~ /^#{@MY_NAME}[:,!]* (?:threshold|threshhold|volume)/
         return "My threshhold is currently " << @THRESHHOLD.to_s
      end

      # Provide help message.
      if input =~ /^#{@MY_NAME}[:,!]* help/
         return "I use http://www.deveiate.org/projects/Linguistics to find plurals, infinitives, articles..."
      end

      # trap restart
      case input
        when /^#{@MY_NAME}[:,!]* restart/, /^restart/
          return "That command is disabled." # for safety
      end

      # Get the agent's response.
      if input =~ /^#{@MY_NAME}[:,! ]*(.*)/ then input = $1.strip end
      puts "input==#{input}"
      r, s = @agent.getResponse(input)
      if s == nil then s = 1 end
      #begin; s = @agent.getScore; rescue NoMethodError; s = 1; end
      s += @SCORE_MOD
      if r.to_s.size > @MAX_RESPONSE
         r = r[0..@MAX_RESPONSE]
      end
      puts "\n[lingbot] score==#{s}; threshhold==#{@THRESHHOLD}"
      puts "r == #{r}"  # DEBUG
      if s > @THRESHHOLD 
        if $0 == __FILE__ then return postprocess(r) end
        puts "[lingbot] r==#{postprocess(r)}; s==#{s}"
        return [postprocess(r), s + 1]
      else
         #puts 'r==' + postprocess(r)
      end

      return ""
   end

   def postprocess input 
      r = input
      return r
   end

   # Respond to natural language method calls.
   def method_missing input
     return getResponse(input.to_s)
   end

end

if $0 == __FILE__
  b = LingBot.new
  print("\n> "); $stdout.flush
  quitwords = [":q", "quit", "exit", "bye"]
  while (line = gets) !~ /^#{quitwords.join('|')}$/ and line !~ /^$/
    print "\n#{b.send(line)}\n"
    print("\n> "); $stdout.flush
  end
  print "Bye!"
end
