################################################################################
# linkbot provides an interface to link grammar.
#
# subbot.org/link/linkbot-api.yaml should be in the same directory as this file.
#
# > subject: "John ate an apple."
# John
#
# > what is the object of: "John loves Mary."
# Mary
#
################################################################################

require 'link'
require 'yaml'

$DEBUG = false # @DEBUG also exists! TODO: replace all $DEBUG with @DEBUG.
#$DEBUG = true

class MyAgent 

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/linkbot-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api],
                     [/^begin log (.*)/, :begin_log],
                     [/^end log (.*)/, :end_log]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0 

     @MY_NAME = "(?i)(?:link|linkagent|linkbot|link agent|link bot|link grammar|lgp)"

     @link = Link.new
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s
     grp2 = $2.to_s
     grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     MyAgent.class_eval(method_def) 
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     puts "[link] input==#{input}" unless not $DEBUG

     if input =~ /#{@MY_NAME}[:, ]*(.*)/
       input = process $1
       #puts "[link] input==#{input}" unless not $DEBUG
       if input =~ /^help/ then return help end
     end

     # filter input
     if input =~ /(\w{60,})/
       puts "[link - WORD LENGTH EXCEEDED]"
       return "#{$1} is too long"
     end
     if input =~ /http:\/\//
       puts "[link - URL SKIPPED]"
       #return "Url skipped"
       return '' 
     end
     if input.size > 300
       puts "[link - INPUT LENGTH EXCEEDED]"
       return "Input is too long."
     end
     if input =~ /\001ACTION/
       puts "[link - skipping ACTION]"
       #return "ACTION skipped."
       return '' 
     end
     if input =~ /\%/ then return "Illegal character" end

     @patterns.each { |pattern|
       if input =~ pattern[0]
          r2 = self.send(pattern[1], pattern[0], input) #rescue r2 = r
          puts "[link - #{Time.now}:1] output==#{r2}" unless not $DEBUG
          return r2
       end
     }
     puts "[link - #{Time.now}:2] output==#{r}" unless not $DEBUG
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     #grp1 = %r{^#{$1}}
     grp1 = %r{(?i-mx:^#{$1})}
     #grp2 = %r{^#{$2}}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process s
      str = s.to_s
      str.strip!
      #str.sub!(/^\[/, '')
      #str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     #exec "ruby #{__FILE__}"  # disabled for safety
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def begin_log(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     file = process $1
     file = "log.txt" unless file != nil and file != ''
     $logger = File.open(file, "w")
     t = Time.now; y = t.year; m = t.month; d = t.day; h= t.hour; min = t.min; s = t.sec
     mins = sprintf("%02d", min)
     secs = sprintf("%02d", s)
     time = "#{y}/#{m}/#{d} #{h}:#{mins}:#{secs}"
     return "Okay I have started logging to #{file} at #{time}"
   end # method_def

   def end_log(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     $logger.close
     $logger == nil
     t = Time.now; y = t.year; m = t.month; d = t.day; h= t.hour; min = t.min; s = t.sec
     mins = sprintf("%02d", min)
     secs = sprintf("%02d", s)
     time = "#{y}/#{m}/#{d} #{h}:#{mins}:#{secs}"
     return "Okay I have stopped logging at #{time}."
   end # method_def

   def getSubject(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     return @link.getSubject(grp1)
   end # method_def

   def getObject(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     return @link.getObject(grp1)
   end # method_def

   def getVerb(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     return @link.getV1(grp1)
   end # method_def

   def help(pattern=nil, input=nil)
     puts "[link.help]: pattern==#{pattern}; input==#{input}"
     return 'I parse sentences into subject, verb, object using link grammar.'
   end # method_def

   def getSVO(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     #return @link.getSVO(grp1)
     subject = self.send("subject: #{grp1}")
     verb    = self.send("verb: #{grp1}")
     object  = self.send("object: #{grp1}")
     if subject =~ /default response/i then subject = "NA" end
     if verb    =~ /default response/i then verb = "NA" end
     if verb    =~ /I can't/i then verb = "NA" end
     if object  =~ /default response/i then object = "NA" end
     if object  =~ /I can't/i then object = "NA" end
     return "S=" << subject << "; " << "V=" << verb << "; " << "O=" << object
   end # method_def

   def parse_constituents(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     s = process $1
     return @link.puts(s)
   end # method_def

   def get_simple_parse(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     s = process $1
     return @link.get_simple_parse(s)
   end # method_def

   def show_rules(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     toReturn = @link.show_rules
     return toReturn unless toReturn == nil or toReturn == ''
     return "No rules"
   end # method_def

   def add_rule(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     if rule == '' then return "No rule found." end
     puts "[linkbot.add_rule] rule == #{$1}" #unless @DEBUG == 0
     return @link.add_rule($1.to_s)
   end # method_def

   def delete_rule(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     number = 0
     begin number = $1.to_i; rescue; return "Rule must be a number!"; end
     toReturn = @link.delete_rule(number)
     if toReturn == nil or toReturn == '' then return "Problem deleting rule!" end
     return toReturn
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  quitwords = [":q", "quit", "exit", "bye"]
  while (line = gets) !~ /^#{quitwords.join('|')}/i and line !~ /^$/ 
    begin; $logger.puts("> #{line}"); rescue; end
    response = bot.send(line.to_s).to_s + "\n\n"
    print response
    begin; $logger.puts(response); rescue; end
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
