require 'test/unit'
require 'oppagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('oppagent.rb')
      FileUtils.copy('oppagent.rb', 'oppagent.rb.save')
      begin 
        FileUtils.copy('oppagent-api.yaml', 'oppagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_opposite_rules
    actual, s = @agent.send("add opposite rule: if vp =~ /has/ then vp = \"hasn't\" end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("show opposite rules")
    assert_match(/hasn't/, actual)
  end

  def test_complex_vp_opposite
    actual, s = @agent.send("what is the opposite of \"I do go\"")
    assert_match(/I do not go/i, actual)

    actual, s = @agent.send("what is the opposite of \"I am being alive\"")
    assert_match(/I am not being alive/i, actual)
  end

  def test_opposite
    actual, s = @agent.send("opposite: I eat.")
    assert_match(/I do not eat/, actual)

    actual, s = @agent.send("opposite: I like traffic lights.")
    assert_match(/I do not *like traffic lights/, actual)

    actual, s = @agent.send("opposite: John eats an apple.")
    assert_match(/John does not eat an apple/, actual)

    actual, s = @agent.send("opposite: \"Every person has enough to eat.\"")
    assert_match(/Not every person has enough to eat./, actual)
    #assert_match(/Some persons do not have enough to eat/i, actual)

    actual, s = @agent.send("opposite: I am mortal.")
    assert_match(/I am not mortal/, actual)

    actual, s = @agent.send("opposite: He has a cold.")
    assert_match(/He does not have a cold/, actual)

    #actual, s = @agent.send("what is the opposite of: \"No individuals are altruistic.\"?")
    #assert_match(/Some individuals are altruistic/i, actual)
    #assert_match(/All individuals are altruistic/i, actual)

    #actual, s = @agent.send("what is the contrary of \"All people are rich\"?")
    #assert_match(/No people are rich/i, actual)

    #actual, s = @agent.send("what is the contradiction of \"All people are rich.\"")
    #assert_match(/Some people are rich/i, actual)
  end

  def teardown
    File.delete('oppagent.rb') unless !File.exist?('oppagent.rb.save')
    begin; File.delete('oppagent-api.yaml'); rescue; end
    if File.exist?('oppagent.rb.save')
      FileUtils.copy('oppagent.rb.save', 'oppagent.rb')
      begin; FileUtils.copy('oppagent-api.yaml.save', 'oppagent-api.yaml'); rescue; end
    end
    if File.exist?('oppagent.rb') then File.delete('oppagent.rb.save') end
    if File.exist?('oppagent-api.yaml.save') then File.delete('oppagent-api.yaml.save') end
  end

end
