require 'test/unit'
require 'oppbot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('oppagent.rb')
      FileUtils.copy('oppagent.rb', 'oppagent.rb.save')
      begin 
        FileUtils.copy('oppagent-api.yaml', 'oppagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/oppbot/i, actual)
  end

  def test_volume
    actual, s = @agent.send("OppBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("OppBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("OppBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("OppBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("OppBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("OppBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("OppBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("OppBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('oppagent.rb') unless !File.exist?('oppagent.rb.save')
    begin; File.delete('oppagent-api.yaml'); rescue; end
    if File.exist?('oppagent.rb.save')
      FileUtils.copy('oppagent.rb.save', 'oppagent.rb')
      begin; FileUtils.copy('oppagent-api.yaml.save', 'oppagent-api.yaml'); rescue; end
    end
    if File.exist?('oppagent.rb') then File.delete('oppagent.rb.save') end
    if File.exist?('oppagent-api.yaml.save') then File.delete('oppagent-api.yaml.save') end
  end

end
