################################################################################
# Gets the sunrise time for seattle wa. Also provides some date and number 
# functions.
#
# > what is the porsi time for bellevue wa?
# Sunrise is 6:59 am, so porsi time is 9:59 am
#
# > what is porsi?
# porsi is the Jain practice of not eating until 3 hours after sunrise.
#
# > 
################################################################################

require 'yaml'
require 'time'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent 

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/porsiagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''
     @DEBUG = 0
     @wordnums = ['zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten', 'eleven', 'twelve', 'thirteen', 'fourteen', 'fifteen', 'sixteen', 'seventeen', 'eighteen', 'nineteen']
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     @patterns.push( [@new_regexp, @new_symbol] )
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     puts "input==#{input}" unless @DEBUG == 0
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     @patterns.each { |p| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          break
       end
     }
     return unless method.to_s != nil
     @patterns.push( [grp1, method] )
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       puts "@new_regexp==#{@new_regexp}" unless @DEBUG == 0

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new # Contains methods and patterns from MyAgent-api.yaml.
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
     } # If more than one pattern maps to a method, only the first is stored.
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/ # try to clean regex
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def what_is_the_sunrise_time(pattern=nil, input=nil)
     Dir.chdir("#{@workingDir}/python")
     time = IO.popen("python sunrise.py")
     timearr = time.readlines 
     sunrise = ''
     begin
       sunrise = timearr[1]
     rescue => err; puts "error: #{err}"; sunrise = '' end
     Dir.chdir(@workingDir)
     if sunrise == '' then return "I don't know" end
     sunrise.chomp!; sunrise.strip!
     @score = 4
     begin
       t = Time.parse(sunrise)
     rescue => err; t = err; score -= 2 end
     return t 
   end # method_def

   def what_is_the_porsi_time_for(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     sunrise = self.send("what is the sunrise time?")[0]
     porsi = sunrise + (60 * 60 * 3) # 3 hours after sunrise
     formatted = "..."
     begin
       formatted = porsi.strftime("porsi for %a, %b %d is at %H:%M")
     rescue => err; puts "error!: #{err}"; end
     @score += 4 unless formatted == "..."
     sunrise = sunrise.strftime("%H:%M")
     return "Sunrise is at #{sunrise} so #{formatted}" 
   end # method_def

   def what_is_porsi(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @score += 4
     return "porsi is the Jain practice of fasting until 1/4 day (or 3 hours) after sunrise."
   end # method_def

   def what_is_the_navkarsi_time(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     sunrise = self.send("what is the sunrise time?")[0]
     navkarsi = sunrise + (60 * 48) # 48 minutes after sunrise
     formatted = "..."
     begin
       formatted = navkarsi.strftime("navkarsi for %a, %b %d is at %H:%M")
     rescue => err; puts "error!: #{err}" end
     @score += 4 unless formatted == "..."
     sunrise = sunrise.strftime("%H:%M")
     return formatted << " (48 minutes after sunrise at #{sunrise})" 
   end # method_def

   def what_is_tomorrows_date(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     require 'date'
     today = Date.today
     tomorrow = today + 1
     @score += 3
     return "Tomorrow's date will be #{tomorrow}."
   end # method_def

   def what_was_yesterdays_date(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     require 'date'
     today = Date.today
     yesterday = today - 1
     @score += 3
     return "Yesterday's date was #{yesterday}."
   end # method_def

   def get_name_of_day(now = 0)
     day = ''
     if now == 7 then day = 'Sunday' end
     if now == 1 then day = 'Monday' end
     if now == 2 then day = 'Tuesday' end
     if now == 3 then day = 'Wednesday' end
     if now == 4 then day = 'Thursday' end
     if now == 5 then day = 'Friday' end
     if now == 6 then day = 'Saturday' end
     return day
   end # method_def

   def what_day_of_the_week_is_it(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     now = Time.now.wday
     day = get_name_of_day(now)
     if day == 0 then return "I don't know what day of the week it is." end
     @score += 3 
     return "Today is #{day}."
   end # method_def

   def isNumeric(s); Float(s) != nil rescue false; end
   
   def what_number_comes_after(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     num = nil
     if isNumeric(grp1) then num = grp1.to_i end
     #puts "num==#{num}"
     if num != nil and num != '' then @score += 3; return "After #{grp1} comes #{num.succ}." end

     startofnum = ''; lastwordnum = ''
     if grp1 =~ /^(#{@wordnums.join('|')})$/ then lastwordnum = $1.to_s end
     #puts "startofnum==#{startofnum}; lastwordnum==#{lastwordnum}"
     case grp1 
       when /(.*)\s(#{@wordnums.join('|')})$/, /(.*)[-,](#{@wordnums.join('|')})$/
         startofnum = $1.to_s; lastwordnum = $2.to_s 
     end

     #puts "startofnum==#{startofnum}; lastwordnum==#{lastwordnum}"
     num = @wordnums.index(lastwordnum)
     if num == nil then return "I don't know what comes after #{grp1}." end

     num = num.next
     number = @wordnums[num]
     @score += 3
     return "I think #{startofnum} #{number} comes after #{grp1}."
   end # method_def

   def what_number_comes_before(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     grp1 = process $1
     num = nil
     if isNumeric(grp1) then num = grp1.to_i end
     if num != nil and num != '' then @score += 3; return "Before #{grp1} comes #{num - 1}." end

     startofnum = ''; lastwordnum = ''
     if grp1 =~ /^(#{@wordnums.join('|')})$/ then lastwordnum = $1.to_s end
     #puts "startofnum==#{startofnum}; lastwordnum==#{lastwordnum}"
     case grp1 
     when /(.*)\s(#{@wordnums.join('|')})$/, /(.*)[-,](#{@wordnums.join('|')})$/
         startofnum = $1.to_s; lastwordnum = $2.to_s 
     end
     #puts "startofnum==#{startofnum}; lastwordnum==#{lastwordnum}"
     num = @wordnums.index(lastwordnum)
     if num == nil then return "I don't know what comes before #{grp1}." end

     num = num - 1
     number = @wordnums[num]
     @score += 3
     return "I think #{startofnum} #{number} comes before #{grp1}."
    end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
