require 'test/unit'
require 'porsiagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('porsiagent.rb')
      FileUtils.copy('porsiagent.rb', 'porsiagent.rb.save')
      begin 
        FileUtils.copy('porsiagent-api.yaml', 'porsiagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_time_functions
    actual, s = @agent.send("what day is it today?")
    now = Time.now
    now =~ /^(.*{3,3}) /
    assert_match(/#{$1}/, actual)

    actual, s = @agent.send("what is the date tomorrow?")
    puts actual

    actual, s = @agent.send("what was yesterday's date?")
    puts actual
  end

  def test_number_functions
    actual, s = @agent.send("what number comes after twelve?")
    assert_match(/thirteen/i, actual)

    actual, s = @agent.send("what number comes before one?")
    assert_match(/zero/i, actual)

    actual, s = @agent.send('what number comes after twenty-three?')
    assert_match(/twenty.four/i, actual)

    actual, s = @agent.send("what number comes before 100?")
    assert_match(/99/, actual)
  end

  def teardown
    File.delete('porsiagent.rb') unless !File.exist?('porsiagent.rb.save')
    begin; File.delete('porsiagent-api.yaml'); rescue; end
    if File.exist?('porsiagent.rb.save')
      FileUtils.copy('porsiagent.rb.save', 'porsiagent.rb')
      begin; FileUtils.copy('porsiagent-api.yaml.save', 'porsiagent-api.yaml'); rescue; end
    end
    if File.exist?('porsiagent.rb') then File.delete('porsiagent.rb.save') end
    if File.exist?('porsiagent-api.yaml.save') then File.delete('porsiagent-api.yaml.save') end
  end

end
