require 'test/unit'
require 'PorsiBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('porsiagent.rb')
      FileUtils.copy('porsiagent.rb', 'porsiagent.rb.save')
      begin 
        FileUtils.copy('porsiagent-api.yaml', 'porsiagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/porsibot/, actual)
  end

  def test_volume
    actual, s = @agent.send("PorsiBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("PorsiBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("PorsiBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("PorsiBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("PorsiBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("PorsiBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("PorsiBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("PorsiBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("porsibot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("porsibot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("porsibot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("porsibot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('porsiagent.rb') unless !File.exist?('porsiagent.rb.save')
    begin; File.delete('porsiagent-api.yaml'); rescue; end
    if File.exist?('porsiagent.rb.save')
      FileUtils.copy('porsiagent.rb.save', 'porsiagent.rb')
      begin; FileUtils.copy('porsiagent-api.yaml.save', 'porsiagent-api.yaml'); rescue; end
    end
    if File.exist?('porsiagent.rb') then File.delete('porsiagent.rb.save') end
    if File.exist?('porsiagent-api.yaml.save') then File.delete('porsiagent-api.yaml.save') end
  end

end
