require 'test/unit'
require 'restartbot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('restartagent.rb')
      FileUtils.copy('restartagent.rb', 'restartagent.rb.save')
      begin 
        FileUtils.copy('restartagent-api.yaml', 'restartagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/restartbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("RestartBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("RestartBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("RestartBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("RestartBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("RestartBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("RestartBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("RestartBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("RestartBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('restartagent.rb') unless !File.exist?('restartagent.rb.save')
    begin; File.delete('restartagent-api.yaml'); rescue; end
    if File.exist?('restartagent.rb.save')
      FileUtils.copy('restartagent.rb.save', 'restartagent.rb')
      begin; FileUtils.copy('restartagent-api.yaml.save', 'restartagent-api.yaml'); rescue; end
    end
    if File.exist?('restartagent.rb') then File.delete('restartagent.rb.save') end
    if File.exist?('restartagent-api.yaml.save') then File.delete('restartagent-api.yaml.save') end
  end

end
