require 'test/unit'
require 'sumbot'
require 'FileUtils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('sumagent.rb')
      FileUtils.copy('sumagent.rb', 'sumagent.rb.save')
      begin 
        FileUtils.copy('sumagent-api.yaml', 'sumagent-api.yaml.save') 
      rescue; end
      puts "copied!!!!!!"
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/sumbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("sumbot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("SumBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("sumBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("Sumbot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    # rule # depends on rules hardcoded in
    actual, s = @agent.send("sumbot: delete rule 2") 
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("sumbot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("sumbot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("sumbot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('sumagent.rb') unless !File.exist?('sumagent.rb.save')
    begin; File.delete('sumagent-api.yaml'); rescue; end
    if File.exist?('sumagent.rb.save')
      FileUtils.copy('sumagent.rb.save', 'sumagent.rb')
      begin; FileUtils.copy('sumagent-api.yaml.save', 'sumagent-api.yaml'); rescue; end
    end
    if File.exist?('sumagent.rb') then File.delete('sumagent.rb.save') end
    if File.exist?('sumagent-api.yaml.save') then File.delete('sumagent-api.yaml.save') end
  end

end
