################################################################################
# Tries to chunk sentences into S-V-O, store them in a graph,
# and make inferences or answer questions about them.
# Uses LinkAgent, MontyAgent, and LogicAgent.
#
# > misery depends on ignorance
# Okay
#
# > what depends on ignorance?
# misery depends on ignorance
#
# > what is depended on by misery?
# ignorance is depended on by misery
#
# > does misery depend on ignorance?
# Yes
#
# > why does misery depend on ignorance?
# Because you said: misery depends on ignorance
#
# >
################################################################################

require 'yaml'
require 'graph'
require 'drb'

$Quitwords = [':q', 'quit', 'exit', 'bye']

class MyAgent

   attr_accessor :DEBUG

   def initialize(args=Array.new)
     @workingDir = File.expand_path(File.dirname(__FILE__))
     @filename = @workingDir + "/syllagent-api.yaml"
     unless load_message = load_synonyms
       @patterns = [ [/^(.*?) is a synonym for (.*)/, :add_synonym], 
                     [/^save syn/, :save_synonyms],
                     [/^load syn/, :load_synonyms], 
                     [/^add pattern (.*), (.*)/, :add_pattern], 
                     [/^>(.*)/, :defining_method],
                     [/^restart/, :restart],
                     [/^show api/, :show_api]
                   ]
       save_synonyms # creates patterns.yaml on first use.
     else
       puts load_message 
     end
     @method_def = ''

     @graph = Graph.new

     #@DEBUG = 0 # off
     @DEBUG = 1 # print debug messages to stdout

     DRb.start_service()
     # assume the following services have already been started!
     @monty = DRbObject.new(nil, 'druby://localhost:9070')
     @link  = DRbObject.new(nil, 'druby://localhost:9050')
     @logic = DRbObject.new(nil, 'druby://localhost:9012')
     
   end

   def def_method method_def
     file = IO.read(__FILE__)
     return "Doesn't have 'end # class'" unless file =~ /(.*)end # class(.*)if \$0 == __FILE__(.*)/m
     grp1 = $1.to_s; grp2 = $2.to_s; grp3 = $3.to_s

     new_filename = get_new_filename
     File.open(new_filename, "w+") { |f|
        f.puts grp1 + method_def + "\n\nend # class\n\nif $0 == __FILE__" + grp3
     }
     #@patterns.push( [@new_regexp, @new_symbol] )
     # replace above line with 'smart push'
     @patterns.each_with_index { |e,i|
       found = false
       puts "[def_method]: e.inspect==#{e.inspect}" unless @DEBUG == 0
       pattern = e[0].to_s
       puts "[def_method]: pattern==#{pattern}" unless @DEBUG == 0
       if pattern =~ Regexp.new(Regexp.escape("(?i-mx:(.*)"))
         puts "[def_method]: pattern matched!"
         @patterns.insert(i, [@new_regexp, @new_symbol])
         found = true
       end
       if found then break end
      }
 
     save_synonyms
     puts "#{method_def}"
     MyAgent.class_eval(method_def)
   end

   # If this file has a '.' + a number before the '.rb' extension 
   # (i.e. foo.0.rb), increment the number (i.e. return foo.1.rb).
   # Otherwise return the name of this file unchanged.
   def get_new_filename
      old_filename = __FILE__
      if old_filename =~ /(.*)\.(\d+)\.rb$/
        incr = $2.to_i + 1
        return new_filename = $1.to_s + "." + incr.to_s + ".rb"
      end
      return old_filename  # Dangerous!
   end

   # Match input against each regexp pattern in @patterns; 
   # if there's a match, call the method associated with the pattern.
   def getResponse input
     r = "Default response."

     @score = 2

     #puts "input==#{input}"
     @patterns.each { |pattern|
       if input =~ pattern[0]
          return [self.send(pattern[1], pattern[0], input), @score] rescue return r
       end
     }
     return r
   end

   def load_synonyms(pattern=nil, input=nil)
     if File.exist?(@filename)
        @patterns = YAML.load_file(@filename)
        return "I have loaded #{@filename}."
     end 
     return false
   end

   def save_synonyms(pattern=nil, input=nil)
      r = "Eep! error saving synonyms! "
      begin
        File.open(@filename, 'w') do |out|
          YAML.dump(@patterns, out)
        end
        return "Okay I have saved the synonyms to disk."
      rescue => error
        return r << error
      end
   end

   def add_synonym(pattern, input)
     return "Problem matching pattern." unless input =~ pattern
     grp1 = %r{(?i-mx:^#{$1})}
     grp2 = %r{(?i-mx:^#{$2})}
     method =''
     index = -1 
     @patterns.each_with_index { |p,i| 
       #puts "p[0]==#{p[0]}; p[0].class==#{p[0].class}; grp2==#{grp2}; grp2.class==#{grp2.class}"
       if p[0].to_s == grp2.to_s
          method = p[1]
          index = i
          break
       end
     }
     return unless method.to_s != nil
     #@patterns.push( [grp1, method] )
     # replace above line with 'smart push'
     @patterns.insert(index, [grp1, method])
 
     save_synonyms
     return "Okay, #{grp1} has been added, and will now call #{method}." 
   end

   def add_pattern(pattern=nil, input=nil)
     if input =~ pattern 
       @new_regexp = $1
       @new_symbol = $2
       @new_regexp.gsub!(/\//, '')
       new_re = %r{#{@new_regexp}}

       # Make the regexp case insensitive.
       new_re_s = new_re.to_s
       if new_re_s =~ /(-mix)/
         new_re_s = $~.pre_match << "i-mx" << $~.post_match
       end
       @new_regexp = %r{#{new_re_s}}
       #puts "@new_regexp==#{@new_regexp}"

       method_def = "   def #{@new_symbol}(pattern=nil, input=nil)\n     return \"#\{input} doesn't match #\{pattern}\" unless input =~ pattern\n     return 'Default response from " + @new_symbol + "'\n   end # method_def"
       return def_method(method_def)
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def defining_method(pattern=nil, input=nil)
     if input =~ pattern
       @method_def << "\n " << $1.to_s
       if @method_def =~ /end # method_def/
         return def_method(@method_def)
       end
       return ''
     else
       return "#{input} doesn't match #{pattern}."
     end
   end

   def method_missing input, *rest
     return getResponse(input.to_s) 
   end

   # Strips delimiters.
   def process str
      str.strip!
      str.sub!(/\?$/, '')
      str.sub!(/^\[/, '')
      str.sub!(/\]$/, '')
      str.sub!(/^(?:"|')/, '')
      str.sub!(/(?:"|')$/, '')
      str.sub!(/\.$/, '')
      str.sub!(/;$/, '')
      str.sub!(/,$/, '')
      str.sub!(/!$/, '')
      str.sub!(/\?$/, '')
      str.sub!(/^\//, '')
      str.sub!(/\/$/, '')
      str.sub!(/\"/, '') # strips double quotes from inside the string
      return str 
   end

   def restart(pattern=nil,input=nil)
     exec "ruby #{__FILE__}" 
   end # method_def

   def show_api(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     apiarr= Array.new
     @patterns.each { |p|
       regexp = p[0].to_s; method = p[1].to_s; found = false
       puts "regexp=#{regexp}; method=#{method}" unless @DEBUG == 0
       apiarr.each { |arr| if arr.include? method then found = true end }
       next unless not found
       apiarr.push [regexp, method] 
       puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     }
     # Try to pretty-print.
     toReturn = ''
     puts "apiarr.inspect=#{apiarr.inspect}" unless @DEBUG == 0
     apiarr.each { |arr|
       puts "1. arr=#{arr}" unless @DEBUG == 0
       regex = arr[0]; method = arr[1] 
       if regex =~ /\(\?(?:[i\-mx]*)(?:\:)?(?:\^)?(.*)\)/
         regex = $1.to_s
       end
       if method == nil or method == '' then method = "Default response" end
       toReturn << method.to_s << " | " << regex.to_s << "\n"
     }
     return toReturn
   end # method_def

   def save_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     file = nil 
     begin
       file = process $1
     rescue; end
     puts "\n... Saving graph (this may take minutes) ..."
     r = @graph.save(file)
     return "#{r}"
   end # method_def

   def load_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     begin
       file = process $1
     rescue; end
     return @graph.load(file)
   end

   def reset_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     @graph = Graph.new(false)
     @save = false
     return "Okay, I have reset the graph."
   end # method_def

   def print_graph(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     s = @graph.to_s
     arr2 = Array.new
     arr = s.split("\n")
     #puts "arr==#{arr}"
     arr.each { |phrase|
       phrase = make_verb_agree(phrase)
       arr2 << phrase
     }
     #return s unless s == nil or s.size == 0
     return arr2.join("\n") unless arr2.empty?
     return "The graph is empty."
   end # method_def

   def make_verb_agree(s)
     puts "[make_verb_agree]: s==#{s}" unless @DEBUG == 0
     begin
       r = @logic.send("make verb agree: #{s}")
       puts "[make_verb_agree2]: r==#{r}" unless @DEBUG == 0
       r.chomp!
       if r =~ /(.*)\d+$/ then return $1.to_s end
       puts "[make_verb_agree3]: r==#{r}" unless @DEBUG == 0
       return r
     rescue; return s; end 
   end

   def a_r_b!(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     sentence = process $1
     puts "[a_r_b!] sentence==#{sentence}" unless @DEBUG == 0
     # sanity check
     if sentence =~ /if.*=~.*then.*end/i then return 'NA' end
     # use link to try to chunk input into s-v-o
     begin
      #TODO: the following creates a lot of overhead for the link agent...
      subj = @link.send("subject: #{sentence}")
      verb = @link.send("verb: #{sentence}")
      obj = @link.send("object: #{sentence}")
      puts "[a_r_b!]: subj==#{subj}; verb = #{verb}; obj = #{obj}" unless @DEBUG == 0 
      if subj =~ /don't know/ or subj =~ /can't find/ then subj = '' end
      if verb =~ /don't know/ or verb =~ /can't find/ then verb = '' end
      if obj =~ /don't know/ or obj =~ /can't find/ then obj = '' end
      if subj != '' and verb != '' and obj != ''
       #verb = get_default_form(verb)
       @graph.a_r_b!(subj.downcase, verb.downcase, obj.downcase) 
       return "Okay: #{subj} - #{verb} - #{obj}"
      end
      return 'Default response'
     rescue => e; return "Can't add to graph! #{e.class}"; end
   end # method_def
   
   alias :a_r_b :a_r_b!
   alias :subject_verbs_object :a_r_b!

   def what_does_subject_verb(pattern=nil,input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from what_does_subject_verb'
   end

   alias :what_does_subject_verb? :what_does_subject_verb
   alias :a_r_? :what_does_subject_verb

   def what_verbs_object(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from what_verbs_object'
   end # method_def

   alias what_verbs_object? what_verbs_object

   def what_is_verbed_by_agent(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from what_is_verbed_by_agent'
   end # method_def

   alias what_is_verbed_by_agent? what_is_verbed_by_agent

   def does_subject_verb_object(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from does_subject_verb_object'
   end # method_def
  
   alias does_subject_verb_object? does_subject_verb_object

   def why_does_subject_verb_object(pattern=nil, input=nil)
     return "#{input} doesn't match #{pattern}" unless input =~ pattern
     return 'Default response from why_does_subject_verb_object'
   end # method_def

end # class

if $0 == __FILE__
  bot = MyAgent.new
  print "\n> "; $stdout.flush
  while (line = gets) !~ /^#{$Quitwords.join('|')}$/i and line !~ /^$/
    response = bot.send(line.to_s).to_s
    if response.class == "Array" then response = response[0] end 
    response = response.chop # remove the score
    puts response + "\n\n"
    print "> "; $stdout.flush
  end
  puts "Bye"
end # of file
