require 'test/unit'
require 'syllagent'
require 'fileutils'

BasicSocket.do_not_reverse_lookup = true

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('syllagent.rb')
      FileUtils.copy('syllagent.rb', 'syllagent.rb.save')
      begin 
        FileUtils.copy('syllagent-api.yaml', 'syllagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
    @agent.send("reset graph")
    @agent.DEBUG = 0
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_add_to_graph
    actual, s = @agent.send("X likes chocolate")
    assert_match(/Okay/i, actual)
  end

  def test_questions
    actual, s = @agent.send("samskara depends on ignorance")
    actual, s = @agent.send("what does samskara depend on?")
    assert_match(/ignorance/i, actual)
    actual, s = @agent.send("what depends on ignorance?")
    #assert_match(/samskara/i, actual)
    actual, s = @agent.send("does samskara depend on ignorance?")
    #assert_match(/yes/i, actual)
    actual, s = @agent.send("what is depended on by ignorance?")
    #assert_match(/samskara/i, actual)
  end

  def teardown
    File.delete('syllagent.rb') unless !File.exist?('syllagent.rb.save')
    begin; File.delete('syllagent-api.yaml'); rescue; end
    if File.exist?('syllagent.rb.save')
      FileUtils.copy('syllagent.rb.save', 'syllagent.rb')
      begin; FileUtils.copy('syllagent-api.yaml.save', 'syllagent-api.yaml'); rescue; end
    end
    if File.exist?('syllagent.rb') then File.delete('syllagent.rb.save') end
    if File.exist?('syllagent-api.yaml.save') then File.delete('syllagent-api.yaml.save') end
  end

end
