require 'test/unit'
require 'syllbot'
require 'fileutils'

BasicSocket.do_not_reverse_lookup = true

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('syllagent.rb')
      FileUtils.copy('syllagent.rb', 'syllagent.rb.save')
      begin 
        FileUtils.copy('syllagent-api.yaml', 'syllagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/syllbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("SyllBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("SyllBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("SyllBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("SyllBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("SyllBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("SyllBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("SyllBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("SyllBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('syllagent.rb') unless !File.exist?('syllagent.rb.save')
    begin; File.delete('syllagent-api.yaml'); rescue; end
    if File.exist?('syllagent.rb.save')
      FileUtils.copy('syllagent.rb.save', 'syllagent.rb')
      begin; FileUtils.copy('syllagent-api.yaml.save', 'syllagent-api.yaml'); rescue; end
    end
    if File.exist?('syllagent.rb') then File.delete('syllagent.rb.save') end
    if File.exist?('syllagent-api.yaml.save') then File.delete('syllagent-api.yaml.save') end
  end

end
