require 'test/unit'
require 'testbot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('testagent.rb')
      FileUtils.copy('testagent.rb', 'testagent.rb.save')
      begin 
        FileUtils.copy('testagent-api.yaml', 'testagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/testbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("TestBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("TestBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("TestBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("TestBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("testbot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("testBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("testagent: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("TestAgent: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)
  end

  def teardown
    File.delete('testagent.rb') unless !File.exist?('testagent.rb.save')
    begin; File.delete('testagent-api.yaml'); rescue; end
    if File.exist?('testagent.rb.save')
      FileUtils.copy('testagent.rb.save', 'testagent.rb')
      begin; FileUtils.copy('testagent-api.yaml.save', 'testagent-api.yaml'); rescue; end
    end
    if File.exist?('testagent.rb') then File.delete('testagent.rb.save') end
    if File.exist?('testagent-api.yaml.save') then File.delete('testagent-api.yaml.save') end
  end

end
