Dir.chdir("..")

require 'test/unit'
require 'controlbot'

# Relevant agents must be started before running this test.
class TestControlbot < Test::Unit::TestCase

  def setup
    @agent = ControlBot.new
    name = @agent.send("what is your name?")
    @agent.send("alice: stfu")
    @agent.send("megahal: stfu")
    #@agent.send("#{name}: set delay to 5.0")
    #@agent.send("#{name}: set sleep to 4.0")
  end

  def test_alice
    print "testing alice... "

    #actual = @agent.send("Alice: if response =~ /JUDGE/ then score = 0 end")
    #assert_match(/Okay/, actual)

    actual = @agent.send("alice: loud")
    assert_match(/threshold/, actual)

    actual = @agent.send("alice: help")
    assert_match(/Wallace/, actual)

    actual = @agent.send("alice: hello!")
    assert_match(/(?:hello|hi)/i, actual)

    actual = @agent.send("alice: soft")
    assert_match(/threshhold/, actual)
  end

  def test_link
    print "testing link... "

    actual = @agent.send('link: what is the subject of "I like traffic lights."')
    assert_match(/\bI\b/, actual)
    actual = @agent.send('link: what is the verb in: "I like traffic lights."')
    assert_match(/\blike\b/, actual)
  end

  def test_monty
    print "testing monty... "

    actual = @agent.send('monty: help')
    assert_match(/montylingua/i, actual)

    actual = @agent.send('subject: John ate an apple.')
    assert_match(/John/, actual)

    actual = @agent.send('monty: what is the object in John ate an apple.')
    assert_match(/an apple/, actual)
  end

  def test_ling
    print "testing ling... "

    actual = @agent.send('plural: goose')
    assert_match(/geese/, actual)

    actual = @agent.send('infinitive: begat')
    assert_match(/beget/, actual)
  end

  def test_isr
    print "testing isr... "

    actual = @agent.send('!sr In America, you can always find a party.')
    assert_match(/In Soviet Russia, the party can always find YOU!/, actual)
  end

  def test_bc
    print "testing bc... "

    actual = @agent.send("bcagent: add A recommends silver")
    assert_match(/Okay/, actual)

    actual = @agent.send("bcagent: if A recommends silver, then B buys silver")
    assert_match(/Okay/, actual)

    actual = @agent.send("bcagent: is 'B buys silver' true?")
    assert_match(/Yes/, actual)

    actual = @agent.send("bcbot: if Z is the parent of X and Z is the parent of Y, then X and Y are siblings.")
    assert_match(/Okay/, actual)

    actual = @agent.send("gagent: forget that Z is the parent of X")
    actual = @agent.send("gagent: forget Z is the parent of Y")

    actual = @agent.send("bcbot: is it true that X and Y are siblings")
    assert_match(/is true if/, actual)

    actual = @agent.send("gagent: is Z the parent of X?")
    assert_match(/no/, actual)

    actual = @agent.send("gagent: Z is the parent of X.")
    actual = @agent.send("gagent: Z is the parent of Y.")

    actual = @agent.send("bcbot: is it the case that X and Y are siblings?")
    assert_match(/Yes/, actual)

  end

  def test_megahal
    print "testing megahal... "

    actual = @agent.send("megahal: help")
    assert_match(/Jason Hutchens/, actual)
  end

  def test_wnbot
    print "testing wnbot... "

    actual = @agent.send("wnbot: help")
    assert_match(/Wordnet/i, actual)
  end

  def test_gagent
    print "testing gagent... "

    actual = @agent.send("gagent: help")
    assert_match(/logic/, actual)
  end

  def test_sumagent
    print "testing sumagent... "

    actual = @agent.send("sumagent: summarize: this!")
    assert_match(/this!/, actual)

    actual = @agent.send("sumbot: summarize: that!")
    assert_match(/that!/, actual)

    actual = @agent.send("sumbot: loud")
    actual = @agent.send("sumbot: loud")
    actual = @agent.send("sumbot: loud")
    actual = @agent.send("summarize: the other!")
    assert_match(/the other!/, actual)
    actual = @agent.send("sumbot: quiet")
    actual = @agent.send("sumbot: quiet")
    actual = @agent.send("sumbot: quiet")
  end

  def test_emoagent
    print "testing emoagent... "

    actual = @agent.send("suppose: P")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("P is \"AI succeeds.\"")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("Q is \"People skills become irrelevant.\"")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("P -> Q")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("R is \"I wasted time learning people skills.\"")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("Q -> R")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("S is \"I feel bad.\"")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("R -> S")
    #assert_match(/Okay/i, actual)

    actual = @agent.send("But S is absurd! We cannot have S.")
    assert_match(/Therefore, not-P./i, actual)
  end


  def teardown
    @agent.send("alice: loud")
    @agent.send("megahal: loud")
    @agent.send("gagent: reset graph")
    @agent.send("emobot: reset graph")
  end

end
