require 'test/unit'
require 'TranslateBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('translateagent.rb')
      FileUtils.copy('translateagent.rb', 'translateagent.rb.save')
      begin 
        FileUtils.copy('translateagent-api.yaml', 'translateagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/translatebot/, actual)
  end

  def test_volume
    actual, s = @agent.send("TranslateBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("TranslateBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("TranslateBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("TranslateBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("TranslateBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("TranslateBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("TranslateBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("TranslateBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("translatebot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("translatebot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("translatebot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("translatebot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('translateagent.rb') unless !File.exist?('translateagent.rb.save')
    begin; File.delete('translateagent-api.yaml'); rescue; end
    if File.exist?('translateagent.rb.save')
      FileUtils.copy('translateagent.rb.save', 'translateagent.rb')
      begin; FileUtils.copy('translateagent-api.yaml.save', 'translateagent-api.yaml'); rescue; end
    end
    if File.exist?('translateagent.rb') then File.delete('translateagent.rb.save') end
    if File.exist?('translateagent-api.yaml.save') then File.delete('translateagent-api.yaml.save') end
  end

end
