require 'test/unit'
require 'waagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('waagent.rb')
      FileUtils.copy('waagent.rb', 'waagent.rb.save')
      begin 
        FileUtils.copy('waagent-api.yaml', 'waagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_show_one_method_api
    actual, s = @agent.send("show show_api's api")
    assert_match(/show api/, actual)
  end

  def test_wa
    puts "\n************************************"
    puts "Note: Must be connected to internet!"
    puts "************************************"

    actual, s = @agent.send(".wa newton")
    assert_match(/Isaac Newton/, actual)

    actual, s = @agent.send("who wrote Moby Dick?")
    assert_match(/Melville/, actual)

    actual, s = @agent.send("who starred in \"East of Eden\"?")
    assert_match(/no continuations/i, actual)

    actual, s = @agent.send("what is 90 degrees in radians?")
    assert_match(/pi.*2/, actual)

    actual, s = @agent.send("2x = 10")
    assert_match(/\b5\b/, actual)
  end

  def teardown
    File.delete('waagent.rb') unless !File.exist?('waagent.rb.save')
    begin; File.delete('waagent-api.yaml'); rescue; end
    if File.exist?('waagent.rb.save')
      FileUtils.copy('waagent.rb.save', 'waagent.rb')
      begin; FileUtils.copy('waagent-api.yaml.save', 'waagent-api.yaml'); rescue; end
    end
    if File.exist?('waagent.rb') then File.delete('waagent.rb.save') end
    if File.exist?('waagent-api.yaml.save') then File.delete('waagent-api.yaml.save') end
  end

end
