require 'test/unit'
require 'wabot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('waagent.rb')
      FileUtils.copy('waagent.rb', 'waagent.rb.save')
      begin 
        FileUtils.copy('waagent-api.yaml', 'waagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/wabot/, actual)
  end

  def test_volume
    actual, s = @agent.send("WaBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("wabot: delete rules")

    actual, s = @agent.send("WaBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("WaBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("WaBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("WaBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("WaBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("WaBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("WaBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("wabot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("wabot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("wabot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("wabot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def test_quote_rule
    puts "******************************"
    puts "MUST BE CONNECTED TO INTERNET!"
    puts "******************************"

    actual, s = @agent.send("Who wrote \"Moby Dick\"?")
    assert_match(/Melville/, actual)
  end

  def teardown
    File.delete('waagent.rb') unless !File.exist?('waagent.rb.save')
    begin; File.delete('waagent-api.yaml'); rescue; end
    if File.exist?('waagent.rb.save')
      FileUtils.copy('waagent.rb.save', 'waagent.rb')
      begin; FileUtils.copy('waagent-api.yaml.save', 'waagent-api.yaml'); rescue; end
    end
    if File.exist?('waagent.rb') then File.delete('waagent.rb.save') end
    if File.exist?('waagent-api.yaml.save') then File.delete('waagent-api.yaml.save') end
  end

end
