require 'test/unit'
require 'WikiBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('wikiagent.rb')
      FileUtils.copy('wikiagent.rb', 'wikiagent.rb.save')
      begin 
        FileUtils.copy('wikiagent-api.yaml', 'wikiagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/wikibot/, actual)
  end

  def test_volume
    actual, s = @agent.send("WikiBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("WikiBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("WikiBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("WikiBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("WikiBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("WikiBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("WikiBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("WikiBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("wikibot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("wikibot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("wikibot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("wikibot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('wikiagent.rb') unless !File.exist?('wikiagent.rb.save')
    begin; File.delete('wikiagent-api.yaml'); rescue; end
    if File.exist?('wikiagent.rb.save')
      FileUtils.copy('wikiagent.rb.save', 'wikiagent.rb')
      begin; FileUtils.copy('wikiagent-api.yaml.save', 'wikiagent-api.yaml'); rescue; end
    end
    if File.exist?('wikiagent.rb') then File.delete('wikiagent.rb.save') end
    if File.exist?('wikiagent-api.yaml.save') then File.delete('wikiagent-api.yaml.save') end
  end

end
