# Provides the top-level interface to Wordnet. 
# This class can function both as a command-line program and 
# as an IRC bot when it is loaded by the RedTurnip IRC bot framework.

begin 
   require 'myagent' # for when we start from the command line
rescue LoadError
   require 'mod/bot/myagent' # for when we start from RedTurnip 
end

class WNBot 

   def initialize
      print "Hello\n"
      @agent = MyAgent.new
      @THRESHHOLD = 1
      @MAX_RESPONSE = 10000
      @MY_NAME = "(?i)(wnbot|wnagent|wordnet|wordnetagent)"
      @SCORE_MOD = 0

      @rules = [] 
      @rules.push("if response =~ /\\(\\d+\\)/ then response.gsub!(/\\(\\d+\\)/,'') end")
      @rules.push("if response =~ /\\(/ then response.gsub!(/\\(/, '') end")
      @rules.push("if response =~ /\\)/ then response.gsub!(/\\)/, '') end")
      @rules.push("if input =~ /^(?:!)?(?:define|dict)/ then score += 2 end")
      @SAFE_LEVEL = "low"
      @Low_safe_level_words = ["low", "default", "0", "off"]
      @High_safe_level_words = ["high", "on", "1"]
   end

   def getResponse(input)

      input = input.to_s
      #print 'INPUT==' + input + "\n"

      # Add post-processing code.
      if input =~ /^#{@MY_NAME} add (.*) to PostProc/
         return "That feature is disabled."
      end

      # Set threshhold.
      if input =~ /^what is your name?/
         name = @MY_NAME
         name.gsub!(/\(\?i\)/, '')
         name.gsub!(/[\(\)]/, '')
         name.gsub!(/\|/, ' or ')
         return name 
      end
      if input =~ /^#{@MY_NAME}[:,!]* (quiet|be quiet|stfu|shut.*up|soft)/
         @THRESHHOLD += 1; @SCORE_MOD -= 1
         return "My threshhold is now #{@THRESHHOLD} and the score_mod is #{@SCORE_MOD}."
      end
      if input =~ /^#{@MY_NAME}[:,!]* (speak.*up|loud|louder|i can't hear you)/
         @THRESHHOLD -= 1; @SCORE_MOD += 1
         return "My threshhold is now #{@THRESHHOLD} and my score_mod is #{@SCORE_MOD}."
      end
      # Return threshhold
      if input =~ /^#{@MY_NAME}[:,!]* what is your (threshhold|threshold|volume)/
         return "My threshhold is currently #{@THRESHHOLD}. Score_mod is #{@SCORE_MOD}."
      end

      # Provide help message.
      if input =~ /^#{@MY_NAME}.? help$/
         return "I provide an interface to Wordnet." 
      end

      # trap restart
      case input
        when /^#{@MY_NAME}[:,!]* restart$/, /^restart$/
          #return "That command is disabled."
          exec("ruby #{__FILE__}")
      end

      # add new if-then rules to post-process responses
      case input
        when /^#{@MY_NAME}[:,!]* (if .*? =~ .*? then .*)$/
          return add_if_rule($1)
      end

      # show the rules
      case input
        when /^#{@MY_NAME}[:,!]*.* show.* rules/
          return show_rules[0]
      end

      # delete all rules
      case input
        when /^#{@MY_NAME}[:,!]*.* delete.* rules/
          return delete_rules
      end

      # delete a rule
      case input
        when /^#{@MY_NAME}[:,!]*.* delete.* rule.* (\d+)/
          return delete_rule($1.to_i)
      end

      # set safe level
      case input
        when /^#{@MY_NAME}[:,! ]*(?:set )?.*safe(?:ty)? level (?:to )?(.*)/
          return set_safe_level($1)
      end

      # get safe level
      case input 
        when /^#{@MY_NAME}[:,! ]*what is.*?safe(?:ty)? level/
          return  get_safe_level
      end

      # Get the agent's response.
      r = @agent.getResponse(input)
      if r.to_s.size > @MAX_RESPONSE
         r = r[0..@MAX_RESPONSE]
      end
      s = @agent.getScore
      r, s = apply_if_rules(r,s,input)
      s += @SCORE_MOD
      puts "\n[wnbot] score==" << s.to_s << "; threshhold==" << @THRESHHOLD.to_s
      if s > @THRESHHOLD 
         return [postprocess(r), s]
      else
         print 'r==' + postprocess(r) + " | #{s}"
      end

      return ""
   end

   def postprocess input 
      r = input
      return r
   end

   # Respond to natural language method calls.
   def method_missing input
     return getResponse(input.to_s)
   end

   # add an if-then rule
   def add_if_rule(rule = '')
     if rule == '' then return "No rule found." end
     if @SAFE_LEVEL =~ /#{@Low_safe_level_words.join('|')}$/ 
       @rules.push(rule)
       return "Okay I have added #{rule}."
     end
     return "I can't add any if rules because my safe level is #{@SAFE_LEVEL}."
   end

   # apply if-then rules
   def apply_if_rules(response, score, input)
     @rules.each { |rule|
       #if @DEBUG > 0
         puts "rule==#{rule}; response==#{response}; score==#{score}"
       #end
       begin
         eval(rule)
       #rescue; puts "Can't apply rule: #{rule}" unless @DEBUG == 0; next; end
       rescue; puts "Can't apply rule: #{rule}"; next; end
     }
     #puts "returning: response==#{response}; score==#{score}"
     return [response, score]
   end

   # return the rules
   def show_rules
     return [@rules.join("\n"), 4]
   end

   # delete a rule by its index
   def delete_rule(i)
     if i < 1 or i > @rules.size
       return "The rules go from 1 to #{@rules.size}."
     end
     rule = @rules[i-1]
     @rules.delete_at(i-1)
     return "Okay, \"#{rule}\" deleted."
   end

   # delete all rules
   def delete_rules
     @rules = []
     return "Okay I have deleted all the rules."
   end

   def set_safe_level(level)
     case level 
       when /\b#{@Low_safe_level_words.join('|')}\b/ 
         @SAFE_LEVEL = "off"
         return "Okay I have turned off the safe level."
       when /\b#{@High_safe_level_words.join('|')}\b/
         @SAFE_LEVEL = "on"
         return "Okay I have turned the safe level on."
     end
     return "I don't understand #{level}."
   end

   def get_safe_level
     return "My safe level is set to #{@SAFE_LEVEL}."
   end # method_def

end


class PostProc
   private_class_method :new
   @@postproc = nil
   @@count = 0
   
   def PostProc.create
      @@postproc = new unless @@postproc
      @@postproc
   end

   def process(input)
      response = input.to_s 
      self.methods.each { |method|
         if method =~ /^newMethod/
            begin
               response = self.send(method, input)
            rescue NoMethodError => boom
               return "Eeep! Error!: " + boom
            end
         end
      }
      return response
   end

   def getNewName
      @@count = @@count + 1
      return "newMethod" << @@count.to_s
   end

   def create_method(name, &block)
      self.class.send(:define_method, name, &block)
   end

end

if $0 == __FILE__
  b = WNBot.new
  print("\n> "); $stdout.flush
  quitwords = [":q", "quit", "exit", "bye"]
  while (line = gets) !~ /^#{quitwords.join('|')}$/ and line !~ /^$/
    r, = b.send(line)
    print "\n" + r + "\n"
    print("\n> "); $stdout.flush
  end
  print "Bye!"
end
