//File: hunters.cpp
//this file contains the main driver program for the hunter simulation
//Programer: Andrew Schlei
//
//Ver 1.0

#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>
#include <windows.h>
#include "Hunter.h"

using namespace std;
//==================FUNCTION PROTOTYPES============================
void displayWorld(int* world,int WORLD_SIZE_Y,int WORLD_SIZE_X,
				  Hunter* hunters);

//=========================MAIN====================================
void main()
{
	//constants
	const int EMPTY = -2;			//used in world matrix
	const int FOOD = -1;			//used in world matrix
	const int X_MAX = 75;			//maximum x dimension
	const int Y_MAX = 22;			//maximum y dimension
	
	//variables entered by user that are not changed by the program
	int WORLD_SIZE_X = 20;		//X dimension of the world
	int WORLD_SIZE_Y = 20;		//Y dimension of the world
	int NUM_HUNTERS = 100;		//total number of hunter agents
	int INITIAL_PLACED_FOOD=100;//amount of food placed in world at start
	int FOOD_VALUE = 10;			//amount of food gained by picking up placed food
	int H_STARTING_FOOD = 20;	//amount of food each agent starts with
	int FOOD_CONSUMED = 1;		//amount of food each hunter must consume 
	int COST_ATTACK = 3;		//amount of food it costs to attack
	int UTILITY_ATTACK = 0;	//utility gained from attacking
	int UTILITY_SHARE = 10;		//utility gained from sharing
	int UTILITY_HOARD = 0;		//utility gained from hoarding
	int VISION_RANGE = 3;		//vision range of the agents
	int MAX_POWER = 20;			//Maximum power for any agent
	int NUM_GENERATIONS = 10;	//number of generations to run
	int NUM_ROUNDS = 30;		//number of rounds in each generation
	int SLEEP_AMOUNT = 100;		//amount of time in ms to sleep	

	//other variables
	int i,gen,rd;			//LCVs
	int randX, randY;		//store randomly generated numbers
	int numlA,numlS,numlH;	//Number of living agents attacking, sharing, and hoarding
	int numdA,numdS,numdH;	//Number of dead agents attacking, sharing, and hoarding
	int avgPower;			//Average power for all agents
	int avgPowA,avgPowS,avgPowH;	//Average power for attacking, sharing, and hoarding
	bool totalExtinction;	//becomes true if all agents in a generation die
	bool error;				//used to detect input errors
	ofstream genOutput;		//output stream for generation summary output
	ofstream logOutput;		//output stream log file




	//seed random number generator with time
	srand((unsigned)time(NULL));




	//get simulation parameters from user
	//get WORLD_SIZE_X
	do
	{
		error = false;
		cout << "Enter x dimension of world (min 1, max "<< X_MAX <<"): ";
		cin >> WORLD_SIZE_X;
		if(WORLD_SIZE_X < 1 || WORLD_SIZE_X > X_MAX || !cin) 
		{
			error = true;
			cout << "Invalid dimension." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}

	} while(error);

	//get WORLD_SIZE_Y
	do
	{
		error = false;
		cout << "Enter y dimension of world (min 1, max "<< Y_MAX <<"): ";
		cin >> WORLD_SIZE_Y;
		if(WORLD_SIZE_Y < 1 || WORLD_SIZE_Y > Y_MAX || !cin)
		{
			error = true;
			cout << "Invalid dimension." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	} while(error);

	cout << "Your world has " << WORLD_SIZE_X*WORLD_SIZE_Y << " total grid squares." << endl;

	//get NUM_HUNTERS
	do
	{
		error = false;
		cout << "Enter the number of hunters you want: ";
		cin >> NUM_HUNTERS;
		if(NUM_HUNTERS < 1 || NUM_HUNTERS > WORLD_SIZE_X*WORLD_SIZE_Y || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get INITIAL_PLACED_FOOD
	do
	{
		error = false;
		cout << "Enter the number of food cells to place each generation: ";
		cin >> INITIAL_PLACED_FOOD;
		if(INITIAL_PLACED_FOOD < 0 || INITIAL_PLACED_FOOD > (WORLD_SIZE_X*WORLD_SIZE_Y)-NUM_HUNTERS || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get FOOD_VALUE
	do
	{
		error = false;
		cout << "Enter the value of each food cell: ";
		cin >> FOOD_VALUE;
		if(FOOD_VALUE < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get H_STARTING_FOOD
	do
	{
		error = false;
		cout << "Enter amount of food each hunter starts with: ";
		cin >> H_STARTING_FOOD;
		if(H_STARTING_FOOD < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get FOOD_CONSUMED
	do
	{
		error = false;
		cout << "Enter amount of food hunters consume each round: ";
		cin >> FOOD_CONSUMED;
		if(FOOD_CONSUMED < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get COST_ATTACK
	do
	{
		error = false;
		cout << "Enter the material cost of attacking (in food): ";
		cin >> COST_ATTACK;
		if(COST_ATTACK < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get UTILITY_ATTACK
	do
	{
		error = false;
		cout << "Enter the utility offset of attacking: ";
		cin >> UTILITY_ATTACK;
		if(!cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get UTILITY_SHARE
	do
	{
		error = false;
		cout << "Enter the utility offset of sharing: ";
		cin >> UTILITY_SHARE;
		if(!cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get UTILITY_HOARD
	do
	{
		error = false;
		cout << "Enter the utility offset of hoarding: ";
		cin >> UTILITY_HOARD;
		if(!cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get VISION_RANGE
	do
	{
		error = false;
		cout << "Enter the vision range of hunters (min 1): ";
		cin >> VISION_RANGE;
		if(VISION_RANGE < 1 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get MAX_POWER
	do
	{
		error = false;
		cout << "Enter the maximum power an agent can have (min 0): ";
		cin >> MAX_POWER;
		if(MAX_POWER < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get NUM_GENERATIONS
	do
	{
		error = false;
		cout << "Enter the number of generations to run (min 1): ";
		cin >> NUM_GENERATIONS;
		if (NUM_GENERATIONS < 1 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get NUM_ROUNDS
	do
	{
		error = false;
		cout << "Enter the number of rounds in each generation (min 1): ";
		cin >> NUM_ROUNDS;
		if (NUM_ROUNDS < 1 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);

	//get SLEEP_AMOUNT
	do
	{
		error = false;
		cout << "Enter sleep delay (in milliseconds, min 0): ";
		cin >> SLEEP_AMOUNT;
		if(SLEEP_AMOUNT < 0 || !cin)
		{
			error = true;
			cout << "Invalid number." << endl;
			cin.clear();
			cin.ignore(100,'\n');
		}
	}while(error);


	//create world matrix
	int* world = new int[WORLD_SIZE_X*WORLD_SIZE_Y];	//stores information about the world
	for (i=0; i< WORLD_SIZE_X*WORLD_SIZE_Y; i++)			
			world[i] = EMPTY;				

	


	//create and initilize array of hunter agents
	Hunter* hunters = new Hunter[NUM_HUNTERS];
	for(i=0; i<NUM_HUNTERS; i++)
	{
		hunters[i].setPower(rand() % (MAX_POWER+1));
		hunters[i].setID(i);
		hunters[i].setFoodCount(H_STARTING_FOOD);
		hunters[i].setAlive(true);
		hunters[i].setStrengthAttack(0);
		hunters[i].setStrengthShare(0);
		hunters[i].setStrengthHoard(0);
		hunters[i].createInteract(NUM_HUNTERS);
		hunters[i].setCostAttack(COST_ATTACK);
		hunters[i].removeTies();
		hunters[i].setUtilityAttack(UTILITY_ATTACK);
		hunters[i].setUtilityShare(UTILITY_SHARE);
		hunters[i].setUtilityHoard(UTILITY_HOARD);
		hunters[i].setVision(VISION_RANGE);
		hunters[i].setMetabolism(FOOD_CONSUMED);
	}




	//display the world
	displayWorld(world, WORLD_SIZE_Y, WORLD_SIZE_X,hunters);
	Sleep(SLEEP_AMOUNT);


	
	
	//open output files
	genOutput.open("gen_output.dat");
	logOutput.open("log_output.dat");

	
	
	
	//Run the simulation
	//= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
	logOutput << "*****Running the simulation...*****" << endl;
	totalExtinction = false;
	gen=0;
	while(gen < NUM_GENERATIONS && !totalExtinction)
	{//Generation Loop
		
		logOutput << "*****Generation " << gen + 1 << " starting...*****" << endl << endl;

		//initialize world
		for (i=0; i< WORLD_SIZE_X*WORLD_SIZE_Y; i++)			
			world[i] = EMPTY;
		logOutput << "World Initialized." << endl;
		
		//setup hunters and place randomly
		for(i=0; i<NUM_HUNTERS; i++)
		{
			hunters[i].setAlive(true);
			hunters[i].setFoodCount(H_STARTING_FOOD);
			hunters[i].removeTies();
			do
			{
				randX = rand() % WORLD_SIZE_X;
				randY = rand() % WORLD_SIZE_Y;
			} while (world[randY*WORLD_SIZE_X+randX] != EMPTY);
			hunters[i].setPos(randX,randY);
			world[randY*WORLD_SIZE_X+randX] = i;
		}
		logOutput << "Hunters Placed." << endl;

		//place initial food randomly
		for(i=0; i<INITIAL_PLACED_FOOD; i++)
		{
			do
			{
				randX = rand() % WORLD_SIZE_X;
				randY = rand() % WORLD_SIZE_Y;
			} while (world[randY*WORLD_SIZE_X+randX] != EMPTY);
			world[randY*WORLD_SIZE_X+randX] = FOOD;
		}
		logOutput << "Food Placed." << endl;

		//pre-run calculations and output
		numlA = 0;
		numlS = 0;
		numlH = 0;
		for(i=0; i < NUM_HUNTERS; i++)
		{
			if(hunters[i].isAttacking()) numlA++;
			else if (hunters[i].isSharing()) numlS++;
			else numlH++;
		}
		genOutput << "***Generation " << gen+1 << "***" << endl;
		genOutput << "Stategies at beginning" << endl;
		genOutput << "----------------------" << endl;
		genOutput << "Number of Agents attacking: "	<< numlA << endl;
		genOutput << "Number of Agents sharing: "	<< numlS << endl;
		genOutput << "Number of Agents hoarding: "	<< numlH << endl;
		genOutput << endl;

		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		for (rd=0; rd<NUM_ROUNDS; rd++)
		{//Round Loop
			
			logOutput << "**Round "  << rd + 1 << " starting...**" << endl;

			//initialize interaction arrays
			for (i=0; i< NUM_HUNTERS; i++)
			{
				hunters[i].initInteract(NUM_HUNTERS);
			}

			//hunters move and gather food
			for (i=0; i < NUM_HUNTERS; i++)
			{
				if (hunters[i].isAlive())
				{
					hunters[i].move(world,WORLD_SIZE_X,WORLD_SIZE_Y,EMPTY,FOOD);
					hunters[i].gather(world,WORLD_SIZE_X,WORLD_SIZE_Y,EMPTY,FOOD,FOOD_VALUE);
				}
			}

			//hunters interact with each other
			for (i=0; i < NUM_HUNTERS; i++)
			{
				hunterInteraction(world,WORLD_SIZE_Y,WORLD_SIZE_X,hunters,i,logOutput);
			}

			//hunters consume food and declared dead
			for (i=0; i < NUM_HUNTERS; i++)
			{
				if (hunters[i].isAlive())
				{
					hunters[i].setFoodCount(hunters[i].getFoodCount()-hunters[i].getMetabolism());
					if(hunters[i].getFoodCount() < 0)	//hunter has died
					{
						hunters[i].setAlive(false);
						world[hunters[i].getY()*WORLD_SIZE_X+hunters[i].getX()] = EMPTY;
					}
				} 
			}

			//display the world
			displayWorld(world,WORLD_SIZE_Y,WORLD_SIZE_X,hunters);

			Sleep(SLEEP_AMOUNT);

			logOutput << "**...Round " << rd + 1 << " ending.**" << endl;
		}//End Round Loop
		//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		//post-run calculations and output
		numlA = 0;
		numdA = 0;
		numlS = 0;
		numdS = 0;
		numlH = 0;
		numdH = 0;
		avgPower = 0;
		avgPowA = 0;
		avgPowS = 0;
		avgPowH = 0;
		for(i=0; i < NUM_HUNTERS; i++)
		{
			avgPower += hunters[i].getPower();
			if(hunters[i].isAttacking()) 
			{
				avgPowA += hunters[i].getPower();
				if(hunters[i].isAlive())
					numlA++;
				else
					numdA++;
			}
			else if (hunters[i].isSharing())
			{
				avgPowS += hunters[i].getPower();
				if (hunters[i].isAlive())
					numlS++;
				else
					numdS++;
			}
			else
			{
				avgPowH += hunters[i].getPower();
				if (hunters[i].isAlive())
					numlH++;
				else
					numdH++;
			}
		}
		avgPower = avgPower / NUM_HUNTERS;
		if (numlA+numdA > 0) avgPowA = avgPowA / (numlA+numdA);
		if (numlS+numdS > 0) avgPowS = avgPowS / (numlS+numdS);
		if (numlH+numdH > 0) avgPowH = avgPowH / (numlH+numdH);

		genOutput << "Stategies at end" << endl;
		genOutput << "----------------------" << endl;
		genOutput << "Number of Agents attacking: "	<< numlA+numdA << " (Alive: " 
				  << numlA << ", Dead: " << numdA << ") (Average Power: " << avgPowA << ")" << endl;
		genOutput << "Number of Agents sharing: "	<< numlS+numdS << " (Alive: "
				  << numlS << ", Dead: " << numdS << ") (Average Power: " << avgPowS << ")" << endl;
		genOutput << "Number of Agents hoarding: "	<< numlH+numdH << " (Alive: "
				  << numlH << ", Dead: " << numdH << ") (Average Power: " << avgPowH << ")" << endl;
		genOutput << "Overall Average Power: " << avgPower << endl;
		genOutput << "Total Survival Rate: " << (double((double(numlA+numlS+numlH))/(double (NUM_HUNTERS)))*100) 
				  << "%" << endl;
		genOutput << endl << endl;

		//Evolution/reproduction stage
		if (numlA+numlS+numlH > 0)
		{
			hunterEvolution(hunters,NUM_HUNTERS,numlA+numlS+numlH);
		}
		else
		{
			totalExtinction = true;
			genOutput << "Simulation ended due to total extinction of all agents." << endl;
		}
		
		logOutput << endl << "*****...Generation " << gen + 1 << " ending.*****" << endl;
		gen++;

	}//End Generation Loop
	//= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 

	logOutput << "*****...ending the simulation*****" << endl;

	
	
	
	//cleanup
	cout << endl;
	genOutput.close();
	logOutput.close();
	delete [] hunters;
	delete [] world;
}




//=========================FUNCTIONS=================================
void displayWorld(int* world,int WORLD_SIZE_Y,int WORLD_SIZE_X, 
				  Hunter* hunters)
//this function is used to display the world on the screen
{
	const EMPTY = -2;
	const FOOD = -1;
	system("cls");
	for (int i=0; i< WORLD_SIZE_Y*WORLD_SIZE_X; i++)			
	{		
		switch(world[i])
		{
			case EMPTY:
				cout << " ";
				break;
			case FOOD:
				cout << "+";
				break;
			default:
				if(hunters[world[i]].isAttacking()) cout << "A";
				else if (hunters[world[i]].isSharing()) cout << "S";
				else if (hunters[world[i]].isHoarding()) cout << "H";
				else cout << "N";
		}
		if ((i+1)%(WORLD_SIZE_X)==0) cout << endl;
	}
}


