require 'test/unit'
require 'huntersagent'
require 'fileutils'

class TestMyAgent < Test::Unit::TestCase

  def setup
    if File.exist?('huntersagent.rb')
      FileUtils.copy('huntersagent.rb', 'huntersagent.rb.save')
      begin 
        FileUtils.copy('huntersagent-api.yaml', 'huntersagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyAgent.new
  end 

  def test_defining_hello
    actual, s = @agent.send("add pattern /^hello/, hello")
    
    actual, s = @agent.send("hello")

    actual, s = @agent.send(">  def hello(pattern=nil, input=nil)")
    actual, s = @agent.send(">    if @count == nil; @count = 1; return 'I am ready for my first lesson.'; end")
    actual, s = @agent.send(">    return 'Hello again.'")
    actual, s = @agent.send(">  end # method_def")

    actual, s = @agent.send("hello")
    assert_match(/lesson/, actual)

    actual, s = @agent.send("hi is a synonym for hello")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("hi")
    assert_match(/again/, actual)
  end

  def test_show_api
    actual, s = @agent.send("show api")
    assert_match(/ is a synonym for /, actual)
  end

  def test_show_one_method_api
    actual, s = @agent.send("show show_api's api")
    assert_match(/show api/, actual)
  end

  def test_run_hunters
    actual, s = @agent.send("run hunters with default of 10")
    assert_match(/----------/, actual)

    actual, s = @agent.send("show output")
    # NOTE: "Strategies" misspelled in hunters.exe
    assert_match(/Stategies at end/m, actual.join('\n'))

    actual, s = @agent.send("repeat last run, please")
    assert_match(/----------/, actual)

    actual, s = @agent.send("show output")
    assert_match(/Generation 10/m, actual.join('\n'))
  end

  def teardown
    File.delete('huntersagent.rb') unless !File.exist?('huntersagent.rb.save')
    begin; File.delete('huntersagent-api.yaml'); rescue; end
    if File.exist?('huntersagent.rb.save')
      FileUtils.copy('huntersagent.rb.save', 'huntersagent.rb')
      begin; FileUtils.copy('huntersagent-api.yaml.save', 'huntersagent-api.yaml'); rescue; end
    end
    if File.exist?('huntersagent.rb') then File.delete('huntersagent.rb.save') end
    if File.exist?('huntersagent-api.yaml.save') then File.delete('huntersagent-api.yaml.save') end
  end

end
