require 'test/unit'
require 'HuntersBot'
require 'fileutils'

class TestMyBot < Test::Unit::TestCase

  def setup
    if File.exist?('huntersagent.rb')
      FileUtils.copy('huntersagent.rb', 'huntersagent.rb.save')
      begin 
        FileUtils.copy('huntersagent-api.yaml', 'huntersagent-api.yaml.save') 
      rescue; end
    end
    @agent = MyBot.new
  end

  def test_name
    actual, s = @agent.send("what is your name?")
    assert_match(/huntersbot/, actual)
  end

  def test_volume
    actual, s = @agent.send("HuntersBot: what is your volume?")
    assert_match(/1/, actual)
  end

  def test_rules
    actual, s = @agent.send("HuntersBot: loud")
    assert_match(/0/, actual)

    actual, s = @agent.send("HuntersBot: if response =~ /default response/i then response = 'foo!' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/foo!/, actual)

    actual, s = @agent.send("HuntersBot: if response =~ /foo!/ then response = 'bar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/bar/, actual)

    actual, s = @agent.send("HuntersBot: delete rule 1")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/default response/i, actual)

    actual, s = @agent.send("HuntersBot: if response =~ /default response/i then response = 'fubar' end")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("testing")
    assert_match(/fubar/, actual)

    actual, s = @agent.send("HuntersBot: delete all rules")
    assert_match(/Okay/, actual)

    actual, s = @agent.send("HuntersBot: show rules")
    assert_match(//, actual)

    actual, s = @agent.send("testing")
    assert_match(/Default response/, actual)

    actual, s = @agent.send("what is the debug mode?")
    assert_match(/off/i, actual)

    actual, s = @agent.send("huntersbot: turn debug mode on")
    assert_match(/on/i, actual)

    actual, s = @agent.send("huntersbot: what is the debug mode?")
    assert_match(/on/i, actual)

    actual, s = @agent.send("huntersbot: please turn off the debug mode.")
    assert_match(/off/i, actual)

    actual, s = @agent.send("huntersbot: what is the debug mode?")
    assert_match(/off/i, actual)
  end

  def teardown
    File.delete('huntersagent.rb') unless !File.exist?('huntersagent.rb.save')
    begin; File.delete('huntersagent-api.yaml'); rescue; end
    if File.exist?('huntersagent.rb.save')
      FileUtils.copy('huntersagent.rb.save', 'huntersagent.rb')
      begin; FileUtils.copy('huntersagent-api.yaml.save', 'huntersagent-api.yaml'); rescue; end
    end
    if File.exist?('huntersagent.rb') then File.delete('huntersagent.rb.save') end
    if File.exist?('huntersagent-api.yaml.save') then File.delete('huntersagent-api.yaml.save') end
  end

end
